/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: base_utils.cc,v 1.4 2007/01/01 03:17:25 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/base_utils.h"

#include <cassert>
#include <cstdlib>


using namespace Ochusha;

namespace Ochusha
{
}


#define BASE64_PAD '='


static char *
encode_base64(const unsigned char *src, size_t length,
	      const char * const alphabet)
{
  if (src == NULL)
    return NULL;

  size_t result_length = (length + 2) / 3 * 4;
  char *result = static_cast<char *>(malloc(result_length + 1));
  if (result == NULL)
    return NULL;	// out of memory

  size_t full_groups = length / 3;
  const unsigned char *tail = src + full_groups * 3;

  char *dst = result;
  while (src < tail)
    {
      *dst++ = alphabet[src[0] >> 2];
      *dst++ = alphabet[(src[0] << 4 | src[1] >> 4) & 0x3f];
      *dst++ = alphabet[(src[1] << 2 | src[2] >> 6) & 0x3f];
      *dst++ = alphabet[src[2] & 0x3f];
      src += 3;
    }
  assert(src == tail);

  size_t trailing = length % 3;
  if (trailing == 1)
    {
      *dst++ = alphabet[*src >> 2];
      *dst++ = alphabet[(*src << 4) & 0x3f];
      *dst++ = BASE64_PAD;
      *dst++ = BASE64_PAD;
    }
  else if (trailing == 2)
    {
      *dst++ = alphabet[src[0] >> 2];
      *dst++ = alphabet[(src[0] << 4 | src[1] >> 4) & 0x3f];
      *dst++ = alphabet[(src[1] << 2) & 0x3f];
      *dst++ = BASE64_PAD;
    }

  *dst = '\0';
  return result;
}


static const char *const base64_alphabet =
  "ABCDEFGHIJKLMNOP"
  "QRSTUVWXYZabcdef"
  "ghijklmnopqrstuv"
  "wxyz0123456789+/";


char *
encode_to_base64(const char *src, size_t length)
{
  return encode_base64(reinterpret_cast<const unsigned char *>(src), length,
		       base64_alphabet);
}


static const char *const base64url_alphabet =
  "ABCDEFGHIJKLMNOP"
  "QRSTUVWXYZabcdef"
  "ghijklmnopqrstuv"
  "wxyz0123456789-_";


char *
encode_to_base64url(const char *src, size_t length)
{
  return encode_base64(reinterpret_cast<const unsigned char *>(src), length,
		       base64url_alphabet);
}


#define IV 0x80
#define PD 0x40


static char *
decode_base64(const unsigned char *src, size_t length,
	      const unsigned char * const dictionary)
{
  if (length % 4 != 0)
    return NULL;

  size_t result_length = length / 4 * 3;
  char *result = static_cast<char *>(malloc(result_length + 1));
  const unsigned char *tail = src + length;
  char *dst = result;

  while (src < tail)
    {
      if (*src & 0x80)
	break;	// invalid source
      unsigned int part = dictionary[*src];
      if (part & (IV | PD))
	break;	// invalid source
      unsigned int word = part << 18;

      if (src[1] & 0x80)
	break;	// invalid source
      part = dictionary[src[1]];
      if (part & (IV | PD))
	break;	// invalid source
      word |= part << 12;

      if (src[2] & 0x80)
	break;	// invalid source
      part = dictionary[src[2]];
      if (part & IV)
	break;	// invalid source
      int num_pad;
      if (part != PD)
	{
	  word |= part << 6;
	  if (src[3] & 0x80)
	    break;	// invalid source
	  part = dictionary[src[3]];
	  if (part & IV)
	    break;	// invalid source
	  if (part != PD)
	    {
	      word |= part;
	      num_pad = 0;
	    }
	  else
	    num_pad = 1;
	}
      else
	{
	  if (src[3] == BASE64_PAD)
	    num_pad = 2;
	  else
	    break;	// invalid source
	}
      *dst++ = word >> 16;
      if (num_pad < 2)
	{
	  *dst++ = (word >> 8) & 0xff;
	  if (num_pad == 0)
	    *dst++ = word & 0xff;
	}

      src += 4;
    }
  if (src == tail)
    {
      *dst = '\0';
      return result;
    }

  // invalid source
  free(result);
  return NULL;
}


static const unsigned char base64_dictionary[] = {
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, 62, IV, IV, IV, 63,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, IV, IV, IV, PD, IV, IV,
  IV,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, IV, IV, IV, IV, IV,
  IV, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, IV, IV, IV, IV, IV,
};


char *
decode_from_base64(const char *src, size_t length)
{
  return decode_base64(reinterpret_cast<const unsigned char *>(src), length,
		       base64_dictionary);
}


static const unsigned char base64url_dictionary[] = {
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, 62, IV, IV,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, IV, IV, IV, PD, IV, IV,
  IV,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, IV, IV, IV, IV, 63,
  IV, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, IV, IV, IV, IV, IV,
};


char *
decode_from_base64url(const char *src, size_t length)
{
  return decode_base64(reinterpret_cast<const unsigned char *>(src), length,
		       base64url_dictionary);
}
