/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_thread.cc,v 1.43 2007/01/04 06:30:19 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/bulletin_board.h"
#include "libochusha/bbs_thread.h"

#include "libochusha/bbs_2ch_utils.h"
#include "libochusha/bbs_jbbs_utils.h"
#include "libochusha/configuration.h"
#include "libochusha/text_utils.h"
#include "libochusha/misc_utils.h"

#include <cstdlib>
#include <cstring>

#include <iostream>

#include <zlib.h>


#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


namespace Ochusha
{


#undef VERBOSE_DESTRUCTOR


Response::Response()
  : name(NULL), mailto(NULL), date(NULL), id(NULL), message(NULL), title(NULL),
    flags(0)
{
}


Response::~Response()
{
  if (name != NULL)
    free(name);

  if (mailto != NULL)
    free(mailto);

  if (date != NULL)
    free(date);

  if (id != NULL)
    free(id);

  if (message != NULL)
    free(message);

  if (title != NULL)
    free(title);
}


void
Response::set_name(const char *value)
{
  if (name != NULL)
    free(name);
  name = value != NULL ? strdup(value) : NULL;
}


void
Response::set_mailto(const char *value)
{
  if (mailto != NULL)
    free(mailto);
  mailto = value != NULL ? strdup(value) : NULL;
}


void
Response::set_date(const char *value)
{
  if (date != NULL)
    free(date);
  date = value != NULL ? strdup(value) : NULL;
}


void
Response::set_id(const char *value)
{
  if (id != NULL)
    free(id);
  id = value != NULL ? strdup(value) : NULL;
}


void
Response::set_message(const char *value)
{
  if (message != NULL)
    free(message);
  message = value != NULL ? strdup(value) : NULL;
}


void
Response::set_title(const char *value)
{
  if (title != NULL)
    free(title);
  title = value != NULL ? strdup(value) : NULL;
}


ResponseCursor::ResponseCursor(BBSThread *thread,
			       Configuration &config, Repository &repository,
			       Buffer *buffer, NetworkAgent *agent,
			       const char *bbs_enc)
  : thread(thread),
    config(config),
    repository(repository),
    iconv_buffer(IconvBuffer::UTF8_ENCODING, bbs_enc),
    buffer(buffer),
    agent(agent)
{
  // ͤξ硢㳰ŪüʾȳƥϹ⡹1000쥹
  // ʤΤǡ2ɤʤhints1000餤ǽ˼äƤޤ
  hints.reserve(1024);
  agent->access_started.connect(this, &ResponseCursor::access_started_cb);
  agent->access_progressed.connect(this, &ResponseCursor::access_progressed_cb);
  agent->access_finished.connect(this, &ResponseCursor::access_finished_cb);
  agent->cache_is_dirty.connect(this, &ResponseCursor::cache_is_dirty_cb);
  agent->access_failed.connect(this, &ResponseCursor::access_failed_cb);
  agent->access_terminated.connect(this, &ResponseCursor::access_terminated_cb);
}


ResponseCursor::~ResponseCursor()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "ResponseCursor(@%p) destructed\n", this);
#endif
}


void
ResponseCursor::clear_hints()
{
  thread->set_number_of_responses_got(0);
  thread->set_dat_filesize(0);
  int lock_id = buffer->get_lock()->wrlock();
  while (!hints.empty())
    hints.pop_back();
  buffer->get_lock()->unlock(lock_id);
}


void
ResponseCursor::set_hints(size_t res_num, int offset, size_t length)
{
  assert(hints.size() >= res_num - 1);
  assert((offset & RESPONSE_HINT_MASK) == 0);
  thread->set_number_of_responses_got(res_num);
  thread->set_dat_filesize(offset + length);
  if (hints.capacity() >= res_num)
    {
      if (hints.size() == res_num - 1)
	hints.push_back(ResponseHint(offset, length));
      else
	hints[res_num - 1] = ResponseHint(offset, length);
      return;
    }
  assert(hints.size() == res_num - 1);
  int lock_id = buffer->get_lock()->wrlock();
  hints.push_back(ResponseHint(offset, length));
  buffer->get_lock()->unlock(lock_id);
}


void
ResponseCursor::set_broken(size_t res_num, int offset, size_t length)
{
  assert(hints.size() >= res_num - 1);
  assert((offset & RESPONSE_HINT_MASK) == 0);
  if (thread->get_number_of_responses_got() < static_cast<int>(res_num))
    {
      thread->set_number_of_responses_got(res_num);
      thread->set_dat_filesize(offset + length);
    }
  if (hints.capacity() >= res_num)
    {
      if (hints.size() == res_num - 1)
	hints.push_back(ResponseHint(RESPONSE_HINT_BROKEN | offset, length));
      else
	hints[res_num - 1]
	  = ResponseHint(RESPONSE_HINT_BROKEN | offset, length);
      return;
    }
  assert(hints.size() == res_num - 1);
  int lock_id = buffer->get_lock()->wrlock();
  hints.push_back(ResponseHint(RESPONSE_HINT_BROKEN | offset, length));
  buffer->get_lock()->unlock(lock_id);
}


void
ResponseCursor::set_deleted(size_t res_num, int offset)
{
  assert(hints.size() >= res_num - 1);
  assert((offset & RESPONSE_HINT_MASK) == 0);
  thread->set_number_of_responses_got(res_num);
  if (hints.capacity() >= res_num)
    {
      if (hints.size() == res_num - 1)
	hints.push_back(ResponseHint(RESPONSE_HINT_DELETED | offset, 0));
      else
	hints[res_num - 1] = ResponseHint(RESPONSE_HINT_DELETED | offset, 0);
      return;
    }

  assert(hints.size() == res_num - 1);
  int lock_id = buffer->get_lock()->wrlock();
  hints.push_back(ResponseHint(RESPONSE_HINT_DELETED | offset, 0));
  buffer->get_lock()->unlock(lock_id);
}


int
ResponseCursor::get_next_offset()
{
  size_t res_num = hints.size();
  if (res_num > 0)
    {
      size_t prev = res_num - 1;
      return (hints[prev].offset & ~RESPONSE_HINT_MASK) + hints[prev].length;
    }
  return 0;
}


bool
ResponseCursor::has_response(size_t res_num, int lock_id)
{
#ifndef NDEBUG
  if (hints.size() >= res_num)
    {
      assert((hints[res_num - 1].offset & RESPONSE_HINT_DELETED) != 0
	      || hints[res_num - 1].length > 0);
      return true;
    }
  return false;
#else
  return hints.size() >= res_num;
#endif
}


Response &
ResponseCursor::get_response(size_t res_num, int lock_id)
{
  assert(has_response(res_num, lock_id));
  size_t index = res_num - 1;

  int offset = hints[index].offset;
  if ((offset & RESPONSE_HINT_MASK) == 0)
    {
      response.clear_flags();
      if (parse_response(res_num, buffer->get_buffer(lock_id) + offset,
			 hints[index].length))
	return response;

      set_broken(res_num, offset, hints[res_num].length);

      response.set_broken();
      return response;
    }

  switch (offset & RESPONSE_HINT_MASK)
    {
    case RESPONSE_HINT_BROKEN:
      response.set_broken();
      return response;

    case RESPONSE_HINT_DELETED:
      response.set_deleted();
      return response;
    }

  // never reached here
#ifndef NDEBUG
  assert(0);
#else
  response.set_not_ready();
  return response;
#endif
}


bool
ResponseCursor::finished(int lock_id)
{
  return buffer->is_fixed();
}


void
ResponseCursor::access_started_cb(NetworkAgent *agent, RefCount *data)
{
  access_started();
}


void
ResponseCursor::access_progressed_cb(NetworkAgent *agent, size_t data_read,
				     uintmax_t total_len, RefCount *data)
{
  access_progressed();
}


void
ResponseCursor::cache_is_dirty_cb(NetworkAgent *agent, RefCount *data)
{
  cache_is_dirty();
}


void
ResponseCursor::access_finished_cb(NetworkAgent *agent, RefCount *data)
{
  access_finished();
}


void
ResponseCursor::access_failed_cb(NetworkAgent *agent, const char *reason,
				 RefCount *data)
{
  access_failed();
}


void
ResponseCursor::access_terminated_cb(NetworkAgent *agent, RefCount *data)
{
  access_terminated();
}


BBSThread::BBSThread(BulletinBoard *board, const char *id, const char *title)
  : board(board),
    id(NULL),
    title(NULL),
    last_modified(NULL),
    date(NULL),
    kako_html(NULL),
    grave(NULL),
    rank(0),
    previous_rank(0),
    number_of_responses_on_server(0),
    number_of_responses_got(0),
    number_of_responses_read(0),
    number_of_new_responses(0),
    bookmark_position(0),
    flags(0),
    ui_flags(0),
    last_name(NULL),
    last_mail(NULL),
    dat_filesize(0),
    alive(false)
{
  set_id(id);
  set_title(title);
}


BBSThread::~BBSThread()
{
  if (id != NULL)
    free(id);

  if (title != NULL)
    free(title);

  if (last_modified != NULL)
    free(last_modified);

  if (date != NULL)
    free(date);

  if (kako_html != NULL)
    free(kako_html);

  if (grave != NULL)
    free(grave);

  if (last_name != NULL)
    free(last_name);

  if (last_mail != NULL)
    free(last_mail);
}


ResponseCursor *
BBSThread::get_responses(Configuration &config,
			 Repository &repository,
			 ThreadPool &pool,
			 NetworkAgentBroker &broker)
{
  assert(board != NULL);
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      return new ResponseCursor2chDAT(this, config, repository,
				      broker.employ_agent());

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      return new ResponseCursorMachiBBS(this, config, repository,
					broker.employ_agent());

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return new ResponseCursorJbbsLivedoor(this, config, repository,
					    broker.employ_agent());

#if 0
    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_MITINOKU:
      return new ResponseCursorJBBS(this, config, repository,
				    broker.employ_agent());
#endif
    }
  return NULL;
}


PostResponseAgent *
BBSThread::get_post_agent(Configuration &config, NetworkAgentBroker &broker)
{
  if (config.get_offline())
    return NULL;

  assert(board != NULL);
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
      return new PostResponseAgent2ch(this, config, broker.employ_agent());

    case OCHUSHA_BBS_TYPE_MACHIBBS:
    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return new PostResponseAgentJbbs(this, config, broker.employ_agent());

#if 0
    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_MITINOKU:
#endif
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      break;
    }
  return NULL;
}


int
BBSThread::get_flags() const
{
  return flags;
}


void
BBSThread::set_flags(int value)
{
  flags = value;
}


void
BBSThread::set_number_of_responses_got(int value)
{
  if (number_of_responses_got == value)
    return;

  number_of_responses_got = value;
  get_board()->notify_minor_modification();
}


void
BBSThread::set_number_of_responses_read(int value)
{
  if (number_of_responses_read == value)
    return;

  number_of_responses_read = value;
  get_board()->notify_minor_modification();
}


void
BBSThread::set_bookmark_position(int value)
{
  if (bookmark_position == value)
    return;
  bookmark_position = value;
  get_board()->notify_major_modification();
}


void
BBSThread::set_dat_dropped(bool value)
{
  flags = (flags & ~DAT_DROPPED) | (value ? DAT_DROPPED : 0);
}


void
BBSThread::set_stopped(bool value)
{
  flags = (flags & ~STOPPED) | (value ? STOPPED : 0);
}


void
BBSThread::set_kako_thread(bool value)
{
  flags = (flags & ~(KAKO_THREAD | DAT_DROPPED))
    | (value ? (KAKO_THREAD | DAT_DROPPED) : 0);
}


void
BBSThread::set_viewer_used(bool value)
{
  flags = (flags & ~(DAT_VIEWER | DAT_DROPPED))
    | (value ? (DAT_VIEWER | DAT_DROPPED) : 0);
}


void
BBSThread::set_id(const char *id)
{
  if (this->id != NULL)
    free(this->id);
  this->id = id != NULL ? strdup(id) : NULL;
}


void
BBSThread::set_title(const char *value)
{
  if (title != NULL)
    free(title);

  title = value != NULL ? wipe_string(value) : NULL;
}


void
BBSThread::set_last_modified(const char *value)
{
  if (last_modified != NULL)
    free(last_modified);
  last_modified = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_date(const char *value)
{
  if (date != NULL)
    free(date);
  date = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_kako_html(const char *value)
{
  if (kako_html != NULL)
    free(kako_html);
  kako_html = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_grave(const char *value)
{
  if (grave != NULL)
    free(grave);
  grave = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_marked(bool value)
{
  ui_flags = (ui_flags & ~MARKED) | (value ? MARKED : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_hidden(bool value)
{
  ui_flags = (ui_flags & ~HIDDEN) | (value ? HIDDEN : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_post_sage(bool value)
{
  ui_flags = (ui_flags & ~LAST_POST_SAGE) | (value ? LAST_POST_SAGE : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_post_be(bool value)
{
  ui_flags = (ui_flags & ~LAST_POST_BE) | (value ? LAST_POST_BE : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_post_viewer(bool value)
{
  ui_flags = (ui_flags & ~LAST_POST_VIEWER) | (value ? LAST_POST_VIEWER : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_name(const char *value)
{
  if (last_name != NULL)
    free(last_name);
  last_name = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_mail(const char *value)
{
  if (last_mail != NULL)
    free(last_mail);
  last_mail = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_number_of_responses_on_server(int value)
{
  if (number_of_responses_on_server < value)
    number_of_new_responses = value - number_of_responses_on_server;
  else
    number_of_new_responses = 0;
  number_of_responses_on_server = value;
}


// ʲget_dat_filename(),open_dat_file(),remove_dat_file()ѡ
// url_bufferPATH_MAXХʬݤƤȲꤷƤ롣
static char *
get_dat_url(BBSThread *thread, char *url_buffer)
{
  assert(url_buffer != NULL);
  assert(thread != NULL);
  BulletinBoard *board = thread->get_board();
  assert(board != NULL);
  URI base_uri(board->get_base_url());
  if (!base_uri.is_valid()
      || base_uri.get_server() == NULL || base_uri.get_path() == NULL)
    return NULL;

  const char *id = thread->get_id();
  if (id == NULL)
    return NULL;
  if (snprintf(url_buffer, PATH_MAX, "%sdat/%s.dat", board->get_base_url(), id)
      < PATH_MAX)
    return url_buffer;
  return NULL;
}


char *
BBSThread::excavate_dat_file(Repository &repository)
{
  BulletinBoard *board = get_board();
  assert(board != NULL);

  URI base_uri(board->get_base_url());
  if (!base_uri.is_valid())
    return NULL;

  char url_buffer[PATH_MAX];
  list<char *>::const_iterator iter = board->get_movement_history_begin();
  list<char *>::const_iterator end_iter = board->get_movement_history_end();
  BBSDataManager &bbs_data = repository.get_bbs_data();
  while (iter != end_iter)
    {
      if (snprintf(url_buffer, PATH_MAX, "%s://%s%sdat/%s.dat",
		   base_uri.get_scheme(), *iter, base_uri.get_path(), get_id())
	  < PATH_MAX)
	{
	  char *filename = bbs_data.find_file(url_buffer);
	  if (filename != NULL)
	    return filename;
	}
      ++iter;
    }

  return NULL;
}


size_t
BBSThread::get_dat_filesize(Repository &repository)
{
  if (dat_filesize != 0)
    return dat_filesize;

  char url_buffer[PATH_MAX];
  char *url = get_dat_url(this, url_buffer);
  char *dat_path;
  if (url == NULL)
    return 0;

  dat_path = repository.get_bbs_data().find_file(url);
  if (dat_path == NULL)
    dat_path = excavate_dat_file(repository);

  if (dat_path == NULL)
    return 0;

  struct stat stat_buf;
  if (stat(dat_path, &stat_buf) != 0)
    stat_buf.st_size = 0;

  free(dat_path);

  set_dat_filesize(stat_buf.st_size);
  return get_dat_filesize();
}


int
BBSThread::open_dat_file(Repository &repository, int flags)
{
  char url_buffer[PATH_MAX];
  char *url = get_dat_url(this, url_buffer);
  if (url == NULL)
    return -1;

  int fd = repository.get_bbs_data().open_file(url, flags);
  if (fd >= 0 || (flags & O_TRUNC) != 0)
    return fd;

  char *filename = excavate_dat_file(repository);
  if (filename == NULL)
    return -1;

  fd = open(filename, flags);

  if (is_alive())
    repository.get_bbs_data().move_file(filename, url);

  free(filename);
  return fd;
}


void
BBSThread::remove_dat_file(Repository &repository)
{
  char url_buffer[PATH_MAX];
  char *url = get_dat_url(this, url_buffer);
  if (url == NULL)
    return;

  char *filename = excavate_dat_file(repository);
  if (filename != NULL)
    repository.get_bbs_data().move_file(filename, url);

  repository.get_bbs_data().invalidate(url);
  return;
}


char *
BBSThread::get_url(int from, int to)
{
  return get_board()->get_url(this, from, to);
}


bool
BBSThread::is_trackback_supported()
{
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return true;

    default:
      break;
    }

  return false;
}


char *
BBSThread::get_trackback_url(int res_num)
{
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
      return BBS2ch::get_trackback_url(this, res_num);

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return BBSJbbsLivedoor::get_trackback_url(this, res_num);

    default:
      break;
    }

  return NULL;
}


bool
BBSThread::is_a_boned(size_t res_num)
{
  list<size_t>::const_iterator iter = a_bone_list.begin();
  list<size_t>::const_iterator end_iter = a_bone_list.end();
  while (iter != end_iter)
    if (*iter++ == res_num)
      return true;

  return false;
}


void
BBSThread::register_a_bone(size_t res_num)
{
  list<size_t>::iterator iter = a_bone_list.begin();
  list<size_t>::iterator end_iter = a_bone_list.end();
  while (iter != end_iter)
    {
      if (*iter == res_num)
	return;

      if (*iter > res_num)
	break;
      ++iter;
    }
  if (iter != end_iter)
    a_bone_list.insert(iter, res_num);
  else
    a_bone_list.push_back(res_num);

  get_board()->notify_major_modification();
}


void
BBSThread::unregister_a_bone(size_t res_num)
{
  a_bone_list.remove(res_num);

  get_board()->notify_major_modification();
}


time_t
BBSThread::get_last_modified_utc()
{
  if (last_modified != NULL)
    return get_utc_from_date(last_modified);
  return -1;
}


time_t
BBSThread::get_birthtime()
{
  // ߤΤȤIDbirthtimeˤʤäƤ褦ʷǼĤˤбƤʤ
  if (id != NULL)
    {
      long value;	// time_t32bitɡġ
      if (sscanf(id, "%ld", &value) == 1)
	return static_cast<time_t>(value);
    }

  return -1;
}


void
BBSThread::reset(Repository &repository)
{
  set_last_modified(NULL);
  set_date(NULL);
  set_kako_html(NULL);
  set_grave(NULL);
  set_number_of_responses_got(0);
  set_number_of_responses_read(0);
  set_dat_filesize(0);
  flags &= ~(DAT_VIEWER | KAKO_THREAD | STOPPED);
  ui_flags = 0;
  set_last_name(NULL);
  set_last_mail(NULL);
  remove_dat_file(repository);
}


void
BBSThread::read_threadlist_element(AttributeTable &attributes)
{
  set_last_modified(attributes.get_string("last_modified"));
  set_kako_html(attributes.get_string("kako_html"));
  set_grave(attributes.get_string("grave"));

  rank = attributes.get_int("rank");
  previous_rank = rank;
  number_of_responses_on_server
    = attributes.get_int("number_of_responses_on_server");

  if (attributes.lookup("number_of_responses_read") == NULL)
    {
      number_of_responses_got
	= attributes.get_int("number_of_responses_got");
      number_of_responses_read = number_of_responses_got;
    }
  else
    {
      number_of_responses_got
	= attributes.get_int("number_of_responses_read");
      number_of_responses_read = number_of_responses_got;
    }

  bookmark_position = attributes.get_int("bookmark_position");

  flags = attributes.get_int("flags");
  ui_flags = attributes.get_int("ui_flags");

  set_last_name(attributes.get_string("last_name"));
  set_last_mail(attributes.get_string("last_mail"));

  const char *a_bone_list = attributes.get_string("a_bone_list");
  if (a_bone_list != NULL)
    {
      const char *cur_pos = a_bone_list;
      while (*cur_pos != '\0')
	{
	  unsigned int res_num;
	  if (sscanf(cur_pos, "%u", &res_num) == 1)
	    register_a_bone(static_cast<size_t>(res_num));
	  cur_pos = strchr(cur_pos, ',');
	  if (cur_pos == NULL)
	    break;
	  ++cur_pos;
	}
    }

  a_bone.read_xml_elements(attributes);
  read_extended_threadlist_element(attributes);
}


#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute)			\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "      <attribute name=\"" #attribute "\">\n"		\
		 "        <boolean val=\"%s\"/>\n"			\
		 "      </attribute>\n",				\
		 (attribute) ? "true" : "false");			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute)				\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "      <attribute name=\"" #attribute "\">\n"		\
		 "        <int val=\"%d\"/>\n"				\
		 "      </attribute>\n", attribute);			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "      <attribute name=\"" #attribute "\">\n"	\
		   "        <string>%s</string>\n"			\
		   "      </attribute>\n",				\
		   append_markup_escaped_text(text, attribute).c_str());\
	}								\
    }									\
  while (0)


void
BBSThread::write_threadlist_element(gzFile threadlist_xml)
{
  if (number_of_responses_got == 0 && (flags & DAT_DROPPED))
    return;	// DATեäƤʤDATϤ줤äѤ˺롣

  gzprintf(threadlist_xml, "    <thread>\n");

  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, id);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, title);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, last_modified);

  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, kako_html);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, grave);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, rank);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, number_of_responses_on_server);
  OUTPUT_ATTRIBUTE_INT(threadlist_xml, number_of_responses_got);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, bookmark_position);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, flags);
  OUTPUT_ATTRIBUTE_INT(threadlist_xml, ui_flags);

  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, last_name);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, last_mail);

  list<size_t>::const_iterator iter = a_bone_list.begin();
  list<size_t>::const_iterator end_iter = a_bone_list.end();

  string tmp_string;
  char tmp_buf[64];
  while (iter != end_iter)
    {
      snprintf(tmp_buf, 64, "%u", static_cast<unsigned int>(*iter));
      if (iter == a_bone_list.begin())
	tmp_string.append(tmp_buf);
      else
	tmp_string.append(",").append(tmp_buf);
      ++iter;
    }
  const char *a_bone_list = tmp_string.c_str();
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, a_bone_list);

  a_bone.write_xml_elements(threadlist_xml, "      ");
  write_extended_threadlist_element(threadlist_xml);

  gzprintf(threadlist_xml, "    </thread>\n");
} 


void
BBSThread::read_extended_threadlist_element(AttributeTable &attributes)
{
}


void
BBSThread::write_extended_threadlist_element(gzFile threadlist_xml) const
{
}


PostResponseAgent::PostResponseAgent(BBSThread *thread, Configuration &config,
				     NetworkAgent *agent)
  : config(config),
    thread(thread),
    agent(agent)
{
}


PostResponseAgent::~PostResponseAgent()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "PostResponseAgent(@%p) destructed\n", this);
#endif
}


void
PostResponseAgent::set_2ch_be_session_id(const char *id)
{
}


void
PostResponseAgent::set_2ch_viewer_session_id(const char *id)
{
}

}
