/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: lock.cc,v 1.18 2007/01/01 03:17:25 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/lock.h"

#include <cassert>
#include <cerrno>
#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>

#include <pthread.h>

#if defined(HAVE_CLOCK_GETTIME) || defined(HAVE_GETTIMEOFDAY)
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# endif
#endif


namespace Ochusha
{

#undef VERBOSE_DESTRUCTOR

static bool lock_manager_destructed = false;
static LockManager lock_manager;


RWLock *
LockManager::rwlock_new()
{
  if (lock_manager_destructed)
    abort();
  return lock_manager.get_rwlock();
}


LockManager::~LockManager()
{
  lock_manager_destructed = true;
  while (!free_rwlock_list.empty())
    {
      RWLock *rwlock = free_rwlock_list.front();
      free_rwlock_list.pop_front();
      if (rwlock->ref_count != 1)
	cerr << "Something wrong with RWLock(@" << rwlock << ")\n";
      delete rwlock;
    }
}


Lock::Lock(Lock *slave)
  : count(0),
    slave(slave)
{
  if (pthread_mutex_init(&mutex, NULL) != 0)
    {
      cerr << "Couldn't initialize a mutex.\n";
      abort();
    }
  memset(&locker, 0, sizeof(pthread_t));
}


Lock::~Lock()
{
  if (pthread_mutex_destroy(&mutex) != 0)
    {
      cerr << "Couldn't destroy a mutex.\n";
      //abort();
    }
}


void
Lock::lock()
{
  if (count > 0 && pthread_equal(locker, pthread_self()))
    {
      ++count;
      return;
    }

#ifdef CHECK_LOCK_ORDER
#if 0
  // recursive2ܰʹߤOKʤΤǡ1ܤlockΤߥå
  if (slave != NULL
      && slave->count > 0 && pthread_equal(slave->locker, pthread_self()))
    {
      cerr << "Lock Order Reversal: slave must be unlocked before locking global.\n";
      abort();
    }
#endif
#endif

  int result;
  if ((result = pthread_mutex_lock(&mutex)) == 0)
    {
      assert(count == 0);
      locker = pthread_self();
      // ˥Хꥢɬפʵġ
#warning "Insert memory barrier here!"
      count = 1;
      return;
    }

  fprintf(stderr, "Couldn't lock a mutex: %s(%d)\n", strerror(result), result);
  abort();
}


bool
Lock::trylock()
{
  if (count > 0 && pthread_equal(locker, pthread_self()))
    {
      ++count;
      return true;
    }

#ifdef CHECK_LOCK_ORDER
  // recursive2ܰʹߤOKʤΤǡ1ܤlockΤߥå
  if (slave != NULL
      && slave->count > 0 && pthread_equal(slave->locker, pthread_self()))
    {
      cerr << "Lock Order Reversal: slave must be unlocked before locking global.\n";
      abort();
    }
#endif

  int result;
  if ((result = pthread_mutex_trylock(&mutex)) == 0)
    {
      assert(count == 0);
      locker = pthread_self();
      // ˥Хꥢɬפʵġ
#warning "Insert memory barrier here!"
      count = 1;
      return true;
    }

  assert(result == EBUSY);
  return false;
}


void
Lock::unlock()
{
  if (count < 1 || !pthread_equal(locker, pthread_self()))
    {
#if 0
      cerr << "Couldn't unlock a Lock held by another thread!\n";
      abort();
#endif
    }

  if (--count > 0)
    return;

  assert(count == 0);
  memset(&locker, 0, sizeof(pthread_t));

  int result;
  if ((result = pthread_mutex_unlock(&mutex)) == 0)
    return;

  fprintf(stderr, "Couldn't unlock a mutex: %s(%d)\n",
	  strerror(result), result);
  abort();
}


Lock *
Lock::clone()
{
  return new Lock();
}


bool
Lock::is_locked_by_self()
{
  return count > 0 && pthread_equal(locker, pthread_self()) != 0;
}


#ifdef CHECK_LOCK_ORDER
Lock *
Lock::get_slave_lock()
{
  static Lock slave_lock;
  return &slave_lock;
}


Lock *
Lock::get_global_lock()
{
  static Lock *global_lock = NULL;
  if (global_lock == NULL)
    global_lock = new Lock(get_slave_lock());
  return global_lock;
}
#endif


Monitor::Monitor()
{
  if (pthread_cond_init(&cond, NULL) != 0)
    {
      cerr << "Couldn't initialize a condition.\n";
      abort();
    }
}


Monitor::~Monitor()
{
  if (pthread_cond_destroy(&cond) != 0)
    {
      cerr << "Couldn't destroy a condition.\n";
      //abort();
    }
}


Lock *
Monitor::clone()
{
  return new Monitor();
}


bool
Monitor::wait()
{
  assert(is_locked_by_self());

  int prev_count = count;
  count = 0;

  if (pthread_cond_wait(&cond, &mutex) != 0)
    {
      cerr << "Couldn't wait a condition.\n";
      abort();
    }

  assert(count == 0);

  locker = pthread_self();
  // ˥Хꥢɬפʵġ
#warning "Insert memory barrier here!"
  count = prev_count;
  return true;
}


bool
Monitor::wait(long ms)
{
  assert(is_locked_by_self());

#if defined(HAVE_CLOCK_GETTIME) || defined(HAVE_GETTIMEOFDAY)
  struct timespec abstime;
  if (ms > 0)
    {
# ifdef HAVE_CLOCKGETTIME
      if (clock_gettime(CLOCK_REALTIME, &abstime) == 0)
	{
	  long nsec = abstime.tv_nsec + (ms % 1000) * 1000000;
	  abstime.tv_sec += (ms / 1000) + (nsec / 1000000000);
	  abstime.tv_nsec = nsec % 1000000000;
	}
      else
	ms = 0;
# else
      struct timeval abstime_now;
      if (gettimeofday(&abstime_now, NULL) == 0)
	{
	  long nsec;
	  abstime.tv_sec = abstime_now.tv_sec;
	  abstime.tv_nsec = abstime_now.tv_usec * 1000;

	  nsec = abstime.tv_nsec + (ms % 1000) * 1000000;
	  abstime.tv_sec += (ms / 1000) + (nsec / 1000000000);
	  abstime.tv_nsec = nsec % 1000000000;
	}
      else
	ms = 0;
# endif
    }

  if (ms == 0)
    return wait();

  int prev_count = count;
  count = 0;

  int result = pthread_cond_timedwait(&cond, &mutex, &abstime);
  if (result != 0 && result != ETIMEDOUT && result != EINTR)
    {
      fprintf(stderr, "Couldn't wait a condition: %s(%d)\n",
	      strerror(result), result);
      abort();
    }

  assert(count == 0);

  locker = pthread_self();
  // ˥Хꥢɬפʵġ
#warning "Insert memory barrier here!"
  count = prev_count;

  return result == 0;
#endif
  return wait();
}


void
Monitor::notify()
{
  if (pthread_cond_signal(&cond) != 0)
    {
      cerr << "Couldn't signal a condition.\n";
      abort();
    }
}


void
Monitor::notify_all()
{
  if (pthread_cond_broadcast(&cond) != 0)
    {
      cerr << "Couldn't broadcast a condition.\n";
      abort();
    }
}


RWLock::RWLock() : ref_count(1)
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_init(&rwlock, NULL) != 0)
    {
      cerr << "Couldn't initialize an rwlock.\n";
      abort();
    }
#else
  if (pthread_mutex_init(&mutex, NULL) != 0)
    {
      cerr << "Couldn't initialize a mutex.\n";
      abort();
    }

  if (pthread_cond_init(&reader_cond, NULL) != 0)
    {
      cerr << "Couldn't initialize a condition.\n";
      abort();
    }

  if (pthread_cond_init(&writer_cond, NULL) != 0)
    {
      cerr << "Couldn't initialize a condition.\n";
      abort();
    }

  number_of_readers = 0;
  number_of_writers = 0;
#endif
}


RWLock::~RWLock()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "RWLock(@" << this << ") destructed.\n";
#endif
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_destroy(&rwlock) != 0)
    {
      cerr << "Couldn't destroy a rwlock.\n";
      abort();
    }
#else
  if (number_of_readers != 0 || number_of_writers != 0)
    {
      cerr << "RWLock is busy.\n";
      abort();
    }

  if (pthread_cond_destroy(&writer_cond) != 0)
    {
      cerr << "Couldn't destroy a condition.\n";
      //abort();
    }

  if (pthread_cond_destroy(&reader_cond) != 0)
    {
      cerr << "Couldn't destroy a condition.\n";
      //abort();
    }

  if (pthread_mutex_destroy(&mutex) != 0)
    {
      cerr << "Couldn't destroy a mutex.\n";
      //abort();
    }
#endif
}


void
RWLock::release_rwlock()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "RWLock(@" << this << ") released.\n"; 
#endif
#ifndef USE_PTHREAD_RWLOCK
  assert(number_of_readers == 0);
  assert(number_of_writers == 0);
#endif
  if (ref_count == 1)
    {
      if (lock_manager_destructed)
	delete this;
      else
	lock_manager.rwlock_free(this);
      return;
    }

  assert(ref_count == 0);

  delete this;
  return;
}


int
RWLock::rdlock()
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_rdlock(&rwlock) != 0)
    {
      cerr << "Couldn't rdlock a rwlock.\n";
      abort();
    }
  return 2;				// Ȥƻʤʡ
#else
  mutex_lock();

  while (number_of_writers > 0)
    reader_cond_wait();

  ++number_of_readers;
  int lock_id = 2;

  // readeråʤΤ¾readerwriterϵ̵̡
  reader_cond_broadcast();
  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::tryrdlock()
{
#ifdef USE_PTHREAD_RWLOCK
  int result = pthread_rwlock_tryrdlock(&rwlock);

  if (result == 0)
    return 2;			// Ȥƻʤʡ

  if (result == EBUSY)
    return 0;			// åƤwriter롣

  cerr << "Couldn't tryrdlock a rwlock.\n";
  abort();
#else
  mutex_lock();

  int lock_id;
  if (number_of_writers == 0)
    {
      ++number_of_readers;
      lock_id = 2;
      // readeråʤΤ¾readerwriterϵ̵̡
      reader_cond_broadcast();
    }
  else
    lock_id = 0;

  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::wrlock()
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_wrlock(&rwlock) != 0)
    {
      cerr << "Couldn't wrlock a rwlock.\n";
      abort();
    }
  return 3;				// Ȥƻʤʡ
#else
  mutex_lock();

  ++number_of_writers;			// waitϿ

  while (number_of_readers > 0)
    writer_cond_wait();			// ƤreaderΥåԤ

  int lock_id = 3;			// writerlock_idϴȤ롣

  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::trywrlock()
{
#ifdef USE_PTHREAD_RWLOCK
  int result = pthread_rwlock_trywrlock(&rwlock);

  if (result == 0)
    return 3;			// Ȥƻʤʡ

  if (result == EBUSY)
    return 0;			// åƤwriter롣

  cerr << "Couldn't tryrdlock a rwlock.\n";
  abort();
#else
  mutex_lock();

  int lock_id;
  if (number_of_readers == 0)
    {
      ++number_of_writers;
      lock_id = 3;			// writerlock_idϴȤ롣
    }
  else
    lock_id = 0;

  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::unlock(int lock_id)
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_unlock(&rwlock) != 0)
    {
      cerr << "Couldn't unlock an rwlock.\n";
      abort();
    }
  return lock_id;
#else
  if (lock_id & 1)
    return release_writer_lock(lock_id);
  return release_reader_lock(lock_id);
#endif
}


#ifndef USE_PTHREAD_RWLOCK
int
RWLock::release_reader_lock(int lock_id)
{
  assert((lock_id & 1) == 0);
  mutex_lock();

  if (--number_of_readers < 0)
    {
      cerr << "Something wrong.  It seems not to be locked.\n";
      abort();	// 
    }

  if (number_of_readers == 0)	// ʬǸreaderäΤǡ
    writer_cond_broadcast();	// writer򵯤ͤꡣ

  mutex_unlock();

  return lock_id;
}


int
RWLock::release_writer_lock(int lock_id)
{
  assert(lock_id & 1);
  mutex_lock();

  if (--number_of_writers < 0)
    {
      cerr << "Something wrong.  It seems not to be locked.\n";
      abort();	// 
    }

  if (number_of_writers == 0)	// ʬǸwriteräΤǡ
    reader_cond_broadcast();	// Եreader򵯤ͤꡣ

  mutex_unlock();

  return lock_id;
}
#endif


}
