/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: lock.h,v 1.16 2007/01/01 03:17:25 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_LOCK_H_
#define _LIBOCHUSHA_LOCK_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "libochusha/ref_count.h"

#include <list>


#ifdef HAVE_PTHREAD_H
# include <iostream>

# include <cassert>
# include <cstdlib>

# include <pthread.h>

# ifdef HAVE_PTHREAD_RWLOCK_INIT
#  define USE_PTHREAD_RWLOCK	1
# endif

#endif

#define CHECK_LOCK_ORDER	1


using namespace std;
namespace Ochusha
{
#ifdef HAVE_PTHREAD_H
  // Recursive Lock
  class Lock
  {
    friend class Monitor;
  private:
    pthread_mutex_t mutex;
    pthread_t locker;
    int count;
#ifdef CHECK_LOCK_ORDER
    Lock *slave;	// ǥХå
#endif


  public:
    Lock(Lock *slave = NULL);
    virtual ~Lock();
    virtual Lock *clone();

    void lock();
    bool trylock();
    void unlock();

    bool is_locked_by_self();

    inline int get_count() const { return count; }
#ifdef CHECK_LOCK_ORDER
    static Lock *get_slave_lock();
    static Lock *get_global_lock();
#else
    static inline Lock *get_slave_lock()
    {
      static Lock slave_lock;
      return &slave_lock;
    }

    static inline Lock *get_global_lock()
    {
      static Lock global_lock;
      return &global_lock;
    }
#endif
  };


  class LockBlock
  {
  private:
    Lock *lock;


  public:
    inline LockBlock(Lock *lock) : lock(lock) { lock->lock(); }
    inline LockBlock(Lock &lock) : lock(&lock) { this->lock->lock(); }
    inline ~LockBlock() { lock->unlock(); }
  };


  class Monitor : public Lock
  {
  private:
    pthread_cond_t cond;


  public:
    Monitor();
    virtual ~Monitor();
    virtual Lock *clone();

    inline void enter() { lock(); }	// ñʤ̾
    inline void exit() { unlock(); }	// ñʤ̾
    bool wait();
    bool wait(long ms);
    void notify();
    void notify_all();
  };


  class RWLock
  {
  private:
    friend class LockManager;
    int ref_count;
#ifdef USE_PTHREAD_RWLOCK
    pthread_rwlock_t rwlock;
#else
    pthread_mutex_t mutex;
    pthread_cond_t reader_cond;	// reader lockwriterβԤ
    pthread_cond_t writer_cond;	// writer lockreaderβԤ

    int number_of_readers;
    int number_of_writers;
#endif
    RWLock();

    void release_rwlock();


  public:
    ~RWLock();
    inline void ref();
    inline void unref();

    int rdlock();
    int tryrdlock();

    int wrlock();
    int trywrlock();

    int unlock(int lock_id);

#ifndef USE_PTHREAD_RWLOCK
  private:
    inline void mutex_lock()
    {
      if (pthread_mutex_lock(&mutex) != 0)
	{
	  cerr << "Couldn't lock a mutex.\n";
	  abort();
	}
    }

    inline void mutex_unlock()
    {
      if (pthread_mutex_unlock(&mutex) != 0)
	{
	  cerr << "Couldn't unlock a mutex.\n";
	  abort();
	}
    }

    inline void reader_cond_wait()
    {
      if (pthread_cond_wait(&reader_cond, &mutex) != 0)
	{
	  cerr << "Couldn't wait a reader condition.\n";
	  abort();
	}
    }

    inline void reader_cond_signal()
    {
      if (pthread_cond_signal(&reader_cond) != 0)
	{
	  cerr << "Couldn't signal a reader condition.\n";
	  abort();
	}
    }

    inline void reader_cond_broadcast()
    {
      if (pthread_cond_broadcast(&reader_cond) != 0)
	{
	  cerr << "Couldn't broadcast a reader condition.\n";
	  abort();
	}
    }

    inline void writer_cond_wait()
    {
      if (pthread_cond_wait(&writer_cond, &mutex) != 0)
	{
	  cerr << "Couldn't wait a writer condition.\n";
	  abort();
	}
    }

    inline void writer_cond_signal()
    {
      if (pthread_cond_signal(&writer_cond) != 0)
	{
	  cerr << "Couldn't signal a writer condition.\n";
	  abort();
	}
    }

    inline void writer_cond_broadcast()
    {
      if (pthread_cond_broadcast(&writer_cond) != 0)
	{
	  cerr << "Couldn't broadcast a writer condition.\n";
	  abort();
	}
    }

    int release_reader_lock(int lock_id);
    int release_writer_lock(int lock_id);
#endif
  };


  class LockManager
  {
  private:
    friend class RWLock;

    Lock lock;

    list<RWLock *> free_rwlock_list;

    inline RWLock *get_rwlock()
    {
      LockBlock lock_obj(lock);
      if (free_rwlock_list.empty())
	return new RWLock();
      RWLock *rwlock = free_rwlock_list.front();
      free_rwlock_list.pop_front();
      return rwlock;
    }

    inline void rwlock_free(RWLock *rwlock)
    {
      LockBlock lock_obj(lock);
      assert(rwlock->ref_count == 1);
      free_rwlock_list.push_front(rwlock);
    }


  public:
    ~LockManager();

    static RWLock *rwlock_new();
  };


  inline void RWLock::ref() { ++ref_count; }
  inline void RWLock::unref()
  { if (--ref_count > 1) return; release_rwlock(); }


#else
  class Lock;
  class Monitor;
  class RWLock;
#endif

}
#endif

#endif	/* _LIBOCHUSHA_LOCK_H_ */
