/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: network_agent.cc,v 1.46 2007/01/04 06:30:19 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/network_agent.h"

#include "libochusha/http_headers.h"
#include "libochusha/misc_utils.h"
#include "libochusha/uri_utils.h"

#include <cstdlib>
#include <cstring>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <poll.h>


namespace Ochusha
{


#define LOG_MESSAGE_LENGTH		1024
#undef VERBOSE_DESTRUCTOR


NetworkAgentBroker::NetworkAgentBroker(PollEventManager &manager,
				       ThreadPool *pool)
  : manager(manager), thread_pool(pool)
{
}


NetworkAgent *
NetworkAgentBroker::employ_agent()
{
  return new NetworkAgent(*this);
}


class EstablishConnectionJob : public WorkerJob
{
private:
  smart_ptr<NetworkAgent> agent;


public:
  EstablishConnectionJob(NetworkAgent *agent) : agent(agent) {}

  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "establish network connection"; }
};


void
EstablishConnectionJob::work(const WorkerThread *employee)
{
  assert(agent->handle != NULL);

  if (employee->get_state() == WorkerThread::DIE_NOW)
    {
      smart_ptr<NetworkAgent> tmp = agent;
      agent->access_terminated(agent, agent->user_data);
      // ¿ʬץཪλʤΤǡϽФʤǤ꽪λ
      agent->broker.connection_pool.release_http_handle(agent->url,
							agent->handle,
							ACCESS_SUCCEEDED);
      agent->handle = NULL;
      agent->user_data = NULL;
      return;
    }

  agent->start_access();
}


class ProcessResponseBodyJob : public WorkerJob
{
private:
  smart_ptr<NetworkAgent> agent;

public:
  ProcessResponseBodyJob(NetworkAgent *agent) : agent(agent) {}
  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "process response body"; }
};


void
ProcessResponseBodyJob::work(const WorkerThread *employee)
{
  if (employee->get_state() == WorkerThread::DIE_NOW)
    {
      smart_ptr<NetworkAgent> tmp = agent;
      agent->access_terminated(agent, agent->user_data);
      // ¿ʬץཪλʤΤǡϽФʤǤ꽪λ
      agent->broker.connection_pool.release_http_handle(agent->url,
							agent->handle,
							ACCESS_SUCCEEDED);
      agent->handle = NULL;
      agent->user_data = NULL;
      return;
    }

  agent->process_response_body_can_block();
}


NetworkAgent::NetworkAgent(NetworkAgentBroker &broker)
  : broker(broker),
    mode(NOT_SPECIFIED),
    state(STATE_INITIAL),
    url(NULL),
    post_body(NULL),
    handle(NULL),
    buffer(NULL),
    user_data(NULL),
    cache_comparison_buffer_length(0),
    cache_check_required(false)
{
}


NetworkAgent::~NetworkAgent()
{
  if (url != NULL)
    free(url);

  assert(handle == NULL);
  assert(!(user_data != NULL));
#ifdef VERBOSE_DESTRUCTOR
  cerr << "NetworkAgent destructed.\n";
#endif
}


void
NetworkAgent::notify_allowed(HTTPHandle *handle)
{
  assert(this->handle == NULL);
  this->handle = handle;
  broker.thread_pool->commit(new EstablishConnectionJob(this));
}


bool
NetworkAgent::notify_busy()
{
  assert(this->handle == NULL);
  return true;
}


void
NetworkAgent::notify_denied()
{
  assert(this->handle == NULL);
  // 絬
  smart_ptr<NetworkAgent> tmp = this;
  const char *errstr = LIB_("Voluntary Restricted for Dead Server");
  access_failed(this, errstr, user_data);

  char message[LOG_MESSAGE_LENGTH];
  snprintf(message, LOG_MESSAGE_LENGTH,
	   LIB_("Access Failed(%s) due to: %s\n"), url, errstr);
  broker.output_log(message);
  user_data = NULL;
  return;
}


const char *
NetworkAgent::get_url()
{
  return url;
}


void
NetworkAgent::establish_connection()
{
  assert(handle == NULL);
  broker.connection_pool.request_connection(this);
}


void
NetworkAgent::refresh_dirty_cache()
{
  char message[LOG_MESSAGE_LENGTH];

  cache_check_required = false;
  cache_comparison_buffer_length = 0;
  buffer->clear0();
  mode = GET_REQUEST;
  state = STATE_INITIAL;
  cache_is_dirty(this, user_data);

  snprintf(message, LOG_MESSAGE_LENGTH, LIB_("Cache is Dirty: %s\n"), url);
  broker.output_log(message);

  broker.connection_pool.release_http_handle(url, handle, ACCESS_SUCCEEDED);
  handle = NULL;
  establish_connection();
}


bool
NetworkAgent::process_response_body_real(int lock_id)
{
  assert(state == STATE_BODY_RECEIVING || state == STATE_BODY_SKIPPING);
  assert(lock_id != 0);

  if (handle->get_status_code() == 416)
    {
      // Requested range not satisfiableʤΤǥåԡ
      buffer->get_lock()->unlock(lock_id);
      refresh_dirty_cache();
      return false;
    }

  short event_mask;
  do
    {
      HTTPStatus http_result = handle->process();

      if (http_result == HTTP_ERROR)
	{
	  buffer->get_lock()->unlock(lock_id);
	  notify_error();
	  // thisdestructƤǽꡣ
	  return false;
	}

      if (state == STATE_BODY_RECEIVING)
	{
	  size_t body_len = handle->get_body_length();
	  if (body_len > 0)
	    {
	      const char *body = handle->get_body();
	      assert(body != NULL);

	      if (cache_check_required)
		{
		  size_t cmp_buf_left
		    = CACHE_COMPARISON_SIZE - cache_comparison_buffer_length;
		  assert(cmp_buf_left > 0);
		  size_t len = MIN(body_len, cmp_buf_left);
		  memcpy((cache_comparison_buffer
			  + cache_comparison_buffer_length), body, len);
		  body_len -= len;
		  body += len;
		  cache_comparison_buffer_length += len;

		  if (cache_comparison_buffer_length == CACHE_COMPARISON_SIZE)
		    {
		      cache_check_required = false;
		      if (memcmp(cache_comparison_buffer,
				 (buffer->get_buffer0() + buffer->get_length0()
				  - CACHE_COMPARISON_SIZE),
				 CACHE_COMPARISON_SIZE) != 0)
			{
			  // å夬äȤˤĤơ
			  buffer->get_lock()->unlock(lock_id);
			  refresh_dirty_cache();
			  return false;
			}
		    }
		  else
		    {
		      // ޤåԤĴ٤ۤɥǡʤ
		      assert(body_len == 0);
		      assert(cache_comparison_buffer_length
			     < CACHE_COMPARISON_SIZE);
		    }
		}

	      if (body_len > 0)
		{
		  if (buffer->append_data(body, body_len, lock_id))
		    handle->flush_response_body_buffer();
		  else
		    {
		      // ¿ʬOut of MemoryΤ褦̿Ūʥ顼
		      buffer->get_lock()->unlock(lock_id);
		      smart_ptr<NetworkAgent> tmp = this;
		      access_failed(this, LIB_("Out of Memory"), user_data);
		      broker.notify_access_status(url, ACCESS_FINISHED);
		      broker.connection_pool.release_http_handle(url, handle,
								 ACCESS_BUSY);
		      handle = NULL;
		      user_data = NULL;
		      return false;
		    }
		}
	    }
	}	// if (state == BODY_RECEIVING)Υ֥å

      if (http_result == HTTP_DONE)
	{
	  if (state == STATE_BODY_RECEIVING)
	    state = STATE_BODY_RECEIVED;
	  else
	    {
	      assert(state == STATE_BODY_SKIPPING);
	      state = STATE_BODY_SKIPPED;
	    }
	  event_mask = 0;
	  break;
	}

      assert(http_result == HTTP_NOT_DONE);
      event_mask = static_cast<short>(handle->get_poll_event_mask());
    }
  while (event_mask == 0);

  buffer->get_lock()->unlock(lock_id);

  size_t len = buffer->get_length0();
  access_progressed(this, len, bytes_total, user_data);

  if (state == STATE_BODY_RECEIVING || state == STATE_BODY_SKIPPING)
    {
      assert(event_mask != 0);
      return true;
    }

  if (!cache_check_required)
    {
      if (state == STATE_BODY_RECEIVED)
	{
	  buffer->fix();
	  smart_ptr<NetworkAgent> tmp = this;
	  response_body_received(this, user_data);
	  notify_finished();
	}
      else
	{
	  assert(state == STATE_BODY_SKIPPED);
	  if (handle->get_status_code() == 304)
	    {
	      buffer->fix();
	      smart_ptr<NetworkAgent> tmp = this;
	      response_body_received(this, user_data);
	      notify_finished();
	    }
	  else
	    notify_failure();
	}
      return false;
    }

  // ɤ߹߽äΤ˥ǡ­ʤΤǡåԡ
  refresh_dirty_cache();
  return false;
}


void
NetworkAgent::process_response_body_can_block()
{
  int lock_id = buffer->get_lock()->wrlock();
  if (!process_response_body_real(lock_id))
    {
      /*
       * Τϡ
       * (1) Ƥresponse body
       * (2) åԤȯ줿ɤľϥꥯȺѡ
       * (3) ʥ顼ȯ
       * ʾ壳ĤξΤ줫Ǥꡢʾ夹뤳ȤϤʤ
       */
      return;
    }

  assert(state == STATE_BODY_RECEIVING || state == STATE_BODY_SKIPPING);

  broker.notify_access_status(url, WAIT_FOR_SERVER);

  short event_mask = static_cast<short>(handle->get_poll_event_mask());
  assert(event_mask != 0);

  broker.manager.request_poll(this, handle->get_socket(),
			      event_mask,
			      PollEventManager::PRIORITY_NORMAL,
			      PollEventManager::TIMEOUT_IO);
}


int
NetworkAgent::process_response_body_nonblock()
{
  int lock_id = buffer->get_lock()->trywrlock();
  if (lock_id == 0)
    {
      // contentionäϡbodyɤɤǽ
      // åɤư
      broker.thread_pool->commit(new ProcessResponseBodyJob(this));
      return EVENT_RESULT_CONTENTION;
    }

  if (!process_response_body_real(lock_id))
    {
      /*
       * Τϡ
       * (1) Ƥresponse body
       * (2) åԤȯ줿ɤľϥꥯȺѡ
       * (3) ʥ顼ȯ
       * ʾ壳ĤξΤ줫Ǥꡢʾ夹뤳ȤϤʤ
       */
      return EVENT_RESULT_DONE;
    }

  assert(state == STATE_BODY_RECEIVING || state == STATE_BODY_SKIPPING);

  broker.notify_access_status(url, WAIT_FOR_SERVER);

  short event_mask = static_cast<short>(handle->get_poll_event_mask());
  assert(event_mask != 0);

  int interval = handle->get_prefered_polling_interval_millisec();
  return EVENT_RESULT_CONTINUE(event_mask, interval);
}


int
NetworkAgent::update_state()
{
  int previous_state = state;

  if (state == STATE_INITIAL)
    {
      if (handle->is_connected())
	{
	  state = STATE_CONNECTED;
	  connection_established(this, user_data);
	}
    }

  if (state == STATE_CONNECTED)
    {
      if (handle->has_request_done())
	{
	  state = STATE_REQUEST_DONE;
	  request_sended(this, user_data);
	}
    }

  if (state == STATE_REQUEST_DONE)
    {
      if (handle->is_status_available())
	{
	  state = STATE_STATUS_RECEIVED;
	  status_received(this, user_data);
	}
    }

  if (state == STATE_STATUS_RECEIVED)
    {
      if (handle->is_headers_available())
	{
	  state = STATE_HEADERS_RECEIVED;

	  if (handle->get_content_length() != 0)
	    bytes_total = handle->get_content_length();

	  response_header_received(this, user_data);
	}
    }

  if (state == STATE_HEADERS_RECEIVED)
    {
      int status_code = handle->get_status_code();

      switch (status_code)
	{
	case 200:
	case 206:
	  state = STATE_BODY_RECEIVING;
	  break;

	case 304:
	  cache_check_required = false;
	  /* FALLTHROUGH */

	default:
	  state = STATE_BODY_SKIPPING;
	  break;
	}

      char message[LOG_MESSAGE_LENGTH];
      const char *reason = handle->get_reason_phrase();
      if (reason != NULL)
	snprintf(message, LOG_MESSAGE_LENGTH,
		 "%s: %d %s\n", url, status_code, reason);
      else
	snprintf(message, LOG_MESSAGE_LENGTH,
		 "%s: %d\n", url, status_code);
      broker.output_log(message);
    }

  return previous_state;
}


void
NetworkAgent::notify_finished()
{
  char message[LOG_MESSAGE_LENGTH];
  snprintf(message, LOG_MESSAGE_LENGTH, LIB_("Access Finished(%s): %d\n"),
	   url, handle->get_status_code());

  smart_ptr<NetworkAgent> tmp = this;
  access_finished(this, user_data);

  broker.output_log(message);
  broker.notify_access_status(url, ACCESS_FINISHED);
  broker.connection_pool.release_http_handle(url, handle, ACCESS_SUCCEEDED);
  handle = NULL;
  user_data = NULL;
}


// ŪHTTPꥯȤ̤ʤäȤΡ(404ʤ)
void
NetworkAgent::notify_failure()
{
  char message[LOG_MESSAGE_LENGTH];
  if (handle != NULL)
    snprintf(message, LOG_MESSAGE_LENGTH,
	     LIB_("Access Failed(%s): due to %d\n"),
	     url, handle->get_status_code());
  else
    snprintf(message, LOG_MESSAGE_LENGTH,
	     LIB_("Access Failed(%s): due to unknown reason, may be terminated at signal handler."),
	     url);

  // access_failedȯԤȡϥɥthis(ʤNetworkAgent)
  // ΤƤƤޤǽ⤤
  smart_ptr<NetworkAgent> tmp = this;
  access_failed(this, message, user_data);

  broker.output_log(message);
  broker.notify_access_status(url, ACCESS_FINISHED);
  broker.connection_pool.release_http_handle(url, handle, ACCESS_SUCCEEDED);
  handle = NULL;
  user_data = NULL;
}


// HTTPⲼؤǤμԤΡ(̿顼)
void
NetworkAgent::notify_error()
{
  char message[LOG_MESSAGE_LENGTH];
  const char *errstr = handle->get_error_string();
  if (handle != NULL)
    errstr = handle->get_error_string();
  else
    errstr = LIB_("Terminated at signal handler?");

  if (errstr == NULL)
    errstr = LIB_("unknown error");

  snprintf(message, LOG_MESSAGE_LENGTH,
	   LIB_("Access Failed(%s) due to: %s\n"), url, errstr);

  // access_failedȯԤȡϥɥthis(ʤNetworkAgent)
  // ΤƤƤޤǽ⤤
  smart_ptr<NetworkAgent> tmp = this;
  access_failed(this, errstr, user_data);
  broker.output_log(message);

  broker.notify_access_status(url, ACCESS_FINISHED);
  broker.connection_pool.release_http_handle(url, handle, ACCESS_FAILED);
  handle = NULL;
  user_data = NULL;
}


int
NetworkAgent::handle_event(int fd, short events, short revents)
{
  // PollEventManagerevent loopåɤƤФ롣
  // ХåեΥåԤǥ֥åƤϤʤʤ
  char message[LOG_MESSAGE_LENGTH];
  //HTTPStatus result;

  if ((events & revents) == 0)
    {
      const char *errstr;
      if (state == STATE_INITIAL)
	errstr = LIB_("Connection Timeout");
      else
	errstr = LIB_("I/O Timeout");

      smart_ptr<NetworkAgent> tmp = this;
      access_failed(this, errstr, user_data);

      snprintf(message, LOG_MESSAGE_LENGTH,
	       LIB_("Access Failed(%s) due to: %s\n"), url, errstr);
      broker.output_log(message);

      broker.notify_access_status(url, ACCESS_FINISHED);
      broker.connection_pool.release_http_handle(url, handle, ACCESS_BUSY);
      handle = NULL;
      user_data = NULL;
      return EVENT_RESULT_DONE;
    }

  if ((revents & (POLLIN | POLLPRI)) != 0)
    broker.notify_access_status(url, TRANSFER_DATA);

  HTTPStatus http_result;
  short event_mask;
  do
    {
      if (state == STATE_BODY_RECEIVING || state == STATE_BODY_SKIPPING)
	return process_response_body_nonblock();

      http_result = handle->process();

      if (http_result == HTTP_ERROR)
	{
	  notify_error();
	  // thisdestructƤǽꡣ
	  return EVENT_RESULT_DONE;
	}

#ifndef NDEBUG
      int previous_state = update_state();
#else
      update_state();
#endif

      if (state == STATE_BODY_RECEIVING)
	{
	  assert(previous_state != STATE_BODY_RECEIVING);
	  return process_response_body_nonblock();
	}

      if (state == STATE_BODY_SKIPPING)
	{
	  assert(previous_state != STATE_BODY_SKIPPING);
	  return process_response_body_nonblock();
	}

      if (state == STATE_TERMINATED)
	return EVENT_RESULT_DONE;

      if (http_result == HTTP_DONE)
	{
	  // HEADꥯȤξʳϤ뤳ȤϤʤ
	  // ơΤȤNetworkAgentHEADꥯȤȤäƤʤġġ
	  assert(handle != NULL);

	  notify_finished();

	  return EVENT_RESULT_DONE;	// ǤΤ
	}

      assert(http_result == HTTP_NOT_DONE);
      event_mask = static_cast<short>(handle->get_poll_event_mask());
    }
  while (event_mask == 0);

  assert(http_result == HTTP_NOT_DONE);
  assert(event_mask != 0);

  int interval = handle->get_prefered_polling_interval_millisec();
  switch (state)
    {
    case STATE_CONNECTED:
      broker.notify_access_status(url, SEND_REQUEST);
      return EVENT_RESULT_CONTINUE(event_mask, interval);

    case STATE_REQUEST_DONE:
    case STATE_STATUS_RECEIVED:
    case STATE_HEADERS_RECEIVED:
      broker.notify_access_status(url, WAIT_FOR_SERVER);
      return EVENT_RESULT_CONTINUE(event_mask, interval);

    case STATE_INITIAL:
    case STATE_BODY_SKIPPING:
    case STATE_BODY_SKIPPED:
    case STATE_BODY_RECEIVING:
    case STATE_BODY_RECEIVED:
    case STATE_TERMINATED:
      // should never reached
      break;
    }

  // should never reached
  assert(state != STATE_INITIAL);
  assert(state != STATE_BODY_SKIPPING);
  assert(state != STATE_BODY_SKIPPED);
  assert(state != STATE_BODY_RECEIVING);
  assert(state != STATE_BODY_RECEIVED);
  assert(state != STATE_TERMINATED);
  abort();
}


void
NetworkAgent::terminate(int fd)
{
  // 
  char message[LOG_MESSAGE_LENGTH];

  state = STATE_TERMINATED;

  smart_ptr<NetworkAgent> tmp = this;
  access_terminated(this, user_data);

  snprintf(message, LOG_MESSAGE_LENGTH,
	   LIB_("Access Terminated(%s)\n"), url);
  broker.output_log(message);

  broker.notify_access_status(url, ACCESS_FINISHED);
  broker.connection_pool.release_http_handle(url, handle, ACCESS_TERMINATED);
  handle = NULL;
  user_data = NULL;
}


// connectionΩ뤿WorkerThreadǥꥯ˰٤ƤФ롣
void
NetworkAgent::start_access()
{
  char message[LOG_MESSAGE_LENGTH];

  // EstablishConnectionJob꿶줿WorkerThreadǼ¹Ԥ졢
  // δؿreturnWorkerThreadϽλ롣
  handle->set_uri(url);
  if (mode == GET_REQUEST || mode == UPDATE_REQUEST)
    {
      snprintf(message, LOG_MESSAGE_LENGTH,
	       LIB_("HTTP Request: GET %s\n"), url);
      handle->set_method(HTTP_REQUEST_METHOD_GET);
    }
  else if (mode == POST_REQUEST)
    {
      snprintf(message, LOG_MESSAGE_LENGTH,
	       LIB_("HTTP Request: POST %s\n"), url);
      handle->set_method(HTTP_REQUEST_METHOD_POST);
      handle->set_body(post_body, strlen(post_body));
    }
  else
    {
      // EstablishConnectionJobreferenceäƤΤǤס
      // smart_ptr<NetworkAgent> tmp = this;
      access_terminated(this, user_data);
      broker.connection_pool.release_http_handle(url, handle, ACCESS_BUSY);
      handle = NULL;
      user_data = NULL;
      return;
    }

  broker.pre_prepare_request(this);
  prepare_request(this, user_data);
  broker.post_prepare_request(this);

  if (mode == UPDATE_REQUEST)
    {
      if (handle->get_request_header(HTTPHeader::If_Modified_Since) != NULL)
	{
	  char range_value[64];	// size_t64bitsǤ28;͵ġġ
	  snprintf(range_value, 64, "bytes=%lu-",
		   static_cast<unsigned long>(buffer->get_length0())
		   - CACHE_COMPARISON_SIZE);
	  handle->set_request_header(HTTPHeader::Range, range_value);
	}
      else
	{
	  // ɤIf-Modified-SinceʤRangeGETꥯȤФ
	  // ɤʤȤ餷
	  mode = GET_REQUEST;
	  cache_check_required = false;
	  cache_comparison_buffer_length = 0;
	  buffer->clear(0);
	  cache_is_dirty(this, user_data);
	}
    }

  handle->set_request_header(HTTPHeader::Connection, "close");
  handle->set_asynchronous(true);
  handle->set_nonblock(true);
  handle->prepare();

  access_started(this, user_data);
  broker.output_log(message);

  URI uri(url);
  assert(uri.is_valid());
  assert(uri.get_server() != NULL);

  broker.notify_access_status(url, LOOK_UP_SERVER);

  assert(state == STATE_INITIAL);

  HTTPStatus http_result;
  short event_mask;
  do
    {
      if (state == STATE_BODY_RECEIVING || state == STATE_BODY_SKIPPING)
	{
	  process_response_body_can_block();
	  // ʤpollingɬפʾrequest_poll()
	  // process_response_body_can_block()˷פ餦
	  return;
	}

      http_result = handle->process();

      if (http_result == HTTP_ERROR)
	{
	  notify_error();
	  // thisdestructƤǽꡣ
	  return;
	}

#ifndef NDEBUG
      int previous_state = update_state();
#else
      update_state();
#endif

      /*
       * O_NONBLOCKʥʤΤǡ̾
       * previous_state == state == STATE_INITIAL
       * http_result == HTTP_NOT_DONE
       * event_mask == POLLOUT
       */

      if (state == STATE_BODY_RECEIVING)
	{
	  assert(previous_state != STATE_BODY_RECEIVING);
	  process_response_body_can_block();
	  // ʤpollingɬפʾrequest_poll()
	  // process_response_body_can_block()˷פ餦
	  return;
	}

      if (state == STATE_BODY_SKIPPING)
	{
	  assert(previous_state != STATE_BODY_SKIPPING);
	  process_response_body_can_block();
	  // ʤpollingɬפʾrequest_poll()
	  // process_response_body_can_block()˷פ餦
	  return;
	}

      if (state == STATE_TERMINATED)
	return;

      if (http_result == HTTP_DONE)
	{
	  // HEADꥯȤξʳϤ뤳ȤϤʤ
	  // ơΤȤNetworkAgentHEADꥯȤȤäƤʤġġ
	  assert(handle != NULL);

	  notify_finished();

	  return;
	}

      assert(http_result == HTTP_NOT_DONE);
      event_mask = static_cast<short>(handle->get_poll_event_mask());
    }
  while (event_mask == 0);

  assert(http_result == HTTP_NOT_DONE);
  assert(event_mask != 0);

  int fd = handle->get_socket();
  assert(fd > 0);

  switch (state)
    {
    case STATE_INITIAL:
      broker.notify_access_status(url, CONNECT_TO_SERVER);
      broker.manager.request_poll(this, fd, event_mask,
				  PollEventManager::PRIORITY_NORMAL,
				  PollEventManager::TIMEOUT_CONNECTION);
      return;

      /*****************************************************************
       * O_NONBLOCKդǥƤΤǰʲϸŪǤʤġġ*
       *****************************************************************/
    case STATE_CONNECTED:
      broker.notify_access_status(url, SEND_REQUEST);
      broker.manager.request_poll(this, fd, event_mask,
				  PollEventManager::PRIORITY_NORMAL,
				  PollEventManager::TIMEOUT_IO);
      return;

    case STATE_REQUEST_DONE:
    case STATE_STATUS_RECEIVED:
    case STATE_HEADERS_RECEIVED:
      broker.notify_access_status(url, WAIT_FOR_SERVER);
      broker.manager.request_poll(this, fd, event_mask,
				  PollEventManager::PRIORITY_NORMAL,
				  PollEventManager::TIMEOUT_IO);
      return;

    case STATE_BODY_SKIPPING:
    case STATE_BODY_SKIPPED:
    case STATE_BODY_RECEIVING:
    case STATE_BODY_RECEIVED:
    case STATE_TERMINATED:
      // should never reached
      break;
    }

  // should never reached
  assert(state != STATE_BODY_SKIPPING);
  assert(state != STATE_BODY_SKIPPED);
  assert(state != STATE_BODY_RECEIVING);
  assert(state != STATE_BODY_RECEIVED);
  assert(state != STATE_TERMINATED);
  abort();
}


void
NetworkAgent::get(const char *url, Buffer *buf, RefCount *data)
{
  assert(buf != NULL);

  if (this->url != NULL)
    free(this->url);
  this->url = strdup(url);
  if (post_body != NULL)
    {
      free(post_body);
      post_body = NULL;
    }
  mode = GET_REQUEST;
  state = STATE_INITIAL;
  buffer = buf;
  cache_check_required = false;
  cache_comparison_buffer_length = 0;
  user_data = data;
  establish_connection();
}


void
NetworkAgent::update(const char *url, Buffer *buf, RefCount *data)
{
  assert(buf != NULL);

  if (this->url != NULL)
    free(this->url);
  this->url = strdup(url);
  if (post_body != NULL)
    {
      free(post_body);
      post_body = NULL;
    }

  if (buf->get_length0() < CACHE_COMPARISON_SIZE)
    {
      // åοĴ٤ۤɥǡʤΤñGET
      mode = GET_REQUEST;
      cache_check_required = false;
      buf->clear(0);
    }
  else
    {
      mode = UPDATE_REQUEST;
      cache_check_required = true;
    }
  cache_comparison_buffer_length = 0;
  state = STATE_INITIAL;
  buffer = buf;
  user_data = data;
  establish_connection();
}


void
NetworkAgent::post(const char *url, const char *body, Buffer *buf,
		   RefCount *data)
{
  assert(buf != NULL);

  if (this->url != NULL)
    free(this->url);
  this->url = strdup(url);
  if (post_body != NULL)
    free(post_body);
  post_body = strdup(body);
  buffer = buf;
  state = STATE_INITIAL;
  mode = POST_REQUEST;
  cache_check_required = false;
  cache_comparison_buffer_length = 0;
  user_data = data;
  establish_connection();
}


bool
NetworkAgent::terminate_access()
{
  if (handle != NULL)
    return handle->close();
  return false;
}


class SaveBufferJob : public WorkerJob
{
private:
  smart_ptr<NetworkAgent> agent;
  int fd;

public:
  SaveBufferJob(NetworkAgent *agent, int fd) : agent(agent), fd(fd) {}
  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "saving buffer to file"; }
};


void
SaveBufferJob::work(const WorkerThread *employee)
{
  agent->get_buffer()->write_file(fd);
}


void
NetworkAgent::save_buffer(int fd)
{
  broker.thread_pool->commit(new SaveBufferJob(this, fd));
}


void
NetworkAgent::set_proxy_url(const char *url)
{
  handle->set_proxy(url);
}


void
NetworkAgent::set_proxy_uri(const URI &uri)
{
  handle->set_proxy(uri);
}


void
NetworkAgent::set_proxy_authinfo(const char *user, const char *password)
{
  handle->set_proxy_authinfo(user, password);
}


void
NetworkAgent::set_header(const char *name, const char *value)
{
  handle->set_request_header(name, value);
}


void
NetworkAgent::set_non_interactive_mode(bool value)
{
  handle->set_policy(value
		     ? HTTP_POLICY_HIGH_THROUGHPUT : HTTP_POLICY_LOW_LATENCY);
}


const char *
NetworkAgent::get_header(const char *name) const
{
  if (handle == NULL)
    return NULL;

  return handle->get_response_header(name);
}


const HTTPHeaders *
NetworkAgent::get_headers() const
{
  if (handle == NULL)
    return NULL;

  return &handle->get_headers();
}


int
NetworkAgent::get_status_code() const
{
  if (handle == NULL)
    return 0;

  return handle->get_status_code();
}


const char *
NetworkAgent::get_reason_phrase() const
{
  const char *reason;
  if (handle != NULL)
    reason = handle->get_reason_phrase();
  else
    reason = NULL;

  return reason != NULL ? reason : "";
}

}

