/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ochusha_procedure.cc,v 1.17 2007/01/01 03:17:25 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/ochusha_procedure.h"

#include "libochusha/bulletin_board.h"
#include "libochusha/misc_utils.h"
#include "libochusha/uri_utils.h"

#include "libtscore/tscore.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <string>


namespace Ochusha
{

ThreadlistRuleInfo::ThreadlistRuleInfo(Cell *name_cell, Cell *label_cell,
				       Cell *desc_cell,
				       Cell *proc_cell, TSCore &core)
  : name(strndup(name_cell->get_string(), name_cell->get_strlen())),
    label(strndup(label_cell->get_string(), label_cell->get_strlen())),
    description(strndup(desc_cell->get_string(), desc_cell->get_strlen())),
    procedure(core, proc_cell)
{
}


ThreadlistRuleInfo::~ThreadlistRuleInfo()
{
  free(name);
  free(label);
  free(description);
}


void
ThreadlistRuleInfo::set_name(Cell *name_cell)
{
  free(name);
  name = strndup(name_cell->get_string(), name_cell->get_strlen());
}


void
ThreadlistRuleInfo::set_label(Cell *label_cell)
{
  free(label);
  label = strndup(label_cell->get_string(), label_cell->get_strlen());
}


void
ThreadlistRuleInfo::set_description(Cell *desc_cell)
{
  free(description);
  description = strndup(desc_cell->get_string(), desc_cell->get_strlen());
}


void
ThreadlistRuleInfo::set_procedure(Cell *proc_cell)
{
  procedure = proc_cell;
}


ExternalToolInfo::ExternalToolInfo(Cell *name_cell, Cell *label_cell,
				   Cell *desc_cell, Cell *tool_type_cell,
				   Cell *template_cell, TSCore &core)
  : name(strndup(name_cell->get_string(), name_cell->get_strlen())),
    label(strndup(label_cell->get_string(), label_cell->get_strlen())),
    description(strndup(desc_cell->get_string(), desc_cell->get_strlen())),
    tool_type(UNKNOWN_TOOL),
    command_line_template(core, template_cell)
{
  set_tool_type(tool_type_cell);
}


ExternalToolInfo::~ExternalToolInfo()
{
  free(name);
  free(label);
  free(description);
}


void
ExternalToolInfo::set_name(Cell *name_cell)
{
  free(name);
  name = strndup(name_cell->get_string(), name_cell->get_strlen());
}


void
ExternalToolInfo::set_label(Cell *label_cell)
{
  free(label);
  label = strndup(label_cell->get_string(), label_cell->get_strlen());
}


void
ExternalToolInfo::set_description(Cell *desc_cell)
{
  free(description);
  description = strndup(desc_cell->get_string(), desc_cell->get_strlen());
}


void
ExternalToolInfo::set_tool_type(Cell *tool_type_cell)
{
  const char *str = tool_type_cell->get_symname();
  if (strcmp(str, "link-tool") == 0)
    tool_type = LINK_TOOL;
  else if (strcmp(str, "image-tool") == 0)
    tool_type = IMAGE_TOOL;
  else if (strcmp(str, "text-tool") == 0)
    tool_type = TEXT_TOOL;
  else
    tool_type = UNKNOWN_TOOL;
}


void
ExternalToolInfo::set_template(Cell *template_cell)
{
  command_line_template = template_cell;
}


NativeProcedure::Info OchushaProcedure::procedures[] = {
  { &OchushaProcedure::url_encode_string,
    "url-encode-string", TST_STRING, 1, 1 },
  { &OchushaProcedure::url_decode_string,
    "url-decode-string", TST_STRING, 1, 1 },
  { &OchushaProcedure::board_get_bbs_type,
    "board-get-bbs-type", TST_NONE, 1, 1 },
  { &OchushaProcedure::board_get_url,
    "board-get-url", TST_NONE, 1, 1 },
  { &OchushaProcedure::board_get_name,
    "board-get-name", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_board,
    "thread-get-board", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_id,
    "thread-get-id", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_title,
    "thread-get-title", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_url,
    "thread-get-url", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_rank,
    "thread-get-rank", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_previous_rank,
    "thread-get-previous-rank", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_number_of_responses_on_server,
    "thread-get-number-of-responses-on-server", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_number_of_responses_read,
    "thread-get-number-of-responses-read", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_number_of_responses_got,
    "thread-get-number-of-responses-got", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_number_of_new_responses,
    "thread-get-number-of-new-responses", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_number_of_responses_available,
    "thread-get-number-of-responses-available", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_number_of_responses_to_read,
    "thread-get-number-of-responses-to-read", TST_NONE, 1, 1 },
  { &OchushaProcedure::thread_get_birthtime,
    "thread-get-birthtime", TST_NONE, 1, 1 },
  { &OchushaProcedure::is_thread_new,
    "thread-new?", TST_NONE, 1, 1 },
  { &OchushaProcedure::is_thread_alive,
    "thread-alive?", TST_NONE, 1, 1 },
  { &OchushaProcedure::is_thread_stopped,
    "thread-stopped?", TST_NONE, 1, 1 },
  { &OchushaProcedure::is_thread_marked,
    "thread-marked?", TST_NONE, 1, 1 },
  { &OchushaProcedure::is_thread_hidden,
    "thread-hidden?", TST_NONE, 1, 1 },
  { &OchushaProcedure::register_threadlist_rule,
    "register-threadlist-rule", TST_STRING TST_STRING TST_STRING TST_PROCEDURE,
    4, 4 },
  { &OchushaProcedure::register_external_tool,
    "register-external-tool",
    TST_STRING TST_STRING TST_STRING TST_SYMBOL TST_ANY,
    5, 5 }
};


OchushaProcedure::OchushaProcedure(TSCore &core) : NativeProcedure(core)
{
  core.define_global_integer("OCHUSHA-BBS-TYPE-UNKNOWN",
			     OCHUSHA_BBS_TYPE_UNKNOWN);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CH",
			     OCHUSHA_BBS_TYPE_2CH);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CHLIKE-EUCJP",
			     OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP);
  core.define_global_integer("OCHUSHA-BBS-TYPE-JBBS",
			     OCHUSHA_BBS_TYPE_JBBS);
  core.define_global_integer("OCHUSHA-BBS-TYPE-MACHIBBS",
			     OCHUSHA_BBS_TYPE_MACHIBBS);
  core.define_global_integer("OCHUSHA-BBS-TYPE-JBBS-LIVEDOOR",
			     OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR);
  core.define_global_integer("OCHUSHA-BBS-TYPE-MITINOKU",
			     OCHUSHA_BBS_TYPE_MITINOKU);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CH-COMPATIBLE",
			     OCHUSHA_BBS_TYPE_2CH_COMPATIBLE);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CH-HEADLINE",
			     OCHUSHA_BBS_TYPE_2CH_HEADLINE);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CH-BE",
			     OCHUSHA_BBS_TYPE_2CH_BE);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CH-RESERVED",
			     OCHUSHA_BBS_TYPE_2CH_RESERVED);
  core.define_global_integer("OCHUSHA-BBS-TYPE-BBSPINK",
			     OCHUSHA_BBS_TYPE_BBSPINK);
  core.define_global_integer("OCHUSHA-BBS-TYPE-2CH-COMAPTIBLE-RO",
			     OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO);
}


OchushaProcedure::~OchushaProcedure()
{
  while (!threadlist_rule_list.empty())
    {
      delete threadlist_rule_list.back();
      threadlist_rule_list.pop_back();
    }
}


void
OchushaProcedure::register_threadlist_rule(Cell *name, Cell *label,
					   Cell *desc, Cell *proc)
{
  vector<ThreadlistRuleInfo *>::iterator iter
    = threadlist_rule_list.begin();
  vector<ThreadlistRuleInfo *>::iterator end_iter
    = threadlist_rule_list.end();

  const char *name_start = name->get_string();
  size_t name_len = name->get_strlen();

  while (iter != end_iter)
    {
      ThreadlistRuleInfo *info = *iter++;
      if (strlen(info->get_name()) == name_len
	  && strncmp(info->get_name(), name_start, name_len) == 0)
	{
	  info->set_label(label);
	  info->set_description(desc);
	  info->set_procedure(proc);
	  return;
	}
    }

  threadlist_rule_list.push_back(new ThreadlistRuleInfo(name, label, desc,
							proc, core));
}


void
OchushaProcedure::register_external_tool(Cell *name, Cell *label,
					 Cell *desc, Cell *tool_type,
					 Cell *command_line_template)
{
  vector<ExternalToolInfo *>::iterator iter
    = external_tool_list.begin();
  vector<ExternalToolInfo *>::iterator end_iter
    = external_tool_list.end();

  const char *name_start = name->get_string();
  size_t name_len = name->get_strlen();

  while (iter != end_iter)
    {
      ExternalToolInfo *info = *iter;
      if (strlen(info->get_name()) == name_len
	  && strncmp(info->get_name(), name_start, name_len) == 0)
	{
	  info->set_label(label);
	  info->set_description(desc);
	  info->set_tool_type(tool_type);
	  info->set_template(command_line_template);
	  return;
	}
      ++iter;
    }

  external_tool_list.push_back(new ExternalToolInfo(name, label, desc,
						    tool_type,
						    command_line_template,
						    core));
}


const ThreadlistRuleInfo *
OchushaProcedure::lookup_threadlist_rule_info(const char *name) const
{
  if (name == NULL)
    return NULL;

  vector<ThreadlistRuleInfo *>::const_iterator iter
    = threadlist_rule_list.begin();
  vector<ThreadlistRuleInfo *>::const_iterator end_iter
    = threadlist_rule_list.end();

  while (iter != end_iter)
    {
      ThreadlistRuleInfo *info = *iter;
      if (strcmp(name, info->get_name()) == 0)
	return info;
      ++iter;
    }

  return NULL;
}


const ExternalToolInfo *
OchushaProcedure::lookup_external_tool_info(const char *name) const
{
  if (name == NULL)
    return NULL;

  vector<ExternalToolInfo *>::const_iterator iter
    = external_tool_list.begin();
  vector<ExternalToolInfo *>::const_iterator end_iter
    = external_tool_list.end();

  while (iter != end_iter)
    {
      ExternalToolInfo *info = *iter;
      if (strcmp(name, info->get_name()) == 0)
	return info;
      ++iter;
    }

  return NULL;
}


Cell *
OchushaProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


Cell *
OchushaProcedure::url_encode_string(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);

  string buf;
  buf.append(arg->get_string(), arg->get_strlen());

  char *result_str = URI::encode_string(buf.c_str());
  Cell *result = vm.mk_string(result_str);
  free(result_str);

  return result;
}


Cell *
OchushaProcedure::url_decode_string(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);

  string buf;
  buf.append(arg->get_string(), arg->get_strlen());

  char *result_str = URI::decode_string(buf.c_str());
  Cell *result = vm.mk_string(result_str);
  free(result_str);

  return result;
}


Cell *
OchushaProcedure::board_get_bbs_type(VirtualMachine &vm, Register &context,
				     unsigned long argument_pointer,
				     unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "board-get-bbs-type: argument must be of foreign object: ", arg);
  BulletinBoard *board = static_cast<BulletinBoard *>(arg->get_foreign_pointer());
  return vm.mk_integer(board->get_bbs_type());
}


Cell *
OchushaProcedure::board_get_url(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "board-get-url: argument must be of foreign object: ", arg);
  BulletinBoard *board = static_cast<BulletinBoard *>(arg->get_foreign_pointer());
  return vm.mk_static_string(board->get_base_url());
}


Cell *
OchushaProcedure::board_get_name(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "board-get-name: argument must be of foreign object: ", arg);
  BulletinBoard *board = static_cast<BulletinBoard *>(arg->get_foreign_pointer());
  return vm.mk_static_string(board->get_name());
}


Cell *
OchushaProcedure::thread_get_board(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-board: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_foreign_object(thread->get_board(), NULL);
}


Cell *
OchushaProcedure::thread_get_id(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-id: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_static_string(thread->get_id());
}


Cell *
OchushaProcedure::thread_get_title(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-title: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_static_string(thread->get_title());
}


Cell *
OchushaProcedure::thread_get_url(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-url: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_static_string(thread->get_url());
}


Cell *
OchushaProcedure::thread_get_rank(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-rank: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_rank());
}


Cell *
OchushaProcedure::thread_get_previous_rank(VirtualMachine &vm,
					   Register &context,
					   unsigned long argument_pointer,
					   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-previous-rank: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_previous_rank());
}


Cell *
OchushaProcedure::thread_get_number_of_responses_on_server(
					VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-number-of-responses-on-server: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_number_of_responses_on_server());
}


Cell *
OchushaProcedure::thread_get_number_of_responses_read(
					VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-number-of-responses-read: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_number_of_responses_read());
}


Cell *
OchushaProcedure::thread_get_number_of_responses_got(
					VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-number-of-responses-got: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_number_of_responses_got());
}


Cell *
OchushaProcedure::thread_get_number_of_new_responses(
					VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-number-of-new-responses: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_number_of_new_responses());
}


Cell *
OchushaProcedure::thread_get_number_of_responses_available(
					VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-number-of-responses-available: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_number_of_responses_available());
}


Cell *
OchushaProcedure::thread_get_number_of_responses_to_read(
					VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-number-of-responses-to-read: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(thread->get_number_of_responses_to_read());
}


Cell *
OchushaProcedure::thread_get_birthtime(VirtualMachine &vm, Register &context,
				       unsigned long argument_pointer,
				       unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-get-birthtime: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return vm.mk_integer(static_cast<long>(thread->get_birthtime()));
}


Cell *
OchushaProcedure::is_thread_new(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-new?: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return Cell::ts_bool(thread->is_alive() && thread->get_previous_rank() == 0);
}


Cell *
OchushaProcedure::is_thread_alive(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-alive?: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return Cell::ts_bool(thread->is_alive());
}


Cell *
OchushaProcedure::is_thread_stopped(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-stopped?: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return Cell::ts_bool(thread->is_stopped());
}


Cell *
OchushaProcedure::is_thread_marked(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-marked?: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return Cell::ts_bool(thread->is_marked());
}


Cell *
OchushaProcedure::is_thread_hidden(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "thread-hidden?: argument must be of foreign object: ", arg);
  BBSThread *thread = static_cast<BBSThread *>(arg->get_foreign_pointer());
  return Cell::ts_bool(thread->is_hidden());
}


Cell *
OchushaProcedure::register_threadlist_rule(VirtualMachine &vm,
					   Register &context,
					   unsigned long argument_pointer,
					   unsigned long nargs, void *data)
{
  Cell *name = context.frame->load_variable(argument_pointer);
  assert(name->is_string());

  Cell *label = context.frame->load_variable(argument_pointer + 1);
  assert(label->is_string());

  Cell *desc = context.frame->load_variable(argument_pointer + 2);
  assert(desc->is_string());

  Cell *proc = context.frame->load_variable(argument_pointer + 3);
  assert(proc->is_procedure());

  OchushaProcedure *procedure = static_cast<OchushaProcedure *>(data);
  procedure->register_threadlist_rule(name, label, desc, proc);

  return Cell::t();
}


Cell *
OchushaProcedure::register_external_tool(VirtualMachine &vm,
					 Register &context,
					 unsigned long argument_pointer,
					 unsigned long nargs, void *data)
{
  Cell *name = context.frame->load_variable(argument_pointer);
  assert(name->is_string());

  Cell *label = context.frame->load_variable(argument_pointer + 1);
  assert(label->is_string());

  Cell *desc = context.frame->load_variable(argument_pointer + 2);
  assert(desc->is_string());

  Cell *tool_type = context.frame->load_variable(argument_pointer + 3);
  assert(tool_type->is_symbol());

  Cell *templ = context.frame->load_variable(argument_pointer + 4);
  if (!templ->is_string() && !templ->is_procedure())
    return signal_error(vm, "register-external-tool: argument #4 must be of a string or procedure that takes a string and returns string: ", templ);

  OchushaProcedure *procedure = static_cast<OchushaProcedure *>(data);
  procedure->register_external_tool(name, label, desc, tool_type, templ);

  return Cell::t();
}

}
