/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sigslot.cc,v 1.12 2007/01/07 05:42:08 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/sigslot.h"


namespace Ochusha
{

_signal_base::_signal_base(Lock *lock)
{
  if (lock == NULL)
    this->lock = Lock::get_global_lock();
  else
    this->lock = lock;
}


_signal_base::~_signal_base()
{
  disconnect_all();
}


void
_signal_base::disconnect_all()
{
  LockBlock lock_obj(lock);
  while (!connected_slots.empty())
    {
      _connection_base *cb = connected_slots.front();
      cb->get_dest()->signal_disconnect(this);
      connected_slots.pop_front();
      delete cb;
    }
}


void
_signal_base::disconnect(has_slots *slot_holder)
{
  LockBlock lock_obj(lock);
  list<_connection_base *>::iterator it = connected_slots.begin();
  list<_connection_base *>::iterator itEnd = connected_slots.end();

  while (it != itEnd)
    {
      if ((*it)->get_dest() == slot_holder)
	{
	  delete *it;
	  connected_slots.erase(it);
	  slot_holder->signal_disconnect(this);
	  return;
	}
      ++it;
    }
}


void
_signal_base::slot_disconnect(has_slots *slot)
{
  LockBlock lock_obj(lock);
  list<_connection_base *>::iterator it = connected_slots.begin();
  list<_connection_base *>::iterator itEnd = connected_slots.end();

  while (it != itEnd)
    {
      list<_connection_base *>::iterator itNext = it;
      ++itNext;
      if ((*it)->get_dest() == slot)
	{
	  delete *it;
	  connected_slots.erase(it);
	}
      it = itNext;
    }
}


void
_signal_base::slot_duplicate(const has_slots *oldtarget, has_slots *newtarget)
{
  LockBlock lock_obj(lock);
  list<_connection_base *>::iterator it = connected_slots.begin();
  list<_connection_base *>::iterator itEnd = connected_slots.end();

  while (it != itEnd)
    {
      if ((*it)->get_dest() == oldtarget)
	{
	  connected_slots.push_back((*it)->duplicate(newtarget));
	}
      ++it;
    }
}


has_slots::has_slots(Lock *lock)
{
  if (lock == NULL)
    this->lock = Lock::get_global_lock();
  else
    this->lock = lock;
}


has_slots::has_slots(const has_slots &hs)
{
  if (hs.lock == Lock::get_global_lock())
    lock = Lock::get_global_lock();
  else
    lock = hs.lock->clone();
  LockBlock lock_obj(lock);
  set<_signal_base *>::const_iterator it = hs.senders.begin();
  set<_signal_base *>::const_iterator itEnd = hs.senders.end();

  while (it != itEnd)
    {
      (*it)->slot_duplicate(&hs, this);
      senders.insert(*it);
      ++it;
    }
}


has_slots::~has_slots()
{
  disconnect_all();
}


void
has_slots::signal_connect(_signal_base *sender)
{
  LockBlock lock_obj(lock);
  senders.insert(sender);
}


void
has_slots::signal_disconnect(_signal_base *sender)
{
  LockBlock lock_obj(lock);
  senders.erase(sender);
}


void
has_slots::disconnect_all()
{
  LockBlock lock_obj(lock);
  set<_signal_base *>::const_iterator it = senders.begin();
  set<_signal_base *>::const_iterator itEnd = senders.end();

  while (it != itEnd)
    {
      (*it)->slot_disconnect(this);
      ++it;
    }

  senders.erase(senders.begin(), senders.end());
}


void
Signal0::emit()
{
  LockBlock lock_obj(lock);

  list<_connection_base *>::const_iterator it = connected_slots.begin();
  list<_connection_base *>::const_iterator itEnd = connected_slots.end();
  while (it != itEnd)
    {
      list<_connection_base *>::const_iterator itNext = it;
      ++itNext;
      dynamic_cast<_connection_base0 *>(*it)->emit();
      it = itNext;
    }
}


}


#ifdef SIG_TEST
#include <iostream>

using namespace Ochusha;

class Switch
{
public:
  Signal0 clicked;
};


class Counter : public has_slots
{
private:
  int count;

public:
  Counter() { count = 0; }
  void increment()
  {
    cerr << "count=" << ++count << endl;
    incremented(this);
  }

  signal1<Counter *> incremented;
};


class CounterWatcher : public has_slots
{
public:
  void show_incremented(Counter *counter)
  {
    cerr << "counter(" << counter << ") incremented.\n";
  }
};


int
main(int argc, char *argv[])
{
  Switch sw;
  Counter counter;
  CounterWatcher watcher;
  sw.clicked.connect(&counter, &Counter::increment);
  sw.clicked();

  counter.incremented.connect(&watcher, &CounterWatcher::show_incremented);
  sw.clicked();
  return 0;
}
#endif
