/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_thread_view.cc,v 1.42 2007/01/06 10:45:02 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "bbs_thread_view.h"

#include "libochusha/bbs_thread.h"
#include "libochusha/constants.h"
#include "libochusha/misc_utils.h"
#include "libochusha/text_utils.h"

#include "controller.h"
#include "image_manager.h"
#include "tab_label.h"

#include "libochushagtk/misc_utils.h"

#include <string>

// #define ENABLE_BENCHMARKING	1

#ifdef ENABLE_BENCHMARKING
#include <sys/time.h>
#endif


namespace Ochusha
{


#define THREAD_VIEW_UPDATE_PRIORITY		GDK_PRIORITY_REDRAW - 5
#define MAX_NUM_OF_RES_TO_APPEND_AT_ONCE	256

#define THREAD_VIEW_EXTRA_WIDTH		30
#define POPUP_OFFSET			8

// #undef VERBOSE_DESTRUCTOR


BBSThreadViewSelectionData::BBSThreadViewSelectionData(BBSThread *thread)
  : thread(thread),
    link(NULL),
    canonical_link(NULL)
{
}


BBSThreadViewSelectionData::~BBSThreadViewSelectionData()
{
  if (link != NULL)
    free(link);

  if (canonical_link != NULL)
    free(canonical_link);
}


static const GtkTargetEntry link_targets[] = {
  { const_cast<char *>("text/uri-list"), 0, 0 }
};


static const GtkTargetEntry empty_targets[] = {
};


const GtkTargetEntry *
BBSThreadViewSelectionData::get_targets()
{
  if (canonical_link == NULL)
    canonical_link = BBSThreadRenderer::canon_link(thread, link);

  return canonical_link != NULL ? link_targets : empty_targets;
}


guint
BBSThreadViewSelectionData::get_n_targets()
{
  if (canonical_link == NULL)
    canonical_link = BBSThreadRenderer::canon_link(thread, link);

  return canonical_link != NULL ? G_N_ELEMENTS(link_targets) : 0;
}


void
BBSThreadViewSelectionData::get_data(GtkClipboard *clipboard,
				     GtkSelectionData *data, guint info)
{
  g_return_if_fail(info == 0);

  g_return_if_fail(canonical_link != NULL);

  char *buf[2] = { canonical_link, NULL };
  gtk_selection_data_set_uris(data, buf);
}


bool
BBSThreadViewSelectionData::clear_data(GtkClipboard *clipboard)
{
  return false;
}


void
BBSThreadViewSelectionData::set_link(const gchar *value)
{
  if (link != NULL)
    free(link);

  link = value != NULL ? strdup(value) : NULL;

  if (canonical_link != NULL)
    {
      free(canonical_link);
      canonical_link = NULL;
    }
}


const char *
BBSThreadViewSelectionData::get_canonical_link()
{
  if (canonical_link == NULL)
    canonical_link = BBSThreadRenderer::canon_link(thread, link);
  return canonical_link;
}


static GQuark view_id = g_quark_from_static_string("BBSThreadView::view");


BBSThreadViewBase::BBSThreadViewBase(Controller &controller,
				     BBSThread *thread)
  : controller(controller),
    thread(thread),
    buffer(ochusha_hypertext_buffer_new(NULL)),
    widget(ochusha_hypertext_view_new(buffer)),
    view(OCHUSHA_HYPERTEXT_VIEW(widget)),
    renderer(controller, thread, buffer, view),
    message_id(0),
    selection_done_id(0),
    popup_menu(false),
    selection_data(thread),
    popup_link(NULL),
    popup_view(NULL)
{
  g_object_set_qdata(G_OBJECT(widget), view_id, this);
  g_signal_connect(widget, "destroy",
		   reinterpret_cast<GCallback>(widget_destroy_cb), this);
  g_signal_connect(widget, "unmap",
		   reinterpret_cast<GCallback>(widget_unmap_cb), this);

  g_signal_connect(widget, "populate_popup",
		   reinterpret_cast<GCallback>(populate_popup_cb), this);
  g_signal_connect(widget, "link_mouse_over",
		   reinterpret_cast<GCallback>(link_mouse_over_cb), this);
  g_signal_connect(widget, "link_mouse_out",
		   reinterpret_cast<GCallback>(link_mouse_out_cb), this);
  g_signal_connect(widget, "link_mouse_press",
		   reinterpret_cast<GCallback>(link_mouse_press_cb), this);
  g_signal_connect(widget, "link_mouse_release",
		   reinterpret_cast<GCallback>(link_mouse_release_cb), this);
  g_signal_connect(widget, "copy_clipboard",
		   reinterpret_cast<GCallback>(copy_clipboard_cb), this);
}


BBSThreadViewBase::~BBSThreadViewBase()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BBSThreadViewBase destructed: %p\n", this);
#endif

  g_object_unref(buffer);

  if (message_id != 0)
    {
      GtkStatusbar *bar = controller.get_statusbar();
      gtk_statusbar_remove(bar, controller.get_link_context_id(), message_id);
    }

  g_return_if_fail(popup_link == NULL);
}


void
BBSThreadViewBase::widget_destroy_cb(GtkWidget *widget,
				     BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  // BBSThreadViewsigslotΥϥɥĤΤdestructorglobal lock
  // 롣å뤿ᡢĴ롣

  // ҥݥåץåפС˾ä
  widget_unmap_cb(widget, view);

  gdk_threads_leave();

  LockBlock lock_obj(Lock::get_global_lock());

  gdk_threads_enter();

  delete view;
}


// ڤؤǸʤʤޤ롣
void
BBSThreadViewBase::widget_unmap_cb(GtkWidget *widget,
				   BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (view->popup_view != NULL)
    {
      BBSThreadViewBase *popup_view = view->popup_view;
      popup_view->hide(true);
      assert(view->popup_view == NULL);
      assert(view->popup_link == NULL);
    }
}


void
BBSThreadViewBase::popup_destroy_cb(GtkWidget *widget, BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  assert(view->popup_link != NULL);
  g_free(view->popup_link);
  view->popup_link = NULL;

  assert(get_view(widget) == view->popup_view);
  view->popup_view = NULL;

  view->hide(false);
}


gboolean
BBSThreadViewBase::link_mouse_over_cb(OchushaHypertextView *widget,
				      GdkEventMotion *event, const gchar *link,
				      BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (view->message_id != 0)
    {
      GtkStatusbar *bar = view->controller.get_statusbar();
      gtk_statusbar_remove(bar, view->controller.get_link_context_id(),
			   view->message_id);
      view->message_id = 0;
    }

  if (strncmp(link, "<RES>", 5) == 0)
    return FALSE;

  if (view->popup_link != NULL && strcmp(view->popup_link, link) != 0)
    {
      if (view->popup_view != NULL)
	{
	  BBSThreadViewBase *popup_view = view->popup_view;
	  popup_view->hide(true);
	  assert(view->popup_view == NULL);
	  assert(view->popup_link == NULL);
	}
      else
	{
	  g_free(view->popup_link);
	  view->popup_link = NULL;
	}
    }

  if (view->get_response_cursor() != NULL)
    {
      // ݥåץåפ䥿ȥݥåץå׻ˤϲ⤹ɬפʤΤ
      // 
      if (view->popup_view == NULL)
	{ // ߥݥåץåɽǤʤ
	  char *popup_link = BBSThreadRenderer::canon_link(view->thread, link);
	  if (popup_link != NULL)
	    {
	      Controller &controller = view->controller;
	      BulletinBoard *board = view->thread->get_board();
	      int from;
	      int to;
	      BBSThread *thread
		= board->lookup_bbs_thread_by_url(popup_link, &from, &to);
	      if (thread == view->thread)
		{
		  if (from == 0)
		    {
		      from = 1;
		      to = 1;
		    }
		  else if (from < 0)
		    {
		      to = thread->get_number_of_responses_got();
		      from += to;
		    }
		  view->popup_view
		    = new BBSThreadPopupView(controller, view->thread,
					     view->get_window(),
					     view->get_response_cursor(),
					     from, to);
		}
	      else
		{
		  BBSTable &table = controller.get_bbs_table();
		  board = table.guess_board(popup_link);
		  if (board != NULL)
		    {
		      board->read_threadlist_xml(controller.get_repository());
		      char title_buf[256];
		      thread = board->lookup_bbs_thread_by_url(popup_link);
		      if (thread != NULL)
			snprintf(title_buf, 256, _("%s@%s board"),
				 thread->get_title(), board->get_name());
		      else if (table.lookup_board_by_url(popup_link) == NULL)
			snprintf(title_buf, 256, _("Unknown thread@%s board"),
				 board->get_name());
		      else
			snprintf(title_buf, 256, _("%s board"),
				 board->get_name());
		      view->popup_view
			= new BBSThreadPopupView(controller, view->thread,
						 view->get_window(),
						 title_buf);
		    }
		  else if (controller.get_prefs().get_enable_popup_image()
			   && controller.get_image_manager().is_image_link(popup_link))
		    {
		      int width
			= controller.get_prefs().get_popup_image_width();
		      int height
			= controller.get_prefs().get_popup_image_height();
		      GtkWidget *image
			= controller.get_image_manager().get_widget(popup_link,
								    width,
								    height);
		      if (image != NULL)
			{
			  view->popup_view
			    = new BBSThreadPopupView(controller, view->thread,
						     view->get_window(),
						     image, popup_link);
			}
		    }
		}
	    }
	  else
	    {
	      if (strncmp(link, "<ID>", 4) == 0
		  && view->controller.get_prefs().get_enable_popup_id())
		{
		  view->popup_view
		    = new BBSThreadPopupView(view->controller, view->thread,
					     view->get_window(),
					     view->get_response_cursor(),
					     link + 4);
		  popup_link = strdup(link);
		}
	    }

	  if (view->popup_view != NULL)
	    {
	      view->popup_link = popup_link;
	      BBSThreadViewBase *popup_view = view->popup_view;
	      popup_view->show();
	      g_signal_connect(popup_view->widget, "destroy",
			       reinterpret_cast<GCallback>(popup_destroy_cb),
			       view);
	    }
	  else if (popup_link != NULL)
	    free(popup_link);
	}
      else
	{ // ɽΥݥåץåפƱ̤٤󥯤ä
	  BBSThreadViewBase *popup_view = view->popup_view;
	  popup_view->show();
	}
    }


  // ʲlink򥹥ơСɽ
  char tmp_buffer[PATH_MAX];
  snprintf(tmp_buffer, PATH_MAX, "%s", link);
  char *tmp_link = strstr(tmp_buffer, "<FAKE>");

  if (tmp_link != NULL)
    *tmp_link = '\0';
  else if (strncmp(tmp_buffer, "<ID>", 4) == 0)
    snprintf(tmp_buffer, PATH_MAX, "%s", link + 4);

  tmp_link = strchr(tmp_buffer, '\n');
  if (tmp_link != NULL)
    *tmp_link = '\0';

  GtkStatusbar *bar = view->controller.get_statusbar();
  view->message_id
    = gtk_statusbar_push(bar, view->controller.get_link_context_id(),
			 tmp_buffer);

  return FALSE;
}


gboolean
BBSThreadViewBase::link_mouse_out_cb(OchushaHypertextView *widget,
				     GdkEventMotion *event,
				     BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (view->message_id != 0)
    {
      GtkStatusbar *bar = view->controller.get_statusbar();
      gtk_statusbar_remove(bar, view->controller.get_link_context_id(),
			   view->message_id);
      view->message_id = 0;
    }

  if (view->popup_view != NULL)
    {
      BBSThreadViewBase *popup_view = view->popup_view;
      popup_view->hide(false);
    }

  return FALSE;
}


void
BBSThreadViewBase::populate_popup_cb(GtkTextView *text_view,
				     GtkMenu *popup_menu,
				     BBSThreadViewBase *view)
{
  view->controller.get_scheme_manager()
    .attach_external_text_tools_menu(text_view, popup_menu);

  view->popup_menu = true;
  assert(view->selection_done_id == 0);
  view->selection_done_id
    = g_signal_connect(popup_menu, "selection-done",
		       reinterpret_cast<GCallback>(selection_done_cb), view);
}


void
BBSThreadViewBase::selection_done_cb(GtkMenu *popup_menu,
				     BBSThreadViewBase *view)
{
  view->popup_menu = false;
  g_return_if_fail(view->selection_done_id != 0);
  g_signal_handler_disconnect(popup_menu, view->selection_done_id);
  view->selection_done_id = 0;
}


void
BBSThreadViewBase::popup_menu_destroy_cb(BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

  g_return_if_fail(view->get_last_event_link() != NULL);

  view->selection_data.set_link(NULL);
  view->popup_menu = false;
}


gboolean
BBSThreadViewBase::link_mouse_press_cb(OchushaHypertextView *widget,
				       GdkEventButton *event, const gchar *link,
				       BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

  view->selection_data.set_link(NULL);

  g_return_val_if_fail(link != NULL, FALSE);

  if (event->button == 3)
    {
      const gchar *menu_root;
      if (strncmp(link, "<RES>", 5) == 0)
	{
	  menu_root = "/ThreadResponseNumberPopupMenu";
	  bool a_boned;
	  guint res_num;
	  if (sscanf(link + 5, "%u", &res_num) == 1)
	    a_boned = view->thread->is_a_boned(res_num);
	  else
	    a_boned = false;
	  view->controller.get_thread_ui().prepare_response_number_popup_menu(a_boned);
	}
      else if (strncmp(link, "<ID>", 4) == 0)
	menu_root = "/ThreadIDPopupMenu";
      else if (strncmp(link, "mailto:", 7) == 0)
	{
	  menu_root = NULL;
	  return FALSE;
	}
      else if (view->controller.get_image_manager().is_image_link(link))
	{
	  menu_root = "/ThreadImagePopupMenu";
	  view->controller.get_thread_ui().prepare_image_popup_menu(link);
	}
      else if (view->controller.get_thread_ui().is_link_in_thread(view->thread,
								  link)
	       || view->controller.is_bbs_link(link))
	menu_root = "/ThreadBBSLinkPopupMenu";
      else
	menu_root = "/ThreadLinkPopupMenu";

      view->selection_data.set_link(link);
      view->controller.update_internal_clipboard(new SelectionDataProxy(&view->selection_data, GTK_WIDGET(widget)));

      GtkUIManager *manager = view->controller.get_ui_manager();
      view->popup_menu = true;
      ochusha_ui_manager_popup_with_data(manager, menu_root,
					 view,
					 (GDestroyNotify)popup_menu_destroy_cb,
					 NULL, NULL,
					 event->button,
					 event->time);
      return TRUE;
    }

  if (strncmp(link, "<RES>", 5) == 0
      || strncmp(link, "<ID>", 4) == 0
      || strncmp(link, "mailto:", 7) == 0)
    return FALSE;

  view->selection_data.set_link(link);

  return FALSE;
}


gboolean
BBSThreadViewBase::link_mouse_release_cb(OchushaHypertextView *widget,
					 GdkEventButton *event,
					 const gchar *link,
					 BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  const char *last_event_link = view->get_last_event_link();
  if (last_event_link == NULL)
    return FALSE;

  if (strcmp(last_event_link, link) == 0 && event->button != 3)
    {
      Preferences &prefs = view->controller.get_prefs();
      bool in_tab = (event->button == 2) ^ prefs.get_default_open_in_tab();
      char *real_link = BBSThreadRenderer::canon_link(view->thread, link);
      if (real_link != NULL)
	{
	  view->controller.open(real_link, in_tab, false, event->button == 1);
	  free(real_link);
	}
      // ξϥݥåץåפ¨¤˾õ٤ġġ
    }

  if (event->button != 3)
    view->selection_data.set_link(NULL);

  return FALSE;
}


void
BBSThreadViewBase::copy_clipboard_cb(OchushaHypertextView *widget,
				     BBSThreadViewBase *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  gchar *text = ochusha_hypertext_view_get_selected_text(widget);
  if (text != NULL)
    {
      view->controller.copy(text);
      g_free(text);
    }
}


BBSThreadViewBase *
BBSThreadViewBase::get_view(GtkWidget *widget)
{
  return (BBSThreadViewBase *)g_object_get_qdata(G_OBJECT(widget), view_id);
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThread *thread,
				       GtkWindow *parent,
				       ResponseCursor *cursor,
				       guint from, guint to)
  : BBSThreadViewBase(controller, thread),
    response_cursor(cursor),
    window(NULL),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // 쥹fromtoޤǤΥ쥹ɽݥåץåפ(Ǥڤꤢ)
  initialize_common(parent);
  int lock_id = cursor->trylock();
  if (lock_id != 0)
    {
      renderer.prepare_for_popup();
      if (from > to)
	to = from;

      int num_res
	= controller.get_prefs().get_maximum_number_of_popup_responses();

      while (from <= to && num_res > 0 && cursor->has_response(from, lock_id))
	{
	  Response &response = cursor->get_response(from, lock_id);
	  if (!response.is_deleted())
	    {
	      if (!response.is_broken())
		renderer.render_popup_response(from, response);
	      else
		renderer.render_popup_broken(from);
	      --num_res;
	    }
	  ++from;
	}

      cursor->unlock(lock_id);
    }
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThread *thread,
				       GtkWindow *parent,
				       ResponseCursor *cursor,
				       const gchar *id)
  : BBSThreadViewBase(controller, thread),
    response_cursor(cursor),
    window(NULL),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // IDidʥ쥹֤򽸤ɽݥåץåפ롣
  initialize_common(parent);
  g_return_if_fail(id != NULL);

  int lock_id = cursor->trylock();
  if (lock_id != 0)
    {
      renderer.prepare_for_popup_id(thread->get_title(), id);
      int count = 0;
      for (guint res_num = 1; cursor->has_response(res_num, lock_id); ++res_num)
	{
	  Response &response = cursor->get_response(res_num, lock_id);
	  if (!response.is_deleted())
	    {
	      const char *res_id = response.get_id();
	      if (res_id != NULL && strstr(res_id, id) != NULL)
		renderer.render_popup_id(res_num, response, ++count);
	    }
	}

      cursor->unlock(lock_id);
    }
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThread *thread,
				       GtkWindow *parent,
				       GtkWidget *image, const gchar *link)
  : BBSThreadViewBase(controller, thread),
    response_cursor(NULL),
    window(NULL),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // linkimageɽݥåץåפ롣
  initialize_common(parent);
  renderer.render_image(image, link);
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThread *thread,
				       GtkWindow *parent,
				       const gchar *title)
  : BBSThreadViewBase(controller, thread),
    response_cursor(NULL),
    window(NULL),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // 쥿̾ɽݥåץåפ롣
  initialize_common(parent);
  renderer.render_popup_title(title);
}


BBSThreadPopupView::~BBSThreadPopupView()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BBSThreadPopupView(@%p) destructed\n", this);
#endif
  assert(Lock::get_slave_lock()->is_locked_by_self());

  if (popup_id != 0)
    g_source_remove(popup_id);

  if (popdown_id != 0)
    g_source_remove(popdown_id);
}


void
BBSThreadPopupView::initialize_common(GtkWindow *parent)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  window = gtk_window_new(GTK_WINDOW_POPUP);
  gtk_window_set_resizable(GTK_WINDOW(window), FALSE);
  gtk_window_set_transient_for(GTK_WINDOW(window), parent);

  GtkWidget *frame = gtk_frame_new(NULL);
  gtk_widget_show(frame);

  gtk_container_add(GTK_CONTAINER(window), frame);

  gtk_widget_show(widget);
  gtk_container_add(GTK_CONTAINER(frame), widget);

  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), GTK_WRAP_NONE);

  g_signal_connect(widget, "enter_notify_event",
		   reinterpret_cast<GCallback>(enter_notify_event_cb), this);
  g_signal_connect(widget, "leave_notify_event",
		   reinterpret_cast<GCallback>(leave_notify_event_cb), this);
  g_signal_connect(widget, "size_request",
		   reinterpret_cast<GCallback>(size_request_cb), this);
}


gboolean
BBSThreadPopupView::enter_notify_event_cb(GtkWidget *widget,
					  GdkEventCrossing *event,
					  BBSThreadPopupView *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  assert(view->popup_id == 0);

  view->pointed = true;

  if (view->popdown_id != 0)
    {
      g_source_remove(view->popdown_id);
      view->popdown_id = 0;
    }

  return FALSE;
}


gboolean
BBSThreadPopupView::leave_notify_event_cb(GtkWidget *widget,
					  GdkEventCrossing *event,
					  BBSThreadPopupView *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  assert(view->popup_id == 0);

  view->pointed = false;

  if (view->popdown_id == 0)
    view->hide(false);

  return FALSE;
}


void
BBSThreadPopupView::size_request_cb(GtkWidget *widget,
				    GtkRequisition *requisition,
				    BBSThreadPopupView *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (requisition->width <= 0 || requisition->height <= 0)
    return;

  gint width;
  if (view->response_cursor != NULL)
    width = requisition->width + THREAD_VIEW_EXTRA_WIDTH;
  else
    width = requisition->width;
  gint height = requisition->height;

  gtk_widget_set_size_request(widget, width, height);
  gtk_window_resize(GTK_WINDOW(view->window), width, height);

  GdkScreen *screen = gtk_widget_get_screen(view->window);
  GdkDisplay *display = gdk_screen_get_display(screen);
  gint x_pos;
  gint y_pos;
  gdk_display_get_pointer(display, NULL, &x_pos, &y_pos, NULL);
  x_pos += POPUP_OFFSET;
  y_pos -= (height + POPUP_OFFSET);
  y_pos = y_pos > 0 ? y_pos : 0;

  gint screen_width = gdk_screen_get_width(screen);
  if (screen_width > 0 && (x_pos + width) > screen_width)
    {
      x_pos = screen_width - width;
      if (x_pos < 0)
	x_pos = 0;
    }

  gtk_window_move(GTK_WINDOW(view->window), x_pos, y_pos);
}


ResponseCursor *
BBSThreadPopupView::get_response_cursor()
{
  return response_cursor;
}


GtkWindow *
BBSThreadPopupView::get_window()
{
  return GTK_WINDOW(window);
}


gboolean
BBSThreadPopupView::show_now(BBSThreadPopupView *view)
{
  gdk_threads_enter();

  view->popup_id = 0;
  gtk_widget_show(view->window);

  gdk_threads_leave();

  return FALSE;
}


gboolean
BBSThreadPopupView::hide_now(BBSThreadPopupView *view)
{
  gdk_threads_enter();

  view->popdown_id = 0;
  if (!view->has_popup() && !view->has_popup_menu())
    {
      // ҥݥåץåפʤ

      gtk_widget_destroy(view->window);
      // λviewdeleteƤ롣
    }
  // ҥݥåץåפ̤ΤäԤġ

  gdk_threads_leave();

  return FALSE;
}


void
BBSThreadPopupView::show()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (popdown_id != 0)
    {
      g_source_remove(popdown_id);
      popdown_id = 0;
      assert(popup_id == 0);
    }
  else if (popup_id != 0)
    return;

  // gtk_widget_map()Ƥʤȡǽsize_requestʥ
  // OchushaHypertextView(ʤGtkTextView)Υ{ 0, 0 }
  // ˤʤäƤޤθäƤallocationΤľˤʤ롣
  // η̡ݥåץåפФƸ줷
  gtk_widget_map(widget);

  guint popup_delay = controller.get_prefs().get_popup_delay();
  if (popup_delay > 0)
    popup_id = g_timeout_add(popup_delay,
			     reinterpret_cast<GSourceFunc>(show_now), this);
  else
    popup_id = g_idle_add(reinterpret_cast<GSourceFunc>(show_now), this);
}


void
BBSThreadPopupView::hide(bool now)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (popup_id != 0 || now)
    {
      gtk_widget_destroy(window);
      return;
    }

  if (popdown_id != 0)
    return;

  if (pointed)
    return;

  guint popdown_delay = controller.get_prefs().get_popdown_delay();
  if (popdown_delay > 0)
    popdown_id = g_timeout_add(popdown_delay,
			       reinterpret_cast<GSourceFunc>(hide_now), this);
  else
    popdown_id = g_idle_add(reinterpret_cast<GSourceFunc>(hide_now), this);
}


BBSThreadView::BBSThreadView(Controller &controller, BBSThread *thread)
  : BBSThreadViewBase(controller, thread),
    response_cursor(NULL),
    number_of_responses_shown(0),
    append_responses_id(0),
    initial_view_position(0),
    my_own_bookmark(NULL),
    my_own_bookmark_position(0),
    post_window(NULL),
    post_window_destroy_id(0),
    tab_label(NULL),
    auto_refresher_id(0),
    previous_interval(0),
    previous_number_of_responses_got(0),
    enable_transparent_a_bone(controller.get_prefs().get_enable_transparent_a_bone()),
    disable_a_bone(controller.get_prefs().get_disable_a_bone()),
    enable_chain_a_bone(controller.get_prefs().get_enable_chain_a_bone())
{
  g_signal_connect(widget, "scroll_view",
		   reinterpret_cast<GCallback>(scroll_view_cb), this);
}


BBSThreadView::~BBSThreadView()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BBSThreadView destructed: %p\n", this);
#endif

  if (auto_refresher_id != 0)
    g_source_remove(auto_refresher_id);

  if (append_responses_id != 0)
    g_source_remove(append_responses_id);

  // 񤭹ߥɥƱĤȤ⤢뤬̯
  if (post_window_destroy_id != 0)
    g_signal_handler_disconnect(post_window, post_window_destroy_id);

  thread->set_number_of_responses_read(thread->get_number_of_responses_got());
  GtkWidget *board_view
    = controller.get_view_manager().get_current_board_view();
  if (board_view != NULL)
    gtk_widget_queue_draw(board_view);
}


ResponseCursor *
BBSThreadView::get_response_cursor()
{
  return response_cursor;
}


GtkWindow *
BBSThreadView::get_window()
{
  return controller.get_toplevel();
}


// 鲶ѤϢ

// OchushaHypertextBuffer/Viewoffset֤widget
// records򹹿
void
BBSThreadView::insert_widget_at_offset(GtkWidget *widget, int offset)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

  GSList *tags = buffer->tags;
  buffer->tags = NULL;
  GtkTextTag *bookmark_tag = controller.get_tag_manager().get_my_bookmark_tag();
  ochusha_hypertext_buffer_push_tag(buffer, bookmark_tag);
  gint b_off = ochusha_hypertext_view_insert_widget_at_offset(view, widget,
							      0, 20, offset);
  ochusha_hypertext_buffer_pop_tags(buffer, bookmark_tag);
  buffer->tags = tags;

  g_return_if_fail(b_off >= 0);

  g_assert(offset == b_off);
  // recordsϥ쥹ֹбƤΤǡoffsetˤĤƾ硣
  // binary searchƴطΤ֤ƬõƤ
  // offsetĴ򤹤ȤŬ;ϤϤ뤬
  // 쥹NˤĤO(N)ʤΤݤʤȤϤʤǤ
  vector<ResponseRecord>::iterator iter = records.begin();
  vector<ResponseRecord>::iterator end_iter = records.end();
  while (iter != end_iter)
    {
      if ((*iter).offset >= offset)
	(*iter).offset++;	// widgetΤä֤offset1餹
      ++iter;
    }
}


// OchushaHypertextBuffer/Viewwidgetrecords
// 
void
BBSThreadView::remove_widget(GtkWidget *widget)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  gint offset = ochusha_hypertext_view_remove_widget(view, widget);
  if (offset < 0)
    fprintf(stderr, "ochusha_hypertext_view_remove_widget() returns %d\n",
	    offset);
  g_return_if_fail(offset >= 0);

  // recordsϥ쥹ֹбƤΤǡoffsetˤĤƾ硣
  // binary searchƴطΤ֤ƬõƤ
  // offsetĴ򤹤ȤŬ;ϤϤ뤬
  // 쥹NˤĤO(N)ʤΤݤʤȤϤʤǤ
  vector<ResponseRecord>::iterator iter = records.begin();
  vector<ResponseRecord>::iterator end_iter = records.end();
  while (iter != end_iter)
    {
      if ((*iter).offset >= offset)
	(*iter).offset--;	// widgetΤä֤offset1餹
      iter++;
    }
}


// 쥹ֹres_numΥ쥹ľ˲ѤΤ
void
BBSThreadView::insert_my_own_bookmark(guint res_num)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (my_own_bookmark != NULL)
    {
      remove_widget(my_own_bookmark);
      my_own_bookmark = NULL;
    }

  if (records.size() < res_num - 1)
    return;

  GtkWidget *label = controller.get_view_manager().get_tab_label(thread);
  if (label != NULL)
    {
      TabLabel *tab_label = TabLabel::get_label(label);
      if (tab_label != NULL && !tab_label->is_error_state())
	tab_label->notice(thread->get_number_of_responses_got()
			  > static_cast<int>(res_num));
    }

  GtkWidget *bookmark = create_bookmark();
  if (bookmark == NULL)
    return;	// ɽ̵֤ԲΥ顼ۤä롣

  ResponseRecord &res = records[res_num - 1];
  ochusha_hypertext_buffer_create_mark_at_offset(buffer, "my-own-bookmark",
						 res.offset);
  insert_widget_at_offset(bookmark, res.offset + res.length);
  my_own_bookmark = bookmark;
  my_own_bookmark_position = res_num;

  controller.get_thread_ui().notify_thread_view_state(this, thread);
}


GtkWidget *
BBSThreadView::create_bookmark()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (controller.get_prefs().get_hide_bookmark())
    return NULL;	// ɽ

  GtkWidget *button
    = gtk_button_new_with_label(controller.get_prefs().get_bookmark_text());
  GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
  gtk_button_set_focus_on_click(GTK_BUTTON(button), FALSE);
  gtk_widget_set_name(button, "mybookmark");

  g_signal_connect(button, "clicked",
		   reinterpret_cast<GCallback>(bookmark_clicked_cb), this);
  return button;
}


void
BBSThreadView::bookmark_clicked_cb(GtkWidget *widget, BBSThreadView *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (view->my_own_bookmark_position == view->records.size())
    return;	// ˤȤϲ⤷ʤ

  view->thread->set_bookmark_position(0);
  view->thread->set_number_of_responses_read(view->records.size());
  view->my_own_bookmark_position = view->records.size();
  view->insert_my_own_bookmark(view->records.size());
  GtkWidget *board_view
    = view->controller.get_view_manager().get_current_board_view();
  if (board_view != NULL)
    gtk_widget_queue_draw(board_view);
}


// ޤǲѤϢ


void
BBSThreadView::scroll_view_cb(GtkWidget *widget, GtkScrollType scroll,
			      BBSThreadView *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  g_signal_emit_by_name(widget->parent, "scroll_child", scroll, FALSE);
}


static bool
check_a_bone(Response &response, ABone &a_bone)
{
  return (a_bone.get_a_bone_by_name() && a_bone.get_a_bone_by_name_regexp().match(response.get_name(), -1))
    || (a_bone.get_a_bone_by_id() && a_bone.get_a_bone_by_id_regexp().match(response.get_id(), -1))
    || (a_bone.get_a_bone_by_content() && a_bone.get_a_bone_by_content_regexp().match(response.get_message(), -1));
}


void
BBSThreadView::append_response(guint res_num, Response &response)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (res_num == 1)
    {
      // 쥹1λϥȥɽ
      const char *title;
      if (!response.is_broken())
	{
	  title = response.get_title();
	  if (thread->get_title() == NULL)
	    thread->set_title(title);
	}
      else
	title = thread->get_title();
      if (title != NULL && *title != '\0')
	renderer.render_header(title);
    }

  assert(records.size() == res_num - 1);

  int offset = ochusha_hypertext_buffer_get_current_offset(buffer);
  if (response.is_deleted())
    {
      records.push_back(ResponseRecord(offset, 0));
      return;
    }

  if (!response.is_broken())
    {
      bool a_bone = !disable_a_bone
	&& (thread->is_a_boned(res_num)
	    || check_a_bone(response, thread->get_a_bone())
	    || check_a_bone(response, thread->get_board()->get_a_bone())
	    || check_a_bone(response, controller.get_config().get_a_bone()));

      if (!a_bone)
	renderer.render_response(res_num, response);
      else
	{
	  if (!enable_transparent_a_bone)
	    renderer.render_a_bone(res_num);
	}
    }
  else
    renderer.render_broken(res_num);

  int tail_offset = ochusha_hypertext_buffer_get_current_offset(buffer);
  records.push_back(ResponseRecord(offset, tail_offset - offset));
}


gboolean
BBSThreadView::append_responses_on_idle(BBSThreadView *view)
{
  // idleϥɥʤΤgdk_threads_leave()
  LockBlock lock_obj(Lock::get_global_lock());

  gdk_threads_enter();

  smart_ptr<ResponseCursor> cursor = view->response_cursor;
  gboolean result;

  int lock_id = cursor->trylock();
  if (lock_id != 0)
    {
#ifdef ENABLE_BENCHMARKING
      struct timeval start_time;
      gettimeofday(&start_time, NULL);
#endif
      // MAX_NUM_OF_RES_TO_APPEND_AT_ONCE쥹⤷ϺǸޤɲ
      int count = 0;
      guint res_num = view->number_of_responses_shown;
      while (cursor->has_response(++res_num, lock_id))
	{
	  Response &response = cursor->get_response(res_num, lock_id);
	  view->append_response(res_num, response);
	  view->number_of_responses_shown = res_num;

	  if (res_num == view->my_own_bookmark_position)
	    view->insert_my_own_bookmark(res_num);

	  if (res_num == view->initial_view_position)
	    {
	      GtkTextMark *mark
		= ochusha_hypertext_buffer_create_mark_at_offset(
					view->buffer, "view-position",
					view->records[res_num - 1].offset);
	      ochusha_hypertext_view_scroll_to_mark(view->view, mark);
	    }

	  if (++count >= MAX_NUM_OF_RES_TO_APPEND_AT_ONCE)
	    break;
	}

      if (view->auto_refresher_id != 0)
	{
	  GtkTextMark *mark
	    = ochusha_hypertext_buffer_create_mark(view->buffer,
						   "tail-view-position");
	  ochusha_hypertext_view_scroll_to_mark(view->view, mark);
	}

      if (!cursor->has_response(res_num, lock_id))
	{ // ΤȤʾɤ߼쥹ʤ
	  view->append_responses_id = 0;
	  result = FALSE;
	  if (cursor->finished(lock_id))
	    {
	      BBSThread *thread = view->thread;
	      GtkWidget *label
		= view->controller.get_view_manager().get_tab_label(thread);
	      if (label != NULL)
		{
		  TabLabel *tab_label = TabLabel::get_label(label);
		  if (tab_label != NULL && !tab_label->is_error_state())
		    tab_label->notice(thread->get_number_of_responses_to_read()
				      > 0);
		}
	    }
	}
      else
	result = TRUE;

      cursor->unlock(lock_id);
#ifdef ENABLE_BENCHMARKING
      struct timeval end_time;
      gettimeofday(&end_time, NULL);
      time_t sec = end_time.tv_sec - start_time.tv_sec;
      suseconds_t usec = end_time.tv_usec - start_time.tv_usec;
      while (usec < 0)
	{
	  sec--;
	  usec += 1000000;
	}
      printf("Rendering of %d responses consumes: %ld sec, %lf msec\n",
	     count, static_cast<long>(sec), static_cast<double>(usec / 1000.0));
#endif
      GtkWidget *board_view
	= view->controller.get_view_manager().get_current_board_view();
      if (board_view != NULL)
	gtk_widget_queue_draw(board_view);
    }
  else
    result = TRUE;	// ϸ

  gdk_threads_leave();

  return result;
}


void
BBSThreadView::state_changed_cb()
{
  // sigslotʥϥɥ
  gdk_threads_enter();

  if (append_responses_id == 0 && response_cursor != NULL)
    // idleؿϿ
    append_responses_id
      = g_idle_add_full(THREAD_VIEW_UPDATE_PRIORITY,
			reinterpret_cast<GSourceFunc>(append_responses_on_idle),
			this, NULL);

  gdk_threads_leave();
}


GtkWidget *
BBSThreadView::create(Controller &controller, BBSThread *thread,
		      TabLabel *label, int initial_pos)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  BBSThreadView *view = new BBSThreadView(controller, thread);
  GtkWidget *widget = view->widget;

  view->initial_view_position = initial_pos;
  view->update(widget, label, false);	// ɽϤ
  return widget;
}


void
BBSThreadView::update(GtkWidget *widget, TabLabel *label, bool automatic)
{
  // gdk_threads_enter()ĶƤ֤ȡ
  assert(Lock::get_slave_lock()->is_locked_by_self());
  BBSThreadView *view = get_view(widget);
  assert(view != NULL);

  view->thread->set_number_of_responses_read(view->thread->get_number_of_responses_got());
  ResponseCursor *cursor = view->response_cursor;
  if (cursor == NULL)
    {
      Controller &controller = view->controller;

      // BBSThread::get_responses()sigslotȤΤglobal lock롣
      // ʤΤǡǰögdk_threads_leave()롣
      gdk_threads_leave();

      smart_ptr<ResponseCursor> new_cursor
	= view->thread->get_responses(controller.get_config(),
				      controller.get_repository(),
				      controller.get_thread_pool(),
				      controller.get_broker());
      new_cursor->access_started.connect(label, &TabLabel::access_started);
      new_cursor->access_progressed.connect(label,
					    &TabLabel::access_progressed);
      new_cursor->access_finished.connect(label, &TabLabel::access_finished);
      new_cursor->access_failed.connect(label, &TabLabel::access_failed);
      new_cursor->access_terminated.connect(label,
					    &TabLabel::access_terminated);
      new_cursor->state_changed.connect(view, &BBSThreadView::state_changed_cb);
      view->response_cursor = new_cursor;

      int bookmark_pos = view->thread->get_number_of_responses_read();
      if (new_cursor != NULL)
	{
	  if (new_cursor->update())
	    view->my_own_bookmark_position = bookmark_pos;
	  else
	    {
	      if (view->my_own_bookmark_position == 0)
		view->my_own_bookmark_position = bookmark_pos;
	      label->access_finished();
	    }
	  if (view->initial_view_position == 0)
	    view->initial_view_position = bookmark_pos;
	}

      gdk_threads_enter();
    }
  else
    {
      if (!automatic)
	{
	  gdk_threads_leave();

	  if (cursor->update())
	    {
	      gdk_threads_enter();
	      size_t bookmark_pos
		= view->thread->get_number_of_responses_read();
	      if (view->my_own_bookmark_position != bookmark_pos)
		{
		  view->my_own_bookmark_position = bookmark_pos;
		  view->insert_my_own_bookmark(view->my_own_bookmark_position);
		  view->initial_view_position = bookmark_pos;
		}
	    }
	  else
	    gdk_threads_enter();
	}
    }

  if (view->append_responses_id == 0 && view->response_cursor != NULL)
    view->append_responses_id
      = g_idle_add_full(THREAD_VIEW_UPDATE_PRIORITY,
			reinterpret_cast<GSourceFunc>(append_responses_on_idle),
			view, NULL);

  if (automatic)
    {
      if (view->auto_refresher_id == 0)
	{
	  GdkColor *fg = view->controller.get_prefs().get_auto_refresh_go();
	  view->tab_label = label->get_widget();
	  gtk_widget_modify_fg(view->tab_label, GTK_STATE_NORMAL, fg);
	  gtk_widget_modify_fg(view->tab_label, GTK_STATE_ACTIVE, fg);
	  view->previous_interval = 10000;
	  view->previous_number_of_responses_got
	    = view->thread->get_number_of_responses_read();
	  view->auto_refresher_id
	    = g_timeout_add(view->previous_interval,
			    reinterpret_cast<GSourceFunc>(auto_refresher_timeout_cb), view);
	}
      else
	{
	  g_source_remove(view->auto_refresher_id);
	  view->auto_refresher_id = 0;
	  gtk_widget_modify_fg(view->tab_label, GTK_STATE_NORMAL, NULL);
	  gtk_widget_modify_fg(view->tab_label, GTK_STATE_ACTIVE, NULL);
	  view->tab_label = NULL;
	}
    }
}


BBSThreadView *
BBSThreadView::get_view(GtkWidget *widget)
{
  return dynamic_cast<BBSThreadView *>(BBSThreadViewBase::get_view(widget));
}


void
BBSThreadView::show()
{
  // ʤˤ⤻
}


void
BBSThreadView::hide(bool now)
{
  // ʤˤ⤻
}


void
BBSThreadView::show_post_window()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (post_window != NULL)
    {
      gtk_window_present(GTK_WINDOW(post_window));
      return;
    }

  ResponseEditorWindow *window = new ResponseEditorWindow(controller, thread);
  post_window = window->get_widget();
  post_window_destroy_id
    = g_signal_connect(post_window, "destroy",
		       reinterpret_cast<GCallback>(post_window_destroy_cb),
		       this);
  gtk_widget_show(post_window);
}


void
BBSThreadView::jump_to_bookmark()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  GtkTextMark *mark = gtk_text_buffer_get_mark(GTK_TEXT_BUFFER(buffer),
					       "my-own-bookmark");
  g_return_if_fail(mark != NULL);

  int current_pos = ochusha_hypertext_view_get_visible_offset(view);
  backward_history.push_front(current_pos);

  while (!forward_history.empty())
    forward_history.pop_front();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_scroll_to_mark(view, mark);
}


void
BBSThreadView::specify_view_position(int position)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  guint pos;
  if (position > 0)
    pos = position;
  else
    pos = records.size() + position;
 if (records.size() < pos)
    return;
 
  int current_pos = ochusha_hypertext_view_get_visible_offset(view);
  backward_history.push_front(current_pos);

  while (!forward_history.empty())
    forward_history.pop_front();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  GtkTextMark *mark
    = ochusha_hypertext_buffer_create_mark_at_offset(buffer, "view-position",
						     records[pos - 1].offset);
  ochusha_hypertext_view_scroll_to_mark(view, mark);
}


void
BBSThreadView::goto_bottom()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  int current_pos = ochusha_hypertext_view_get_visible_offset(view);
  backward_history.push_front(current_pos);

  while (!forward_history.empty())
    forward_history.pop_front();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_scroll_to_end(view);
}


void
BBSThreadView::goto_top()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  int current_pos = ochusha_hypertext_view_get_visible_offset(view);
  backward_history.push_front(current_pos);

  while (!forward_history.empty())
    forward_history.pop_front();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_scroll_to_start(view);
}


void
BBSThreadView::goto_next()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  g_return_if_fail(!forward_history.empty());

  gint current_pos = ochusha_hypertext_view_get_visible_offset(view);
  backward_history.push_front(current_pos);

  gint offset = forward_history.front();
  forward_history.pop_front();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_make_offset_visible(view, offset);
}


void
BBSThreadView::goto_prev()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  g_return_if_fail(!backward_history.empty());

  gint current_pos = ochusha_hypertext_view_get_visible_offset(view);
  forward_history.push_front(current_pos);

  gint offset = backward_history.front();
  backward_history.pop_front();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_make_offset_visible(view, offset);
}


void
BBSThreadView::post_window_destroy_cb(GtkWidget *widget, BBSThreadView *view)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  view->post_window = 0;
  view->post_window_destroy_id = 0;
}


gboolean
BBSThreadView::auto_refresher_timeout_cb(BBSThreadView *view)
{
  gdk_threads_enter();

  GtkTextView *text_view = GTK_TEXT_VIEW(view->view);
  GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(text_view);
  GdkRectangle rect;
  GtkTextIter iter;
  gint y;
  gint height;

  gtk_text_view_get_visible_rect(text_view, &rect);
  gtk_text_buffer_get_end_iter(text_buffer, &iter);
  gtk_text_view_get_line_yrange(text_view, &iter, &y, &height);

  bool skip_refresh;
  if (rect.y + rect.height < y)
    {
      // ɽƤʤϼư򥹥å
      skip_refresh = TRUE;
      GdkColor *fg = view->controller.get_prefs().get_auto_refresh_wait();
      gtk_widget_modify_fg(view->tab_label, GTK_STATE_NORMAL, fg);
      gtk_widget_modify_fg(view->tab_label, GTK_STATE_ACTIVE, fg);
    }
  else
    {
      skip_refresh = FALSE;
      GdkColor *fg = view->controller.get_prefs().get_auto_refresh_go();
      gtk_widget_modify_fg(view->tab_label, GTK_STATE_NORMAL, fg);
      gtk_widget_modify_fg(view->tab_label, GTK_STATE_ACTIVE, fg);
    }

  int num_got = view->thread->get_number_of_responses_got();
  if (view->previous_number_of_responses_got != num_got)
    {
      if (!skip_refresh)
	{
	  view->previous_interval -= 1000;
	  if (view->previous_interval < 10000)
	    view->previous_interval = 10000;
	}
      else
	num_got = view->previous_number_of_responses_got;
    }
  else
    {
      view->previous_interval += 10000;
      if (view->thread->is_stopped() || view->previous_interval > 300000)
	{
	  // ߤ
	  gtk_widget_modify_fg(view->tab_label, GTK_STATE_NORMAL, NULL);
	  gtk_widget_modify_fg(view->tab_label, GTK_STATE_ACTIVE, NULL);
	  view->previous_interval = 0;
	}
    }
  view->previous_number_of_responses_got = num_got;

  if (view->previous_interval != 0)
    view->auto_refresher_id
      = g_timeout_add(view->previous_interval,
		      reinterpret_cast<GSourceFunc>(auto_refresher_timeout_cb),
		      view);
  else
    view->auto_refresher_id = 0;

  if (!skip_refresh)
    {
      gdk_threads_leave();

      bool result = view->response_cursor->update();
      if (result)
	{
	  gdk_threads_enter();

	  if (view->append_responses_id == 0)
	    view->append_responses_id
	      = g_idle_add_full(THREAD_VIEW_UPDATE_PRIORITY,
				reinterpret_cast<GSourceFunc>(append_responses_on_idle), view, NULL);

	  gdk_threads_leave();
	}
      return FALSE;
    }

  gdk_threads_leave();
  return FALSE;
}

}
