/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main_window.cc,v 1.39 2007/01/01 03:18:25 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "main_window.h"

#include "controller.h"
#include "session_client.h"
#include "ui_constants.h"
#include "view_manager.h"

#include "libochushagtk/searchbar.h"
#include "libochushagtk/stock.h"
#include "libochushagtk_lgpl/mod_notebook.h"

#include "ochusha48.h"


#define DEFAULT_WINDOW_WIDTH	600
#define DEFAULT_WINDOW_HEIGHT	400
#define DEFAULT_SIDEBAR_WIDTH	150


namespace Ochusha
{


static void
find_again_notify_sensitive_cb(GtkWidget *button, GParamSpec *arg1,
			       GtkAction *action)
{
  gtk_action_set_sensitive(action, GTK_WIDGET_IS_SENSITIVE(button));
}


MainWindow::MainWindow(Controller &controller)
  : controller(controller),
    action_group(create_action_group()),
    history_view(NULL),
    post_history_view(NULL),
    shown_sidebar_items(0),
    view_manager(controller),
    maximized(false)
{
  static GdkGeometry geometry = { 0, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0,
				  GDK_GRAVITY_NORTH_WEST };

  const char *id = controller.get_session().get_sm_client_id();
  if (id != NULL)
    gdk_set_sm_client_id(id);

  setup_main_ui();

  toplevel = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
  g_signal_connect(toplevel, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &toplevel);
  g_signal_connect(toplevel, "delete_event",
		   reinterpret_cast<GCallback>(delete_event_cb), this);
  g_signal_connect(toplevel, "window_state_event",
		   reinterpret_cast<GCallback>(window_state_event_cb), this);

  GtkAccelGroup *accel_group
    = gtk_ui_manager_get_accel_group(controller.get_ui_manager());
  gtk_window_add_accel_group(toplevel, accel_group);

  gtk_window_set_title(toplevel, _("Ochusha"));
  gtk_window_set_icon(toplevel,
		      gdk_pixbuf_new_from_inline(-1, ochusha48_inline, FALSE,
						 NULL));

  // 礭
  int width;
  const char *value = controller.get_session().get_state("MainWindow::width");
  if (value == NULL || sscanf(value, "%d", &width) == 0 || width == 0)
    width = DEFAULT_WINDOW_WIDTH;

  int height;
  value = controller.get_session().get_state("MainWindow::height");
  if (value == NULL || sscanf(value, "%d", &height) == 0 || height == 0)
    height = DEFAULT_WINDOW_HEIGHT;

  gtk_widget_set_size_request(GTK_WIDGET(toplevel), width, height);

  int x;
  value = controller.get_session().get_state("MainWindow::x");
  if (value == NULL || sscanf(value, "%d", &x) != 1)
    x = -1;

  int y;
  value = controller.get_session().get_state("MainWindow::y");
  if (value == NULL || sscanf(value, "%d", &y) != 1)
    y = -1;

  if (x >= 0 && y >= 0)
    {
      gtk_widget_realize(GTK_WIDGET(toplevel));
      gtk_window_move(toplevel, x, y);
      gtk_widget_queue_resize(GTK_WIDGET(toplevel));
    }

  int m = 0;
  value = controller.get_session().get_state("MainWindow::maximized");
  if (value != NULL && sscanf(value, "%d", &m) == 1 && m != 0)
    {
      maximized = true;
      gtk_window_maximize(toplevel);
    }

  main_box = GTK_BOX(gtk_vbox_new(FALSE, 0));
  gtk_container_add(GTK_CONTAINER(toplevel), GTK_WIDGET(main_box));
  gtk_window_set_geometry_hints(toplevel, GTK_WIDGET(main_box),
				&geometry, GDK_HINT_MIN_SIZE);

  menubar = gtk_ui_manager_get_widget(controller.get_ui_manager(), "/MainMenu");
  gtk_widget_show(menubar);
  gtk_box_pack_start(main_box, menubar, FALSE, TRUE, 0);

  navigation_toolbar = gtk_ui_manager_get_widget(controller.get_ui_manager(),
						 "/NavigationToolbar");
  GtkToggleAction *action = controller.get_prefs().show_navigation_toolbar;
  if (navigation_toolbar != NULL)
    {
      gtk_toolbar_set_style(GTK_TOOLBAR(navigation_toolbar),
			    GTK_TOOLBAR_ICONS);
      gtk_box_pack_start(main_box, navigation_toolbar, FALSE, TRUE, 0);

      if (gtk_toggle_action_get_active(action))
	gtk_widget_show(navigation_toolbar);
      else
	gtk_widget_hide(navigation_toolbar);
      g_signal_connect(action, "toggled",
		       reinterpret_cast<GCallback>(show_navigation_toolbar_toggled_cb), this);
    }
  else
    gtk_action_set_sensitive(GTK_ACTION(action), FALSE);

  main_container = gtk_hpaned_new();
  g_signal_connect(main_container, "notify::position-set",
		   reinterpret_cast<GCallback>(notify_position_set_cb), this);
  gtk_widget_show(main_container);
  gtk_box_pack_start(main_box, main_container, TRUE, TRUE, 1);

  value = controller.get_session().get_state("MainWindow::sidebar_width");
  if (value != NULL && sscanf(value, "%d", &width) == 1)
    gtk_paned_set_position(GTK_PANED(main_container), width);


  // ɥСν
  GtkWidget *widget = mod_notebook_new();
  g_signal_connect(widget, "button_press_event",
		   reinterpret_cast<GCallback>(sidebar_button_press_event_cb),
		   this);
  sidebar_book = MOD_NOTEBOOK(widget);
  mod_notebook_set_scrollable(sidebar_book, TRUE);
  mod_notebook_set_show_tabs(sidebar_book, TRUE);
  mod_notebook_set_minimum_tab_label_size(sidebar_book, 0);
  mod_notebook_set_tab_shrinkable(sidebar_book, TRUE);
  mod_notebook_tooltips_enable(sidebar_book);


  // İ
  boardlist_view = controller.boardlist_ui.get_widget();
  action = controller.get_prefs().show_boardlist_sidebar;
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(boardlist_view);
      shown_sidebar_items |= BOARDLIST_SIDEBAR;
    }
  else
    gtk_widget_hide(boardlist_view);
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_boardlist_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, boardlist_view,
			   gtk_label_new(_("Boardlist")));


  // ֥åޡ
  bookmarks_view = controller.bookmarks_ui.get_widget();
  action = controller.get_prefs().show_bookmarks_sidebar;
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(bookmarks_view);
      shown_sidebar_items |= BOOKMARKS_SIDEBAR;
    }
  else
    gtk_widget_hide(bookmarks_view);
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_bookmarks_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, bookmarks_view,
			   gtk_label_new(_("Bookmarks")));

#if 0
  // 
  history_view = controller.history_ui.get_widget();
  action = controller.get_prefs().show_history_sidebar;
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_hide(history_view);
      gtk_widget_show(history_view);
    }
  else
    shown_sidebar_items |= HISTORY_SIDEBAR;
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_history_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, history_view,
			   gtk_label_new(_("History")));


  // 
  post_history_view = controller.post_history_ui.get_widget();
  action = controller.get_prefs().show_post_history_sidebar;
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(post_history_view);
      gtk_widget_hide(post_history_view);
    }
  else
    shown_sidebar_items |= POST_HISTORY_SIDEBAR;
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_post_history_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, post_history_view,
			   gtk_label_new(_("Post History")));
#endif

  if ((shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    {
      int item;
      value = controller.get_session().get_state("MainWindow::current_sidebar_item");
      if (value != NULL && sscanf(value, "%d", &item) == 1)
	{
	  GtkWidget *widget;
	  if (item == BOARDLIST_SIDEBAR)
	    widget = boardlist_view;
	  else if (item == BOOKMARKS_SIDEBAR)
	    widget = bookmarks_view;
	  else if (item == HISTORY_SIDEBAR)
	    widget = history_view;
	  else if (item == POST_HISTORY_SIDEBAR)
	    widget = post_history_view;
	  else
	    widget = NULL;

	  if (widget != NULL)
	    {
	      gint page = mod_notebook_page_num(sidebar_book, widget);
	      if (page != -1)
		mod_notebook_set_current_page(sidebar_book, page);
	    }
	}
      
      gtk_widget_show(widget);
    }
  else
    gtk_widget_hide(widget);
  gtk_paned_add1(GTK_PANED(main_container), widget);


  widget = view_manager.get_widget();
  gtk_widget_show(widget);
  gtk_paned_add2(GTK_PANED(main_container), widget);


  statusbar = GTK_STATUSBAR(gtk_statusbar_new());
  //gtk_statusbar_set_has_resize_grip(statusbar, FALSE);


  // СϢ
  searchbar = ochusha_searchbar_new();
  gtk_widget_grab_focus(searchbar);
  gtk_box_pack_start(main_box, searchbar, FALSE, FALSE, 0);

  OchushaSearchbar *sb = OCHUSHA_SEARCHBAR(searchbar);
  ochusha_searchbar_set_enable_incremental_search(sb, TRUE);
  GtkAction *find_again
    = gtk_action_group_get_action(action_group, "FindAgain");
  gtk_action_set_sensitive(find_again,
			   GTK_WIDGET_IS_SENSITIVE(sb->find_next_button));
  g_signal_connect_swapped(find_again, "activate",
			   reinterpret_cast<GCallback>(gtk_button_clicked),
			   sb->find_next_button);
  g_signal_connect(sb->find_next_button, "notify::sensitive",
		   reinterpret_cast<GCallback>(find_again_notify_sensitive_cb),
		   find_again);

  int b;
  value = controller.get_session().get_state("Searchbar::match_case");
  if (value != NULL && sscanf(value, "%d", &b) == 1)
    gtk_toggle_button_set_active(sb->match_case_button, b);

  value = controller.get_session().get_state("Searchbar::use_regexp");
  if (value != NULL && sscanf(value, "%d", &b) == 1)
    gtk_toggle_button_set_active(sb->use_regexp_button, b);

  g_signal_connect(sb->match_case_button, "toggled",
		   reinterpret_cast<GCallback>(match_case_toggled_cb), this);
  g_signal_connect(sb->use_regexp_button, "toggled",
		   reinterpret_cast<GCallback>(use_regexp_toggled_cb), this);

  searchbar_context_id
    = gtk_statusbar_get_context_id(statusbar, "searchbar-target");
  boardlist_context_id
    = gtk_statusbar_get_context_id(statusbar, "boardlist-info");
  link_context_id
    = gtk_statusbar_get_context_id(statusbar, "link-target");
  network_context_id
    = gtk_statusbar_get_context_id(statusbar, "network-access-status");

  gdk_threads_leave();

  controller.get_broker()
    .notify_access_status.connect(this, &MainWindow::notify_access_status_cb);

  gdk_threads_enter();

  // ơС
  status_box = GTK_BOX(gtk_hbox_new(FALSE, 0));


  widget = gtk_event_box_new();
  g_signal_connect(widget, "button_press_event",
		   reinterpret_cast<GCallback>(connection_image_pressed_cb),
		   &controller);

  connection_image
    = gtk_image_new_from_stock(controller.get_config().get_offline()
			       ? GTK_STOCK_DISCONNECT : GTK_STOCK_CONNECT,
			       GTK_ICON_SIZE_MENU);
  gtk_widget_show(connection_image);
  gtk_container_add(GTK_CONTAINER(widget), connection_image);
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(widget), FALSE);
  gtk_widget_show(widget);
  gtk_box_pack_start(status_box, widget, FALSE, FALSE, 2);


  widget = gtk_event_box_new();
  g_signal_connect(widget, "button_press_event",
		   reinterpret_cast<GCallback>(edit_image_pressed_cb),
		   controller.prefs.enable_embeded_response_editor);

  g_signal_connect_swapped(view_manager.thread_pane, "hide",
			   reinterpret_cast<GCallback>(gtk_widget_hide),
			   widget);
  g_signal_connect_swapped(view_manager.thread_pane, "show",
			   reinterpret_cast<GCallback>(gtk_widget_show),
			   widget);

  edit_image
    = gtk_image_new_from_stock(controller.prefs.get_enable_embeded_response_editor()
			       ? OCHUSHA_STOCK_EDIT_ENABLED
			       : OCHUSHA_STOCK_EDIT_DISABLED,
			       GTK_ICON_SIZE_MENU);


  gtk_widget_show(edit_image);
  gtk_container_add(GTK_CONTAINER(widget), edit_image);
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(widget), FALSE);
  //gtk_widget_show(widget);
  gtk_box_pack_end(GTK_BOX(statusbar), widget, FALSE, FALSE, 0);


  gtk_widget_show(GTK_WIDGET(statusbar));
  gtk_box_pack_start(status_box, GTK_WIDGET(statusbar), TRUE, TRUE, 0);
  gtk_widget_show(GTK_WIDGET(status_box));
  gtk_box_pack_end(main_box, GTK_WIDGET(status_box), FALSE, FALSE, 0);

  g_signal_connect(controller.work_offline_action, "toggled",
		   reinterpret_cast<GCallback>(work_offline_toggled_cb), this);
  g_signal_connect(controller.prefs.enable_embeded_response_editor, "toggled",
		   reinterpret_cast<GCallback>(embeded_response_editor_toggled_cb),
		   this);

  gtk_widget_show(GTK_WIDGET(main_box));
}


MainWindow::~MainWindow()
{
  g_object_unref(action_group);
}


GtkActionGroup *
MainWindow::create_action_group()
{
  static GtkActionEntry menubar_entries[] = {
    {
      "FileMenu", NULL,
      N_("_File"), NULL,
      NULL,
      NULL,
    },
    {
      "EditMenu", NULL,
      N_("_Edit"), NULL,
      NULL,
      NULL,
    },
    {
      "ViewMenu", NULL,
      N_("_View"), NULL,
      NULL,
      NULL,
    },
    {
      "GoMenu", NULL,
      N_("_Go"), NULL,
      NULL,
      NULL,
    },
    {
      "BoardlistMenu", NULL,
      N_("Boar_dlist"), NULL,
      N_("Boardlist"),
      NULL,
    },
    {
      "BookmarksMenu", NULL,
      N_("_Bookmarks"), NULL,
      N_("Bookmarks"),
      NULL,
    },
    {
      "ToolsMenu", NULL,
      N_("_Tools"), NULL,
      N_("Tools"),
      NULL,
    },
    {
      "HelpMenu", NULL,
      N_("_Help"), NULL,
      NULL,
      NULL,
    },
    {
      "ViewToolbarMenu", NULL,
      N_("_Toolbar"), NULL,
      NULL,
      NULL,
    },
    {
      "BoardlistViewMenu", NULL,
      N_("Boar_dlist"), NULL,
      N_("Boardlist"),
      NULL,
    },
    {
      "ThreadlistViewMenu", NULL,
      N_("Thread_list"), NULL,
      N_("Threadlist"),
      NULL
    },
    {
      "ThreadViewMenu", NULL,
      N_("_Thread"), NULL,
      N_("Thread"),
      NULL
    },
    {
      "ViewSidebarMenu", NULL,
      N_("Sid_ebar"), NULL,
      NULL,
      NULL,
    },
    {
      "FontsMenu", NULL,
      N_("_Fonts"), NULL,
      NULL,
      NULL,
    },
    {
      "FindAgain", NULL,
      N_("Find A_gain"), NULL,
      N_("Find Again"),
      NULL,
    },
    {
      "GoBack", GTK_STOCK_GO_BACK,
      N_("_Back"), "<alt>Left",
      N_("Back"),
      NULL	/* reinterpret_cast<GCallback>(go_back_cb) */
    },
    {
      "GoForward", GTK_STOCK_GO_FORWARD,
      N_("_Forward"), "<alt>Right",
      N_("Forward"),
      NULL	/* reinterpret_cast<GCallback>(go_forward_cb) */
    }
  };

  action_group = gtk_action_group_new("MenubarActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group,
			       menubar_entries,
			       G_N_ELEMENTS(menubar_entries),
			       this);

  /* ̤ʵǽ򻦤 */
  GtkAction *action = gtk_action_group_get_action(action_group, "GoBack");
  gtk_action_set_sensitive(action, FALSE);

  action = gtk_action_group_get_action(action_group, "GoForward");
  gtk_action_set_sensitive(action, FALSE);


  /* GtkUIManager˻Ź */
  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);

  return action_group;
}


void
MainWindow::setup_main_ui()
{
  static const char *fallback_ui =
    "<ui>"
    "  <menubar name=\"MainMenu\">"
    "    <menu action=\"FileMenu\">"
    "      <menuitem action=\"OpenURL\"/>"
    "      <menuitem action=\"CloseThisPage\"/>"
    "      <separator/>"
    "      <menuitem action=\"ClearCookies\"/>"
    "      <separator/>"
    "      <menuitem action=\"Login2chBe\"/>"
#if 0
    "      <menuitem action=\"Login2chViewer\"/>"
#endif
    "      <menuitem action=\"WorkOffline\"/>"
    "      <separator/>"
    "      <menuitem action=\"Quit\"/>"
    "    </menu>"
    "    <menu action=\"EditMenu\">"
    "      <menuitem action=\"NewThread\"/>"
    "      <menuitem action=\"WriteResponse\"/>"
    "      <separator/>"
    "      <menuitem action=\"SearchBoardlist\"/>"
    "      <menuitem action=\"SearchThreadlist\"/>"
    "      <menuitem action=\"SearchThread\"/>"
    "      <separator/>"
    "      <menuitem action=\"Preferences\"/>"
    "    </menu>"
    "    <menu action=\"ViewMenu\">"
    "      <menu action=\"ViewToolbarMenu\">"
#if 0
    "        <menuitem action=\"NavigationToolbar\"/>"
    "        <menuitem action=\"BoardlistToolbar\"/>"
    "        <menuitem action=\"BookmarksToolbar\"/>"
#endif
    "        <menuitem action=\"ThreadlistToolbar\"/>"
    "        <menuitem action=\"ThreadToolbar\"/>"
    "      </menu>"
    "      <menu action=\"ViewSidebarMenu\">"
    "        <menuitem action=\"BoardlistSidebar\"/>"
    "        <menuitem action=\"BookmarksSidebar\"/>"
#if 0
    "        <menuitem action=\"HistorySidebar\"/>"
    "        <menuitem action=\"PostHistorySidebar\"/>"
#endif
    "      </menu>"
    "      <menu action=\"BoardlistViewMenu\">"
    "        <menuitem action=\"ShowHiddenBoardlistItems\"/>"
    "      </menu>"
    "      <menu action=\"ThreadlistViewMenu\">"
    "        <menuitem action=\"DefaultThreadlistRule\"/>"
    "        <menuitem action=\"ShowHiddenThreads\"/>"
    "      </menu>"
    "      <menu action=\"ThreadViewMenu\">"
    "        <menuitem action=\"ShowMailto\"/>"
    "        <menuitem action=\"InlineImage\"/>"
    "        <menuitem action=\"PopupImage\"/>"
    "        <separator/>"
    "        <menuitem action=\"TransparentABone\"/>"
#if 0
    "        <menuitem action=\"EnableChainABone\"/>"
#endif
    "        <menuitem action=\"DisableABone\"/>"
    "        <separator/>"
    "        <menuitem action=\"EmbededResponseEditor\"/>"
    "      </menu>"
    "      <menu action=\"FontsMenu\">"
    "        <menuitem action=\"SelectListViewFonts\"/>"
    "        <menuitem action=\"SelectThreadViewFonts\"/>"
    "      </menu>"
    "      <separator/>"
    "      <menuitem action=\"Stop\"/>"
    "      <menuitem action=\"RefreshThreadlist\"/>"
    "      <menuitem action=\"RefreshThread\"/>"
#if 0
    "      <menuitem action=\"AutoRefreshThread\"/>"
#endif
    "    </menu>"
#if 0
    "    <menu action=\"GoMenu\">"
    "      <menuitem action=\"GoBack\"/>"
    "      <menuitem action=\"GoForward\"/>"
    "      <separator/>"
    "    </menu>"
#endif
    "    <menu action=\"BoardlistMenu\">"
    "      <menuitem action=\"AddBoard\"/>"
    "      <menuitem action=\"ManageBoardlist\"/>"
    "      <separator/>"
    "      <menuitem action=\"RefreshBoardlist\"/>"
    "    </menu>"
    "    <menu action=\"BookmarksMenu\">"
    "      <menuitem action=\"BookmarkCurrentBoard\"/>"
    "      <menuitem action=\"BookmarkCurrentThread\"/>"
    "      <menuitem action=\"ManageBookmarks\"/>"
    "    </menu>"
    "    <menu action=\"ToolsMenu\">"
    "      <menuitem action=\"Notification\"/>"
    "      <menuitem action=\"DebugConsole\"/>"
    "      <menuitem action=\"SchemeConsole\"/>"
    "      <separator/>"
    "      <menuitem action=\"BoardInfo\"/>"
    "      <menuitem action=\"ThreadInfo\"/>"
    "    </menu>"
    "    <menu action=\"HelpMenu\">"
    "      <menuitem action=\"Help\"/>"
    "      <separator/>"
    "      <menuitem action=\"About\"/>"
    "    </menu>"
    "  </menubar>"
    "  <toolbar name=\"BoardViewToolbar\">"
    "    <toolitem action=\"RefreshThreadlist\"/>"
    "    <toolitem action=\"SearchThreadlist\"/>"
    "    <separator/>"
    "    <toolitem action=\"ThreadlistRule\"/>"
    "    <separator/>"
    "    <toolitem action=\"NewThread\"/>"
    "    <separator/>"
    "    <toolitem action=\"BoardArrowLeft\"/>"
    "    <toolitem action=\"BoardArrowRight\"/>"
    "    <toolitem action=\"ThreadlistTabs\"/>"
    "  </toolbar>"
    "  <toolbar name=\"ThreadViewToolbar\">"
    "    <toolitem action=\"GotoPrev\"/>"
    "    <toolitem action=\"GotoNext\"/>"
    "    <toolitem action=\"RefreshThread\"/>"
    "    <toolitem action=\"GotoTop\"/>"
    "    <toolitem action=\"GotoBottom\"/>"
    "    <toolitem action=\"JumpToBookmark\"/>"
    "    <toolitem action=\"SearchThread\"/>"
    "    <toolitem action=\"WriteResponse\"/>"
    "    <separator/>"
    "    <toolitem action=\"ThreadArrowLeft\"/>"
    "    <toolitem action=\"ThreadArrowRight\"/>"
    "    <toolitem action=\"ThreadTabs\"/>"
    "  </toolbar>"
    "  <popup name=\"BoardTabPopupMenu\">"
    "    <menuitem action=\"BookmarkBoardTab\"/>"
    "    <separator/>"
    "    <menuitem action=\"RefreshBoardTab\"/>"
    "    <separator/>"
    "    <menuitem action=\"CloseBoardTab\"/>"
    "  </popup>"
    "  <popup name=\"ThreadTabPopupMenu\">"
    "    <menuitem action=\"BookmarkThreadTab\"/>"
    "    <separator/>"
    "    <menuitem action=\"RefreshThreadTab\"/>"
    "    <menuitem action=\"AutoRefreshThreadTab\"/>"
    "    <separator/>"
    "    <menuitem action=\"CloseThreadTab\"/>"
    "  </popup>"
    "  <popup name=\"FoldertreePopupMenu\">"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <separator/>"
    "    <menuitem action=\"NewFolder\"/>"
    "    <separator/>"
    "    <menuitem action=\"CopyTreeItemName\"/>"
    "    <separator/>"
    "    <menuitem action=\"KillTreeItem\"/>"
    "    <separator/>"
    "    <menuitem action=\"TreeItemProperties\"/>"
    "  </popup>"
    "  <popup name=\"BoardlistPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocation\"/>"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"NewFolder\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <menuitem action=\"CopyTreeItemName\"/>"
    "    <menuitem action=\"CopyTreeItemNameURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"HideTreeItem\"/>"
    "    <separator/>"
    "    <menuitem action=\"KillTreeItem\"/>"
    "    <separator/>"
    "    <menuitem action=\"TreeItemProperties\"/>"
    "  </popup>"
    "  <popup name=\"UglyBoardlistPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocation\"/>"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <menuitem action=\"CopyTreeItemName\"/>"
    "    <menuitem action=\"CopyTreeItemNameURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"ToggleTreeItemHiddenState\"/>"
    "    <separator/>"
    "    <menuitem action=\"KillTreeItem\"/>"
    "    <separator/>"
    "    <menuitem action=\"TreeItemProperties\"/>"
    "  </popup>"
    "  <popup name=\"BookmarksPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocation\"/>"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"NewFolder\"/>"
    "    <separator/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <menuitem action=\"CopyTreeItemName\"/>"
    "    <menuitem action=\"CopyTreeItemNameURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"KillTreeItem\"/>"
    "    <separator/>"
    "    <menuitem action=\"TreeItemProperties\"/>"
    "  </popup>"
    "  <popup name=\"ThreadlistPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocation\"/>"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <menuitem action=\"CopyThreadTitle\"/>"
    "    <menuitem action=\"CopyThreadTitleURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"MarkThread\"/>"
    "    <menuitem action=\"UnmarkThread\"/>"
    "    <menuitem action=\"HideThread\"/>"
    "    <separator/>"
    "    <menuitem action=\"ResetThread\"/>"
    "    <separator/>"
    "    <menuitem action=\"ThreadProperties\"/>"
    "  </popup>"
    "  <popup name=\"UglyThreadlistPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocation\"/>"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <menuitem action=\"CopyThreadTitle\"/>"
    "    <menuitem action=\"CopyThreadTitleURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"MarkThread\"/>"
    "    <menuitem action=\"UnmarkThread\"/>"
    "    <menuitem action=\"ToggleThreadHiddenState\"/>"
    "    <separator/>"
    "    <menuitem action=\"ResetThread\"/>"
    "    <separator/>"
    "    <menuitem action=\"ThreadProperties\"/>"
    "  </popup>"
    "  <popup name=\"ThreadBBSLinkPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocation\"/>"
    "    <menuitem action=\"OpenLinkLocationInTab\"/>"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <separator/>"
    "  </popup>"
    "  <popup name=\"ThreadLinkPopupMenu\">"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <menuitem action=\"ExternalLinkTools\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"SaveLinkAs\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "  </popup>"
    "  <popup name=\"ThreadImagePopupMenu\">"
    "    <menuitem action=\"OpenLinkLocationWithWebBrowser\"/>"
    "    <menuitem action=\"ExternalLinkTools\"/>"
    "    <menuitem action=\"ExternalImageTools\"/>"
    "    <separator/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"TenureLinkImage\"/>"
    "    <menuitem action=\"SaveLinkAs\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <menuitem action=\"CopyFileLocation\"/>"
    "    <separator/>"
    "    <menuitem action=\"ABoneLinkImage\"/>"
    "    <menuitem action=\"ABoneLinkServer\"/>"
    "  </popup>"
    "  <popup name=\"ThreadResponseNumberPopupMenu\">"
    "    <menuitem action=\"WriteResponseToThisResponse\"/>"
    "    <menuitem action=\"InsertBookmark\"/>"
    "    <menuitem action=\"BookmarkLinkLocation\"/>"
    "    <menuitem action=\"CopyLinkLocation\"/>"
    "    <separator/>"
    "    <menuitem action=\"ABoneResponse\"/>"
    "    <menuitem action=\"UnABoneResponse\"/>"
    "  </popup>"
    "  <popup name=\"ThreadIDPopupMenu\">"
    "    <menuitem action=\"ABoneIDInThisThread\"/>"
    "    <menuitem action=\"ABoneIDInThisBoard\"/>"
    "  </popup>"
    "</ui>";

  char *filename = controller.get_repository().find_file(OCHUSHA_UI_XML);
  GError *error = NULL;
  if (filename != NULL)
    {
      if (!gtk_ui_manager_add_ui_from_file(controller.get_ui_manager(),
					   filename, &error))
	{
	  g_message("Building UI from file \"%s\" failed: %s\n"
		    "Fallback to built-in UI.\n", filename, error->message);
	  free(filename);
	}
      else
	{
	  free(filename);
	  goto done;
	}
    }

  if (!gtk_ui_manager_add_ui_from_string(controller.get_ui_manager(),
					 fallback_ui, -1, &error))
    {
      g_message("*FATAL* Building built-in UI failed: %s\n", error->message);
      g_error_free(error);
      exit(EXIT_FAILURE);
    }

 done:
  controller.boardlist_ui.setup_menubar();
  controller.bookmarks_ui.setup_menubar();
}


void
MainWindow::notify_position_set_cb(GtkPaned *paned, GParamSpec *arg1,
				   MainWindow *window)
{
#if 0
  fprintf(stderr, "MainWindow::notify_position_set_cb: position=%d\n",
	  gtk_paned_get_position(paned));
#endif
  char value[64];
  snprintf(value, 64, "%d", gtk_paned_get_position(paned));
  window->controller.get_session().set_state("MainWindow::sidebar_width",
					     value);
}


void
MainWindow::match_case_toggled_cb(GtkToggleButton *button, MainWindow *window)
{
  char value[64];
  snprintf(value, 64, "%d", gtk_toggle_button_get_active(button));
  window->controller.get_session().set_state("Searchbar::match_case", value);
}


void
MainWindow::use_regexp_toggled_cb(GtkToggleButton *button, MainWindow *window)
{
  char value[64];
  snprintf(value, 64, "%d", gtk_toggle_button_get_active(button));
  window->controller.get_session().set_state("Searchbar::use_regexp", value);
}


void
MainWindow::delete_event_cb(GtkWidget *widget, GdkEventAny *event,
			    MainWindow *window)
{
  window->controller.quit();
}


gboolean
MainWindow::window_state_event_cb(GtkWidget *widget, GdkEventWindowState *event,
				  MainWindow *window)
{
  int changed_mask = event->changed_mask;
  if ((changed_mask & static_cast<int>(GDK_WINDOW_STATE_MAXIMIZED)) == 0)
    return FALSE;

  int new_state = event->new_window_state;
  window->maximized
    = (new_state & static_cast<int>(GDK_WINDOW_STATE_MAXIMIZED)) != 0;

  char value[64];
  snprintf(value, 64, "%d", window->maximized);

  window->controller.get_session().set_state("MainWindow::maximized", value);

  return FALSE;
}


void
MainWindow::show_navigation_toolbar_toggled_cb(GtkToggleAction *action,
					       MainWindow *window)
{
  g_return_if_fail(window->navigation_toolbar != NULL);

  if (gtk_toggle_action_get_active(action))
    gtk_widget_show(window->navigation_toolbar);
  else
    gtk_widget_hide(window->navigation_toolbar);
}


gboolean
MainWindow::connection_image_pressed_cb(GtkWidget *event_box,
					GdkEventButton *event,
					Controller *controller)
{
  gboolean val = !gtk_toggle_action_get_active(controller->work_offline_action);
  gtk_toggle_action_set_active(controller->work_offline_action, val);

  return FALSE;
}


void
MainWindow::work_offline_toggled_cb(GtkToggleAction *action,
				    MainWindow *window)
{
  if (window->connection_image != NULL)
    gtk_image_set_from_stock(GTK_IMAGE(window->connection_image),
			     gtk_toggle_action_get_active(action)
			     ? GTK_STOCK_DISCONNECT : GTK_STOCK_CONNECT,
			     GTK_ICON_SIZE_MENU);
}


gboolean
MainWindow::edit_image_pressed_cb(GtkWidget *event_box,
				  GdkEventButton *event,
				  GtkToggleAction *action)
{
  gboolean value = !gtk_toggle_action_get_active(action);
  gtk_toggle_action_set_active(action, value);

  return FALSE;
}


void
MainWindow::embeded_response_editor_toggled_cb(GtkToggleAction *action,
					       MainWindow *window)
{
  if (window->edit_image != NULL)
    gtk_image_set_from_stock(GTK_IMAGE(window->edit_image),
			     gtk_toggle_action_get_active(action)
			     ? OCHUSHA_STOCK_EDIT_ENABLED
			     : OCHUSHA_STOCK_EDIT_DISABLED,
			     GTK_ICON_SIZE_MENU);
}


gboolean
MainWindow::sidebar_button_press_event_cb(ModNotebook *notebook,
					  GdkEventButton *event,
					  MainWindow *window)
{
  if (event->button != 2)
    return FALSE;

  gint page = mod_notebook_get_mouse_event_page_num(window->sidebar_book,
					reinterpret_cast<GdkEventAny *>(event));
  if (page != -1)
    {
      GtkWidget *widget = mod_notebook_get_nth_page(window->sidebar_book, page);
      if (widget == window->boardlist_view)
	window->controller.get_prefs().set_show_boardlist_sidebar(false);
      else if (widget == window->bookmarks_view)
	window->controller.get_prefs().set_show_bookmarks_sidebar(false);
      else if (widget == window->history_view)
	window->controller.get_prefs().set_show_history_sidebar(false);
      else if (widget == window->post_history_view)
	window->controller.get_prefs().set_show_post_history_sidebar(false);
      return TRUE;
    }
  return FALSE;
}


void
MainWindow::show_boardlist_sidebar_toggled_cb(GtkToggleAction *action,
					      MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->boardlist_view);
      window->shown_sidebar_items |= BOARDLIST_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->boardlist_view);
      g_assert(page != -1);
      mod_notebook_set_current_page(window->sidebar_book, page);
    }
  else
    {
      gtk_widget_hide(window->boardlist_view);
      window->shown_sidebar_items &= ~BOARDLIST_SIDEBAR;
    }
  if ((window->shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    gtk_widget_show(GTK_WIDGET(window->sidebar_book));
  else
    gtk_widget_hide(GTK_WIDGET(window->sidebar_book));
}


void
MainWindow::show_bookmarks_sidebar_toggled_cb(GtkToggleAction *action,
					      MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->bookmarks_view);
      window->shown_sidebar_items |= BOOKMARKS_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->bookmarks_view);
      g_assert(page != -1);
      mod_notebook_set_current_page(window->sidebar_book, page);
    }
  else
    {
      gtk_widget_hide(window->bookmarks_view);
      window->shown_sidebar_items &= ~BOOKMARKS_SIDEBAR;
    }
  if ((window->shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    gtk_widget_show(GTK_WIDGET(window->sidebar_book));
  else
    gtk_widget_hide(GTK_WIDGET(window->sidebar_book));
}


void
MainWindow::show_history_sidebar_toggled_cb(GtkToggleAction *action,
					    MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->history_view);
      window->shown_sidebar_items |= HISTORY_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->history_view);
      g_assert(page != -1);
      mod_notebook_set_current_page(window->sidebar_book, page);
    }
  else
    {
      gtk_widget_hide(window->history_view);
      window->shown_sidebar_items &= ~HISTORY_SIDEBAR;
    }
  if ((window->shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    gtk_widget_show(GTK_WIDGET(window->sidebar_book));
  else
    gtk_widget_hide(GTK_WIDGET(window->sidebar_book));
}


void
MainWindow::show_post_history_sidebar_toggled_cb(GtkToggleAction *action,
						 MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->post_history_view);
      window->shown_sidebar_items |= POST_HISTORY_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->post_history_view);
      g_assert(page != -1);
      mod_notebook_set_current_page(window->sidebar_book, page);
    }
  else
    {
      gtk_widget_hide(window->post_history_view);
      window->shown_sidebar_items &= POST_HISTORY_SIDEBAR;
    }
  if ((window->shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    gtk_widget_show(GTK_WIDGET(window->sidebar_book));
  else
    gtk_widget_hide(GTK_WIDGET(window->sidebar_book));
}


AccessMessage::AccessMessage(const char *url, guint message_id)
  : url(strdup(url)),
    message_id(message_id)
{
}


AccessMessage::AccessMessage(const AccessMessage &orig)
  : url(strdup(orig.url)),
    message_id(orig.message_id)
{
}


AccessMessage::~AccessMessage()
{
  free(url);
}


void
MainWindow::notify_access_status_cb(const char *url, NetworkAccessStatus status)
{
  gdk_threads_enter();

  list<AccessMessage>::iterator iter = access_message_list.begin();
  list<AccessMessage>::iterator end_iter = access_message_list.end();
  while (iter != end_iter)
    {
      if (strcmp((*iter).url, url) == 0)
	break;
      iter++;
    }

  if (iter != end_iter)
    gtk_statusbar_remove(statusbar, network_context_id, (*iter).message_id);

  if (status == ACCESS_FINISHED)
    {
      g_return_if_fail(iter != end_iter);
      access_message_list.erase(iter);

      if (access_message_list.begin() == access_message_list.end())
	gtk_action_set_sensitive(controller.stop_action, FALSE);

      gdk_threads_leave();

      return;
    }

  URI uri(url);
  gchar message[128];

  const char *server = uri.get_server();
  if (server == NULL)
    server = "unknown";	// ꤨ͡

  switch (status)
    {
    case LOOK_UP_SERVER:
      snprintf(message, 128, _("Looking up %s..."), server);
      break;

    case CONNECT_TO_SERVER:
      snprintf(message, 128, _("Connecting to %s..."), server);
      break;

    case SEND_REQUEST:
      snprintf(message, 128, _("Sending request to %s..."), server);
      break;

    case WAIT_FOR_SERVER:
      snprintf(message, 128, _("Waiting for %s..."), server);
      break;

    case TRANSFER_DATA:
      snprintf(message, 128, _("Transferring data from %s..."), server);
      break;

    default:
      abort();	// ꤨ͡
    }

  guint message_id = gtk_statusbar_push(statusbar, network_context_id, message);

  if (iter != end_iter)
    {
      (*iter).message_id = message_id;

      gdk_threads_leave();

      return;
    }

  gtk_action_set_sensitive(controller.stop_action, TRUE);

  access_message_list.push_front(AccessMessage(url, message_id));

  gdk_threads_leave();
}


void
MainWindow::show()
{
  gtk_widget_show(GTK_WIDGET(toplevel));
}


void
MainWindow::save_states()
{
}


void
MainWindow::prepare_session()
{
  g_return_if_fail(toplevel != NULL);

  GtkWidget *widget = GTK_WIDGET(toplevel);
  g_return_if_fail(GDK_IS_WINDOW(widget->window));

  SessionClient &session = controller.get_session();
  char buf[64];
  if (!maximized || session.get_state("MainWindow::x") == NULL)
    {
      int sx = gdk_screen_width();
      int sy = gdk_screen_height();
      int x, y;
      gdk_window_get_root_origin(widget->window, &x, &y);
      x %= sx;
      y %= sy;

      snprintf(buf, 64, "%d", x);
      session.set_state("MainWindow::x", buf);
      snprintf(buf, 64, "%d", y);
      session.set_state("MainWindow::y", buf);
      snprintf(buf, 64, "%d", widget->allocation.width);
      session.set_state("MainWindow::width", buf);
      snprintf(buf, 64, "%d", widget->allocation.height);
      session.set_state("MainWindow::height", buf);
    }

  if (shown_sidebar_items != 0)
    {
      gint page = mod_notebook_get_current_page(sidebar_book);
      g_return_if_fail(page != -1);
      GtkWidget *widget = mod_notebook_get_nth_page(sidebar_book, page);
      g_return_if_fail(widget != NULL);

      int current_sidebar_item;
      if (widget == boardlist_view)
	current_sidebar_item = BOARDLIST_SIDEBAR;
      else if (widget == bookmarks_view)
	current_sidebar_item = BOOKMARKS_SIDEBAR;
      else if (widget == history_view)
	current_sidebar_item = HISTORY_SIDEBAR;
      else if (widget == post_history_view)
	current_sidebar_item = POST_HISTORY_SIDEBAR;
      else
	current_sidebar_item = 0;

      if (current_sidebar_item != 0)
	{
	  snprintf(buf, 64, "%d", current_sidebar_item);
	  session.set_state("MainWindow::current_sidebar_item", buf);
	}
    }
}

}
