/* -*- c++ -*-
 * Copyright (c) 2003-2007 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: threadlist_view.cc,v 1.31 2007/01/06 14:57:53 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "threadlist_view.h"

#include "libochusha/ochusha_procedure.h"
#include "libochusha/thread_pool.h"

#include "controller.h"
#include "session_client.h"
#include "tab_label.h"

#include "libochushagtk/misc_utils.h"
#include "libochushagtk/searchable.h"
#include "libochushagtk/tree_selector_view.h"

#include <gtk/gtk.h>

#include <climits>


namespace Ochusha
{

#define POPUP_MENU_DELAY_TIMEOUT		250	// ms
#define THREADLIST_UPDATE_PRIORITY		GDK_PRIORITY_REDRAW + 5
#define MAX_NUM_OF_THREADS_TO_APPEND_AT_ONCE	500
#undef VERBOSE_DESTRUCTOR


enum {
  THREAD_DATA = 0,
  THREAD_COUNT,
  THREAD_WEIGHT,
  THREAD_SEARCH_WEIGHT,
  N_COLUMNS
};


enum {
  TARGET_INTERNAL_LIST_OF_THREADS = 0,
  TARGET_TEXT_URI_LIST,
  TARGET_UTF8_STRING
};


static const GtkTargetEntry all_targets[] = {
  { const_cast<char *>("internal-list-of-threads"),
    GTK_TARGET_SAME_APP, TARGET_INTERNAL_LIST_OF_THREADS },
  { const_cast<char *>("text/uri-list"), 0, TARGET_TEXT_URI_LIST },
  { const_cast<char *>("UTF8_STRING"), 0, TARGET_UTF8_STRING },
  { const_cast<char *>("STRING"), 0, TARGET_UTF8_STRING },
  { const_cast<char *>("TEXT"), 0, TARGET_UTF8_STRING },
  { const_cast<char *>("COMPOUND_TEXT"), 0, TARGET_UTF8_STRING }
};


static const GtkTargetEntry empty_targets[] = {
};


ThreadlistViewSelectionData::ThreadlistViewSelectionData(GtkTreeSelection *selection)
  : selection(selection)
{
  g_signal_connect(selection, "changed",
		   reinterpret_cast<GCallback>(changed_cb), this);
}


ThreadlistViewSelectionData::~ThreadlistViewSelectionData()
{
}


void
ThreadlistViewSelectionData::accumulate_threads(GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  ThreadlistViewSelectionData *selection_data
    = static_cast<ThreadlistViewSelectionData *>(data);
  selection_data->thread_list.push_back(thread);
}


void
ThreadlistViewSelectionData::changed_cb(GtkTreeSelection *selection,
					ThreadlistViewSelectionData *data)
{
  data->thread_list.clear();

  gtk_tree_selection_selected_foreach(selection, accumulate_threads, data);
}


const GtkTargetEntry *
ThreadlistViewSelectionData::get_targets()
{
  return thread_list.size() > 0 ? all_targets : empty_targets;
}


guint
ThreadlistViewSelectionData::get_n_targets()
{
  return thread_list.size() > 0 ? 6 : 0;
}


void
ThreadlistViewSelectionData::get_data(GtkClipboard *clipboard,
				      GtkSelectionData *data, guint info)
{
  set_data(data, info);
}


bool
ThreadlistViewSelectionData::clear_data(GtkClipboard *clipboard)
{
  return false;
}


bool
ThreadlistViewSelectionData::set_data(GtkSelectionData *data, guint info)
{
  switch (info)
    {
    case TARGET_INTERNAL_LIST_OF_THREADS:
      break;

    case TARGET_TEXT_URI_LIST:
      {
	char **uri_list
	  = static_cast<char **>(malloc(sizeof(char *)
					* (thread_list.size() + 1)));

	vector<BBSThread *>::iterator iter = thread_list.begin();
	vector<BBSThread *>::iterator end_iter = thread_list.end();
	int i = 0;
	while (iter != end_iter)
	  uri_list[i++] = (*iter++)->get_url();
	uri_list[i] = NULL;
	gboolean result = gtk_selection_data_set_uris(data, uri_list);
	while (--i >= 0)
	  free(uri_list[i]);
	free(uri_list);
	return result != FALSE;
      }

    case TARGET_UTF8_STRING:
      {
	string buf;
	vector<BBSThread *>::iterator iter = thread_list.begin();
	vector<BBSThread *>::iterator end_iter = thread_list.end();
	while (iter != end_iter)
	  {
	    BBSThread *thread = *iter++;
	    const char *title = thread->get_title();
	    if (title != NULL)
	      buf.append(title);
	    else
	      buf.append(_("Anonymous Thread"));
	    buf.append("\n");
	    char *url = thread->get_url();
	    buf.append(url).append("\n");
	    free(url);
	  }
	return gtk_selection_data_set_text(data, buf.c_str(), -1) != FALSE;
      }
    }

  return false;
}


SelectionData *
ThreadlistViewSelectionData::create_proxy()
{
  GtkWidget *widget = GTK_WIDGET(gtk_tree_selection_get_tree_view(selection));
  return new SelectionDataProxy(this, widget);
}


enum {
  SORT_BY_COUNT,
  SORT_BY_BOARD,
  SORT_BY_RANK,
  SORT_BY_MARK,
  SORT_BY_TITLE,
  SORT_BY_N_RES_VERBOSE,
  SORT_BY_N_RES_VERBOSE2,
  SORT_BY_N_RES_AVAILABLE,
  SORT_BY_N_RES_READ,
  SORT_BY_N_RES_UNGOT,
  SORT_BY_N_RES_TO_READ,
  SORT_BY_N_NEW_RESPONSES,
  SORT_BY_BIRTHTIME,
  SORT_BY_LAST_MODIFIED,
  SORT_BY_WEIGHT,
  SORT_BY_SEARCH_WEIGHT
};


#define BOARD_WIDTH		"ThreadlistView::board::width"
#define TITLE_WIDTH		"ThreadlistView::title::width"
#define RANK_WIDTH		"ThreadlistView::rank::width"
#define MARK_WIDTH		"ThreadlistView::mark::width"
#define COUNT_WIDTH		"ThreadlistView::count::width"
#define N_RES_VERBOSE_WIDTH	"ThreadlistView::n_res_verbose::width"
#define N_RES_VERBOSE2_WIDTH	"ThreadlistView::n_res_verbose2::width"
#define N_RES_AVAILABLE_WIDTH	"ThreadlistView::n_res_available::width"
#define N_RES_READ_WIDTH	"ThreadlistView::n_res_read::width"
#define N_RES_UNGOT_WIDTH	"ThreadlistView::n_res_ungot::width"
#define N_RES_TO_READ_WIDTH	"ThreadlistView::n_res_to_read::width"
#define N_NEW_RES_WIDTH		"ThreadlistView::n_new_res::width"
#define RANK_DIFF_WIDTH		"ThreadlistView::rank_diff::width"
#define BIRTHTIME_WIDTH		"ThreadlistView::birthtime::width"
#define LAST_MODIFIED_WIDTH	"ThreadlistView::last_modified::width"
#define WEIGHT_WIDTH		"ThreadlistView::weight::width"
#define SEARCH_WEIGHT_WIDTH	"ThreadlistView::search_weight::width"


// GtkWidgetThreadlistView֥ȤФ˻Ȥ
static GQuark view_id = g_quark_from_static_string("ThreadlistView::view");

static GQuark sort_column_id
  = g_quark_from_static_string("ThreadlistView::sort_column");
static GQuark sort_order_id
  = g_quark_from_static_string("ThreadlistView::sort_order");

static int compare_count(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			 gpointer unused);
static int compare_board(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			 gpointer unused);
static int compare_rank(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			gpointer unused);
static int compare_mark(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			gpointer unused);
static int compare_title(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			 gpointer unused);
static int compare_n_res_verbose(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);
static int compare_n_res_verbose2(GtkTreeModel *model, GtkTreeIter *a,
				  GtkTreeIter *b, gpointer unused);
static int compare_n_res_avail(GtkTreeModel *model, GtkTreeIter *a,
			       GtkTreeIter *b, gpointer unused);
static int compare_n_res_read(GtkTreeModel *model, GtkTreeIter *a,
			      GtkTreeIter *b, gpointer unused);
static int compare_n_res_ungot(GtkTreeModel *model, GtkTreeIter *a,
			       GtkTreeIter *b, gpointer unused);
static int compare_n_res_to_read(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);
static int compare_n_new_res(GtkTreeModel *model, GtkTreeIter *a,
			     GtkTreeIter *b, gpointer unused);
static int compare_birthtime(GtkTreeModel *model, GtkTreeIter *a,
			     GtkTreeIter *b, gpointer unused);
static int compare_last_modified(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);
static int compare_weight(GtkTreeModel *model, GtkTreeIter *a,
			  GtkTreeIter *b, gpointer unused);
static int compare_search_weight(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);


void
ThreadlistView::widget_destroy_cb(GtkWidget *widget, ThreadlistView *view)
{
  // ThreadlistViewsigslotΥϥɥĤΤdestructorglobal lock
  // 롣å뤿ᡢĴ롣
  gdk_threads_leave();

  LockBlock lock_obj(Lock::get_global_lock());

  gdk_threads_enter();

  delete view;
}


static gboolean
title_match_func(OchushaTreeSelectorView *view,
		 GtkTreeModel *model, GtkTreeIter *iter,
		 OchushaSearchKey *key, gpointer unused)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_val_if_fail(thread != NULL, FALSE);
  const char *title = thread->get_title();
  return ochusha_search_key_match_text(key, title != NULL ? title : "", -1);
}


ThreadlistView::ThreadlistView(Controller &controller, BulletinBoard *board,
			       TabLabel *label)
  : controller(controller),
    model(GTK_TREE_MODEL(gtk_list_store_new(N_COLUMNS,
					    G_TYPE_POINTER,	// BBSThread *
					    G_TYPE_INT,		// ɲý
					    G_TYPE_LONG,	// 
					    G_TYPE_LONG))),	// 
    widget(ochusha_tree_selector_view_new_with_model(model)),
    selection_data(gtk_tree_view_get_selection(GTK_TREE_VIEW(widget))),
    weight_column(NULL),
    last_sort_column(NULL),
    last_event_path(NULL),
    press_start_x(0),
    press_start_y(0),
    press_start_x_root(0),
    press_start_y_root(0),
    press_start_button(0),
    press_start_state(0),
    popup_delay_timeout_id(0),
    board(board),
    rule_info(NULL),
    append_threads_id(0),
    number_of_threads(0)
{
  GtkTreeSortable *sortable = GTK_TREE_SORTABLE(model);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_COUNT, compare_count,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_BOARD, compare_board,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_RANK, compare_rank,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_MARK, compare_mark,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_TITLE, compare_title,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_VERBOSE, compare_n_res_verbose,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_VERBOSE2,
				  compare_n_res_verbose2,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_AVAILABLE,
				  compare_n_res_avail,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_READ, compare_n_res_read,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_UNGOT, compare_n_res_ungot,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_TO_READ, compare_n_res_to_read,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_NEW_RESPONSES, compare_n_new_res,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_BIRTHTIME, compare_birthtime,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_LAST_MODIFIED, compare_last_modified,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_WEIGHT, compare_weight,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_SEARCH_WEIGHT, compare_search_weight,
				  NULL, NULL);

  gtk_tree_sortable_set_sort_column_id(sortable,
				       SORT_BY_WEIGHT,
				       GTK_SORT_ASCENDING);
  GtkTreeSelection *selection
    = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

  g_signal_connect(selection, "changed",
		   reinterpret_cast<GCallback>(selection_changed_cb), this);

  g_object_set_qdata(G_OBJECT(widget), view_id, this);
  g_signal_connect(widget, "destroy",
		   reinterpret_cast<GCallback>(widget_destroy_cb), this);

  setup_columns();
  ochusha_tree_selector_view_set_search_equal_func(OCHUSHA_TREE_SELECTOR_VIEW(widget),
						   title_match_func,
						   NULL, NULL);
  g_signal_connect(widget, "row_button_press",
		   reinterpret_cast<GCallback>(row_button_press_cb), this);
  g_signal_connect(widget, "row_button_release",
		   reinterpret_cast<GCallback>(row_button_release_cb), this);
  g_signal_connect(widget, "popup_menu",
		   reinterpret_cast<GCallback>(popup_menu_cb), this);
  g_signal_connect(widget, "motion_notify_event",
		   reinterpret_cast<GCallback>(motion_notify_cb), this);
  g_signal_connect(widget, "column_resized",
		   reinterpret_cast<GCallback>(column_resized_cb), this);

  // DnDݡȤμ¸Ƥ
  initialize_for_drag_source();
  // ޤDnDϢ

  update(widget, label, true);	// ȤϤ
}


ThreadlistView::~ThreadlistView()
{
  if (popup_delay_timeout_id != 0)
    g_source_remove(popup_delay_timeout_id);

  if (append_threads_id != 0)
    g_source_remove(append_threads_id);

#if VERBOSE_DESTRUCTOR
  fprintf(stderr, "ThreadlistView destructed: %p\n", this);
#endif

  g_object_unref(model);
}


void
ThreadlistView::column_set_mark(GtkTreeViewColumn *column,
				GtkCellRenderer *renderer,
				GtkTreeModel *model, GtkTreeIter *iter,
				gpointer data)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  GdkColor *bg = NULL;
  ThreadlistView *view = static_cast<ThreadlistView *>(data);
  SchemeManager &scheme = view->controller.get_scheme_manager();
  Cell *func = scheme.decorate_list_entry.get_cell();
  if (func != NULL)
    {
      /* SchemeؿƤǿ򤹤 */
      Cell *color_pair = scheme.get_core().evalf0(func, "%p", thread);
      if (color_pair != NULL && color_pair->is_pair())
	{
	  Cell *bg_obj = color_pair->get_cdr();
	  if (bg_obj->is_foreign_object())
	    bg = static_cast<GdkColor *>(bg_obj->get_foreign_pointer());
	  else
	    bg = NULL;
	}
    }

  g_object_set(renderer,
	       "active", thread->is_marked(),
	       "cell_background_gdk", bg,
	       NULL);
}


void
ThreadlistView::column_set_text(GtkCellRenderer *renderer, BBSThread *thread,
				const gchar *text, gpointer data)
{
  bool hidden = thread->is_hidden();

  GdkColor *fg = NULL;
  GdkColor *bg = NULL;
  ThreadlistView *view = static_cast<ThreadlistView *>(data);
  SchemeManager &scheme = view->controller.get_scheme_manager();
  Cell *func = scheme.decorate_list_entry.get_cell();

  if (func != NULL)
    {
      /* SchemeؿƤǿ򤹤 */
      Cell *color_pair = scheme.get_core().evalf0(func, "%p", thread);
      if (color_pair != NULL && color_pair->is_pair())
	{
	  Cell *fg_obj = color_pair->get_car();
	  if (fg_obj->is_foreign_object())
	    fg = static_cast<GdkColor *>(fg_obj->get_foreign_pointer());
	  else
	    fg = NULL;

	  Cell *bg_obj = color_pair->get_cdr();
	  if (bg_obj->is_foreign_object())
	    bg = static_cast<GdkColor *>(bg_obj->get_foreign_pointer());
	  else
	    bg = NULL;
	}
    }

  g_object_set(renderer,
	       "text", text,
	       "strikethrough", hidden,
	       "strikethrough_set", hidden,
	       "foreground_gdk", fg,
	       "cell_background_gdk", bg,
	       NULL);
}


void
ThreadlistView::column_set_board_name(GtkTreeViewColumn *column,
				      GtkCellRenderer *renderer,
				      GtkTreeModel *model, GtkTreeIter *iter,
				      gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  column_set_text(renderer, thread, thread->get_board()->get_name(), view);
}


void
ThreadlistView::column_set_title(GtkTreeViewColumn *column,
				 GtkCellRenderer *renderer,
				 GtkTreeModel *model, GtkTreeIter *iter,
				 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  const char *title = thread->get_title();
  if (title != NULL)
    {
      column_set_text(renderer, thread, title, view);
      return;
    }
  char title_buf[256];
  snprintf(title_buf, 256, _("(tmp)%s"), thread->get_id());
  column_set_text(renderer, thread, title_buf, view);
}


void
ThreadlistView::column_set_rank(GtkTreeViewColumn *column,
				GtkCellRenderer *renderer,
				GtkTreeModel *model, GtkTreeIter *iter,
				gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar rank_buf[16];
  const gchar *text;
  if (thread->is_alive())
    {
      snprintf(rank_buf, 16, "%d", thread->get_rank());
      text = rank_buf;
    }
  else if (thread->is_kako_thread())
    text = _("Past");
  else if (thread->is_viewer_used())
    text = _("$");
  else
    text = _("Dead");

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_verbose(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];

  int num_read = thread->get_number_of_responses_got();
  int num_to_read = thread->get_number_of_responses_to_read();
  int num_avail = thread->get_number_of_responses_available();
  if (num_read == 0)
    snprintf(text, 64, "%d", num_avail);
  else if (num_to_read == 0)
    snprintf(text, 64, "%d (0)", num_avail);
  else
    snprintf(text, 64, "%d (+%d)", num_avail, num_to_read);

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_verbose2(GtkTreeViewColumn *column,
					  GtkCellRenderer *renderer,
					  GtkTreeModel *model,
					  GtkTreeIter *iter,
					  gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];

  int num_read = thread->get_number_of_responses_got();
  int num_to_get = thread->get_number_of_responses_to_get();
  int num_avail = thread->get_number_of_responses_available();
  if (num_read == 0)
    snprintf(text, 64, "%d", num_avail);
  else if (num_to_get == 0)
    snprintf(text, 64, "%d (0)", num_avail);
  else
    snprintf(text, 64, "%d (+%d)", num_avail, num_to_get);

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_avail(GtkTreeViewColumn *column,
				       GtkCellRenderer *renderer,
				       GtkTreeModel *model, GtkTreeIter *iter,
				       gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_available());
  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_read(GtkTreeViewColumn *column,
				      GtkCellRenderer *renderer,
				      GtkTreeModel *model, GtkTreeIter *iter,
				      gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_got());
  column_set_text(renderer, thread, text, view);
}

   
void
ThreadlistView::column_set_n_res_ungot(GtkTreeViewColumn *column,
				       GtkCellRenderer *renderer,
				       GtkTreeModel *model, GtkTreeIter *iter,
				       gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_to_get());
  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_to_read(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_to_read());
  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_new_res(GtkTreeViewColumn *column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *model, GtkTreeIter *iter,
				     gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[64];
  const gchar *text;
  int num = thread->get_number_of_new_responses();
  if (num == 0)
    text = "";
  else
    {
      snprintf(text_buf, 64, "+%d", num);
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_rank_diff(GtkTreeViewColumn *column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *model, GtkTreeIter *iter,
				     gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[64];
  const gchar *text;

  int rank = thread->get_rank();
  int prev_rank = thread->get_previous_rank();
  if (rank == prev_rank || rank == 0)
    text = "";	// ѲʤǤ륹졣
  else if (prev_rank == 0)
    text = _("New Face!");
  else if (rank > prev_rank)
    text = _("down");
  else
    {
      snprintf(text_buf, 64, _("%d Rank Up"), prev_rank - rank);
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_birthtime(GtkTreeViewColumn *column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *model, GtkTreeIter *iter,
				     gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[128];
  const gchar *text;

  time_t utc = thread->get_birthtime();
  if (utc == -1)
    text = "";	// Τ͡
  else
    {
#ifdef HAVE_LOCALTIME_R
      struct tm tm;
      size_t result;
      localtime_r(&utc, &tm);
      result = strftime(text_buf, 128, "%Y-%m-%d %T", &tm);
#else	/* gdk_threads_enter()ǼƤĤ */
      struct tm *tm = localtime(&utc);
      size_t result = strftime(text_buf, 128, "%Y-%m-%d %T", tm);
#endif
      if (result == 0)
	text_buf[0] = '\0';
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}

   
void
ThreadlistView::column_set_last_modified(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[128];
  const gchar *text;

  time_t utc = thread->get_last_modified_utc();
  if (utc == -1)
    text = "";	/* Τ͡ */
  else
    {
#ifdef HAVE_LOCALTIME_R
      struct tm tm;
      size_t result;
      localtime_r(&utc, &tm);
      result = strftime(text_buf, 128, "%Y-%m-%d %T", &tm);
#else	/* gdk_threads_enter()ǼƤĤ */
      struct tm *tm = localtime(&utc);
      size_t result = strftime(text_buf, 128, "%Y-%m-%d %T", tm);
#endif
      if (result == 0)
	text_buf[0] = '\0';
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_weight(GtkTreeViewColumn *column,
				  GtkCellRenderer *renderer,
				  GtkTreeModel *model, GtkTreeIter *iter,
				  gpointer view)
{
  BBSThread *thread = NULL;
  long weight = 0;
  gtk_tree_model_get(model, iter,
		     THREAD_DATA, &thread,
		     THREAD_WEIGHT, &weight,
		     -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%ld", weight + LONG_MAX);

  column_set_text(renderer, thread, text, view);
}

   
void
ThreadlistView::column_set_search_weight(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  long weight = 0;
  gtk_tree_model_get(model, iter,
		     THREAD_DATA, &thread,
		     THREAD_SEARCH_WEIGHT, &weight,
		     -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%ld", weight + LONG_MAX);

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_clicked_cb(GtkTreeViewColumn *column,
				  OchushaTreeSelectorView *widget)
{
  gint *id_p = static_cast<gint *>(g_object_get_qdata(G_OBJECT(column),
						      sort_column_id));
  g_return_if_fail(id_p != NULL);

  GtkSortType *order_p = static_cast<GtkSortType *>(g_object_get_qdata(G_OBJECT(column),
								       sort_order_id));
  g_return_if_fail(order_p != NULL);

  ThreadlistView *view = ThreadlistView::get_view(GTK_WIDGET(widget));
  g_return_if_fail(view != NULL);

  GtkTreeSortable *sortable
    = GTK_TREE_SORTABLE(gtk_tree_view_get_model(GTK_TREE_VIEW(widget)));

  gint current_id;
  GtkSortType order;
  if (gtk_tree_sortable_get_sort_column_id(sortable, &current_id, &order))
    {
      if (current_id == *id_p)
	{
	  gtk_tree_view_column_set_sort_indicator(column, FALSE);
	  gtk_tree_sortable_set_sort_column_id(sortable,
					       SORT_BY_WEIGHT,
					       GTK_SORT_ASCENDING);

	  view->last_sort_column = view->weight_column;
	  if (view->weight_column != NULL)
	    gtk_tree_view_column_set_sort_indicator(view->weight_column, TRUE);
	}
      else
	{
	  if (view->last_sort_column != NULL)
	    gtk_tree_view_column_set_sort_indicator(view->last_sort_column,
						    FALSE);

	  gtk_tree_view_column_set_sort_indicator(column, TRUE);
	  gtk_tree_sortable_set_sort_column_id(sortable,
					       *id_p, *order_p);
	  view->last_sort_column = column;
	}
      return;
    }
  fprintf(stderr, "Why non sortable column clickable?\n");
}


void
ThreadlistView::set_sort_column_id(GtkTreeViewColumn *column, gint column_id,
				   GtkSortType order, GtkWidget *view)
{
  gint *id_p = g_new0(gint, 1);
  *id_p = column_id;
  g_object_set_qdata_full(G_OBJECT(column), sort_column_id,
			  id_p, reinterpret_cast<GDestroyNotify>(g_free));

  GtkSortType *order_p = g_new0(GtkSortType, 1);
  *order_p = order;
  g_object_set_qdata_full(G_OBJECT(column), sort_order_id,
			  order_p, reinterpret_cast<GDestroyNotify>(g_free));

  gtk_tree_view_column_set_clickable(column, TRUE);
  g_signal_connect(column, "clicked",
		   reinterpret_cast<GCallback>(column_clicked_cb), view);
}


void
ThreadlistView::restore_session(GtkTreeViewColumn *column, const char *name)
{
  SessionClient &session = controller.get_session();
  const char *value = session.get_state(name);
  if (value != NULL)
    {
      gint width;
      if (sscanf(value, "%d", &width) == 1)
	{
#if 0
	  // ݤ졽ɽʤʤ롣
	  column->use_resized_width = 1;
	  column->width = width;
#endif
	  // max_widthȤꤷƤ뤬column_resizedʥ
	  // Υϥɥ-1ꤷľƤΤmax_widthresize
	  // ¤뤳ȤϤʤ
	  gtk_tree_view_column_set_max_width(column, width);
	}
    }
}


void
ThreadlistView::append_board_name_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Board"), renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_board_name,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_BOARD, GTK_SORT_ASCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, BOARD_WIDTH);
}


void
ThreadlistView::append_title_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Thread Title"), renderer,
					       NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_title,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_TITLE, GTK_SORT_ASCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, TITLE_WIDTH);
}


void
ThreadlistView::append_rank_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Rank"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_rank,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_RANK, GTK_SORT_ASCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, RANK_WIDTH);
}


void
ThreadlistView::append_mark_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_toggle_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Mark"), renderer, NULL);
  gtk_cell_renderer_toggle_set_radio(GTK_CELL_RENDERER_TOGGLE(renderer),
				     FALSE);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_mark,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_MARK, GTK_SORT_ASCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, MARK_WIDTH);
}


void
ThreadlistView::append_n_res_verbose_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Res"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_verbose,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_VERBOSE, GTK_SORT_DESCENDING,
		     widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_VERBOSE_WIDTH);
}


void
ThreadlistView::append_n_res_verbose2_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Res"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_verbose2,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_VERBOSE2, GTK_SORT_DESCENDING,
		     widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_VERBOSE2_WIDTH);
}


void
ThreadlistView::append_n_res_avail_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Res"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_avail,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_AVAILABLE, GTK_SORT_DESCENDING,
		     widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_AVAILABLE_WIDTH);
}


void
ThreadlistView::append_n_res_read_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Got"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_read,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_READ, GTK_SORT_DESCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_READ_WIDTH);
}


void
ThreadlistView::append_n_res_ungot_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Ungot"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_ungot,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_UNGOT, GTK_SORT_DESCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_UNGOT_WIDTH);
}


void
ThreadlistView::append_n_res_to_read_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Unread"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_to_read,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_TO_READ, GTK_SORT_DESCENDING,
		     widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_TO_READ_WIDTH);
}


void
ThreadlistView::append_n_new_res_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#NewRes"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_new_res,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_NEW_RESPONSES, GTK_SORT_DESCENDING,
		     widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_NEW_RES_WIDTH);
}


void
ThreadlistView::append_rank_diff_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Rank Diffs"),
					       renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_rank_diff,
					  this, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, RANK_DIFF_WIDTH);
}


void
ThreadlistView::append_birthtime_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Birthtime"), renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_birthtime,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_BIRTHTIME, GTK_SORT_DESCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, BIRTHTIME_WIDTH);
}


void
ThreadlistView::append_last_modified_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Last Modified"),
					       renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_last_modified,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_LAST_MODIFIED, GTK_SORT_DESCENDING,
		     widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, LAST_MODIFIED_WIDTH);
}


void
ThreadlistView::append_weight_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Weight"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_weight,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_WEIGHT, GTK_SORT_ASCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, WEIGHT_WIDTH);
  weight_column = column;
}


void
ThreadlistView::append_search_weight_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("S Weight"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_search_weight,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_SEARCH_WEIGHT, GTK_SORT_ASCENDING, widget);
  gtk_tree_view_append_column(GTK_TREE_VIEW(widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, SEARCH_WEIGHT_WIDTH);
}


void
ThreadlistView::setup_columns()
{
  const char *column = controller.get_prefs().get_threadlist_view_contents();
  g_return_if_fail(strlen(column) > 0);

  while (*column != '\0')
    {
      switch (*column)
	{
	case 'A':
	  append_n_res_avail_column();
	  break;

	case 'B':
	  append_board_name_column();
	  break;

	case 'b':
	  append_birthtime_column();
	  break;

	case 'D':
	  append_rank_diff_column();
	  break;

	case 'L':
	  append_last_modified_column();
	  break;

	case 'M':
	  append_mark_column();
	  break;

	case 'N':
	  append_n_new_res_column();
	  break;

	case 'n':
	  append_n_res_read_column();
	  break;

	case 'R':
	  append_rank_column();
	  break;

	case 'T':
	  append_title_column();
	  break;

	case 'U':
	  append_n_res_ungot_column();
	  break;

	case 'u':
	  append_n_res_to_read_column();
	  break;

	case 'V':
	  append_n_res_verbose_column();
	  break;

	case 'v':
	  append_n_res_verbose2_column();
	  break;

	case 'W':
	  append_weight_column();
	  break;

	case 'w':
	  append_search_weight_column();
	  break;

	default:
	  fprintf(stderr, "'%c' doesn't match any column type\n", *column);
	  break;
	}
      column++;
    }
}


BBSThread *
ThreadlistView::get_thread(GtkTreeModel *model, GtkTreeIter *iter)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  return thread;
}


// ThreadlistView֥Ȥget_widget()֤GtkWidgetȰϡ
GtkWidget *
ThreadlistView::create(Controller &controller, BulletinBoard *board,
		       TabLabel *label)
{
  return (new ThreadlistView(controller, board, label))->widget;
}


void
ThreadlistView::append_thread(BBSThread *thread)
{
  // gdk_threads_enter()ʴĶƤӽФ٤
  SchemeManager &scheme = controller.get_scheme_manager();
  long weight = 0;
  if (rule_info != NULL)
    {
      Cell *cell = scheme.get_core().applyf(rule_info->get_procedure(),
					    "%p", thread);
      if (cell != NULL)
	{
	  if (cell->is_number())
	    weight = cell->get_integer() - LONG_MAX;
	  else if (cell == Cell::f())
	    return;
	}
    }

  shown_threads.push_back(thread);
  GtkTreeIter iter;
  gtk_list_store_insert_with_values(reinterpret_cast<GtkListStore *>(model),
				    &iter,
				    G_MAXINT32,
				    THREAD_DATA, thread,
				    THREAD_COUNT, ++number_of_threads,
				    THREAD_WEIGHT, weight,
				    THREAD_SEARCH_WEIGHT, 0L,
				    -1);
}


gboolean
ThreadlistView::append_threads_on_idle(ThreadlistView *view)
{
  LockBlock lock_obj(Lock::get_global_lock());
  gdk_threads_enter();
  smart_ptr<ThreadlistIterator> iter = view->list_iter;
  gboolean result;

  // NĤ⤷̵ʤޤɲ
  int count = 0;
  while (iter->has_more_threads())
    {
      view->append_thread(iter->next_thread());
      if (++count >= MAX_NUM_OF_THREADS_TO_APPEND_AT_ONCE)
	break;
    }

  if (!iter->has_more_threads())
    {
      view->append_threads_id = 0;
      result = FALSE;
      if (iter->finished())
	view->list_iter = NULL;
    }
  else
    result = TRUE;

  gdk_threads_leave();

  return result;
}


void
ThreadlistView::state_changed_cb()
{
  // sigslotʥϥɥ
  gdk_threads_enter();
  if (append_threads_id == 0 && list_iter != NULL)
    {
      // idleؿϿ
      append_threads_id
	= g_idle_add_full(THREADLIST_UPDATE_PRIORITY,
			  reinterpret_cast<GSourceFunc>(append_threads_on_idle),
			  this, NULL);
    }
  gdk_threads_leave();
}


void
ThreadlistView::update(GtkWidget *widget, TabLabel *label, bool reload)
{
  // gdk_threads_enter()ĶƤ֤ȡ
  ThreadlistView *view = get_view(widget);
  assert(view != NULL);

  if (view->list_iter != NULL)
    {
      cerr << "Threadlist is busy: " << view->board->get_name() << endl;
      return;	// 
    }

  Controller &controller = view->controller;
  // ϡ1ܤ̣ʤ
  view->board->read_threadlist_xml(controller.get_repository());

  const char *rule_name = view->board->get_rule();
  const OchushaProcedure *procedure
    = controller.get_scheme_manager().ochusha_procedure;
  if (rule_name != NULL)
    view->rule_info = procedure->lookup_threadlist_rule_info(rule_name);
  else
    view->rule_info = NULL;

  if (view->rule_info == NULL)
    view->rule_info = procedure->lookup_threadlist_rule_info(controller.get_prefs().get_default_threadlist_rule());

  gtk_list_store_clear(GTK_LIST_STORE(view->model));
  view->number_of_threads = 0;

  while (view->shown_threads.size() > 0)
    view->shown_threads.pop_back();

  gdk_threads_leave();
  view->list_iter = view->board->get_threadlist(controller.get_config(),
						controller.get_repository(),
						controller.get_thread_pool(),
						controller.get_broker(),
						reload);

  view->list_iter->access_started.connect(label, &TabLabel::access_started);
  view->list_iter->access_progressed.connect(label,
					     &TabLabel::access_progressed);
  view->list_iter->access_finished.connect(label, &TabLabel::access_finished);
  view->list_iter->access_failed.connect(label, &TabLabel::access_failed);
  view->list_iter->access_terminated.connect(label,
					     &TabLabel::access_terminated);

  view->list_iter->state_changed.connect(view,
					 &ThreadlistView::state_changed_cb);

  if (view->list_iter->finished())
    label->access_finished();

  gdk_threads_enter();

  if (view->append_threads_id == 0 && view->list_iter != NULL)
    view->append_threads_id
      = g_idle_add_full(THREADLIST_UPDATE_PRIORITY,
			reinterpret_cast<GSourceFunc>(append_threads_on_idle),
			view, NULL);
}


ThreadlistView *
ThreadlistView::get_view(GtkWidget *widget)
{
  return static_cast<ThreadlistView *>(g_object_get_qdata(G_OBJECT(widget),
							  view_id));
}  


GtkTreeSelection *
ThreadlistView::get_selection()
{
  return gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
}


void
ThreadlistView::initialize_for_drag_source()
{
  int buttons = GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK;
  int actions = GDK_ACTION_COPY;

  gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(widget),
					 static_cast<GdkModifierType>(buttons),
					 all_targets, G_N_ELEMENTS(all_targets),
					 static_cast<GdkDragAction>(actions));
  g_signal_connect(widget, "drag_data_get",
		   reinterpret_cast<GCallback>(drag_data_get_cb), this);
}


void
ThreadlistView::drag_data_get_cb(GtkWidget *widget,
				 GdkDragContext *drag_context,
				 GtkSelectionData *data, guint info, guint time,
				 ThreadlistView *view)
{
  g_return_if_fail(GTK_IS_TREE_VIEW(widget));
  view->selection_data.set_data(data, info);
}


gboolean
ThreadlistView::row_button_press_cb(OchushaTreeSelectorView *widget,
				    GtkTreePath *path,
				    GtkTreeViewColumn *column,
				    GdkEventButton *event,
				    ThreadlistView *view)
{
  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  if (event->button == 3)
    {
      GtkTreeIter iter;
      GtkTreeView *tree_view = GTK_TREE_VIEW(widget);
      GtkTreeModel *model = gtk_tree_view_get_model(tree_view);

      if (!gtk_tree_model_get_iter(model, &iter, path))
	return FALSE;

      view->press_start_x = static_cast<gint>(event->x);
      view->press_start_y = static_cast<gint>(event->y);

      view->press_start_x_root = static_cast<gint>(event->x_root);
      view->press_start_y_root = static_cast<gint>(event->y_root);
      view->press_start_button = event->button;

      if (view->popup_delay_timeout_id != 0)
	g_source_remove(view->popup_delay_timeout_id);

      view->popup_delay_timeout_id
	= g_timeout_add(POPUP_MENU_DELAY_TIMEOUT,
			reinterpret_cast<GSourceFunc>(popup_delay_timeout_cb),
			view);
    }
  else if ((event->state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) == 0)
    view->last_event_path = gtk_tree_path_copy(path);

  return FALSE;
}


void
ThreadlistView::selection_changed_cb(GtkTreeSelection *selection,
				     ThreadlistView *view)
{
  view->controller.update_internal_clipboard(view->selection_data.create_proxy());
}


gboolean
ThreadlistView::popup_menu_cb(GtkWidget *widget, ThreadlistView *view)
{
  if (view->popup_delay_timeout_id != 0)
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
    }

  view->popup_context_menu(0, gtk_get_current_event_time());
  return TRUE;
}


gboolean
ThreadlistView::motion_notify_cb(GtkWidget *widget, GdkEventMotion *event,
				 ThreadlistView *view)
{
  if (view->popup_delay_timeout_id != 0
      && gtk_drag_check_threshold(widget,
				  view->press_start_x, view->press_start_y,
				  static_cast<gint>(event->x),
				  static_cast<gint>(event->y)))
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
    }
  return FALSE;
}


void
ThreadlistView::popup_context_menu(guint button, guint32 event_time)
{
  bool show_all = controller.get_prefs().get_show_hidden_threads();
  const char *menu_root
    = show_all ? "/UglyThreadlistPopupMenu" : "/ThreadlistPopupMenu";
  ochusha_ui_manager_popup_with_data(controller.get_ui_manager(), menu_root,
				     this, NULL, NULL, NULL,
				     button, event_time);
}


gboolean
ThreadlistView::popup_delay_timeout_cb(ThreadlistView *view)
{
  gdk_threads_enter();

  view->popup_context_menu(view->press_start_button,
			   gtk_get_current_event_time());
  view->popup_delay_timeout_id = 0;

  gdk_threads_leave();
  return FALSE;
}


gboolean
ThreadlistView::row_button_release_cb(OchushaTreeSelectorView *widget,
				      GtkTreePath *path,
				      GtkTreeViewColumn *column,
				      GdkEventButton *event,
				      ThreadlistView *view)
{
  if (view->popup_delay_timeout_id != 0)
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
      view->popup_context_menu(0, event->time);
    }

  GtkTreeIter iter;
  GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(widget));

  gboolean result = FALSE;
  BBSThread *thread = NULL;
  if (view->last_event_path == NULL
      || gtk_tree_path_compare(view->last_event_path, path) != 0
      || !gtk_tree_model_get_iter(model, &iter, path))
    goto done;

  gtk_tree_model_get(model, &iter, THREAD_DATA, &thread, -1);
  if (thread != NULL)
    {
      gint *id_p = static_cast<gint *>(g_object_get_qdata(G_OBJECT(column),
							  sort_column_id));
      if (id_p != NULL && *id_p == SORT_BY_MARK)
	{
	  thread->set_marked(!thread->is_marked());
	  gtk_tree_model_row_changed(model, path, &iter);
	}
      else
	{
	  Controller &controller = view->controller;
	  bool in_tab = ((event->button == 2)
			 ^ (controller.get_prefs().get_default_open_in_tab()));
	  controller.open(thread, 0, in_tab, false);
	}
    }

 done:
  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  return result;
}


void
ThreadlistView::column_resized_cb(OchushaTreeSelectorView *widget,
				  GtkTreeViewColumn *column,
				  gboolean use_resized_width,
				  gint resized_width,
				  ThreadlistView *view)
{
  gint *id_p = static_cast<gint *>(g_object_get_qdata(G_OBJECT(column),
						      sort_column_id));
  const char *name;

  if (id_p == NULL)
    name = RANK_DIFF_WIDTH;	// rank_diffΤϤġ
  else switch (*id_p)
    {
    case SORT_BY_COUNT:
      name = COUNT_WIDTH;
      break;

    case SORT_BY_BOARD:
      name = BOARD_WIDTH;
      break;

    case SORT_BY_RANK:
      name = RANK_WIDTH;
      break;

    case SORT_BY_MARK:
      name = MARK_WIDTH;
      break;

    case SORT_BY_TITLE:
      name = TITLE_WIDTH;
      break;

    case SORT_BY_N_RES_VERBOSE:
      name = N_RES_VERBOSE_WIDTH;
      break;

    case SORT_BY_N_RES_VERBOSE2:
      name = N_RES_VERBOSE2_WIDTH;
      break;

    case SORT_BY_N_RES_AVAILABLE:
      name = N_RES_AVAILABLE_WIDTH;
      break;

    case SORT_BY_N_RES_READ:
      name = N_RES_READ_WIDTH;
      break;

    case SORT_BY_N_RES_UNGOT:
      name = N_RES_UNGOT_WIDTH;
      break;

    case SORT_BY_N_RES_TO_READ:
      name = N_RES_TO_READ_WIDTH;
      break;

    case SORT_BY_N_NEW_RESPONSES:
      name = N_NEW_RES_WIDTH;
      break;

    case SORT_BY_BIRTHTIME:
      name = BIRTHTIME_WIDTH;
      break;

    case SORT_BY_LAST_MODIFIED:
      name = LAST_MODIFIED_WIDTH;
      break;

    case SORT_BY_WEIGHT:
      name = WEIGHT_WIDTH;
      break;

    case SORT_BY_SEARCH_WEIGHT:
      name = SEARCH_WEIGHT_WIDTH;
      break;

    default:
      name = NULL;
    }

  if (name != NULL)
    {
      SessionClient &session = view->controller.get_session();
      if (use_resized_width)
	{
	  char buf[64];
	  snprintf(buf, 64, "%d", resized_width);
	  session.set_state(name, buf);
	}
      else
	session.remove_state(name);
    }

  gtk_tree_view_column_set_max_width(column, -1);
  if (name == NULL)
    fprintf(stderr, "Cannot correctly handle id=%d\n", *id_p);
}


static int
compare_count(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	      gpointer unused)
{
  int a_count;
  int b_count;
  gtk_tree_model_get(model, a, THREAD_COUNT, &a_count, -1);
  gtk_tree_model_get(model, b, THREAD_COUNT, &b_count, -1);
  return a_count - b_count;
}


static int
compare_board(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;
  result = strcmp(a_thread->get_board()->get_name(),
		  b_thread->get_board()->get_name());
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_rank(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	     gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;
  if (a_thread->is_alive() && b_thread->is_alive())
    result = a_thread->get_rank() - b_thread->get_rank();
  else if (a_thread->is_alive())
    return -1;
  else if (b_thread->is_alive())
    return 1;
  else
    return compare_title(model, a, b, unused);

  if (result != 0)
    return result;

  return compare_title(model, a, b, unused);
}


static int
compare_mark(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	     gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;
  result = b_thread->is_marked() - a_thread->is_marked();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_title(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  const char *a_title = a_thread->get_title();
  const char *b_title = b_thread->get_title();

  if (a_title == NULL && b_title == NULL)
    return 0;
  if (a_title == NULL)
    return 1;
  if (b_title == NULL)
    return -1;

  result = strcmp(a_title, b_title);
  if (result != 0)
    return result;
  return compare_count(model, a, b, unused);
}


static int
compare_n_res_verbose(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  int a_read = a_thread->get_number_of_responses_got();
  int b_read = b_thread->get_number_of_responses_got();
  if ((a_read > 0 || b_read > 0) && (a_read == 0 || b_read == 0))
    return a_read - b_read;

  result = a_thread->get_number_of_responses_to_read()
    - b_thread->get_number_of_responses_to_read();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_verbose2(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		       gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  int a_read = a_thread->get_number_of_responses_got();
  int b_read = b_thread->get_number_of_responses_got();
  if ((a_read > 0 || b_read > 0) && (a_read == 0 || b_read == 0))
    return a_read - b_read;

  result = a_thread->get_number_of_responses_to_get()
    - b_thread->get_number_of_responses_to_get();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_avail(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		    gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_available()
    - b_thread->get_number_of_responses_available();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_read(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		   gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_got()
    - b_thread->get_number_of_responses_got();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_ungot(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		    gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_to_get()
    - b_thread->get_number_of_responses_to_get();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_to_read(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_to_read()
    - b_thread->get_number_of_responses_to_read();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_n_new_res(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		  gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_new_responses()
    - b_thread->get_number_of_new_responses();
  if (result != 0)
    return result;
  return compare_weight(model, a, b, unused);
}


static int
compare_birthtime(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		  gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  time_t result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_birthtime() - b_thread->get_birthtime();
  if (result != 0)
    return static_cast<int>(result);
  return compare_title(model, a, b, unused);
}


static int
compare_last_modified(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  time_t result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_last_modified_utc()
    - b_thread->get_last_modified_utc();
  if (result != 0)
    return static_cast<int>(result);
  return compare_weight(model, a, b, unused);
}


static int
compare_weight(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	       gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  long a_weight;
  long b_weight;
  int result = 0;
  gtk_tree_model_get(model, a,
		     THREAD_DATA, &a_thread,
		     THREAD_WEIGHT, &a_weight,
		     -1);
  gtk_tree_model_get(model, b,
		     THREAD_DATA, &b_thread,
		     THREAD_WEIGHT, &b_weight,
		     -1);
  result = static_cast<int>(a_weight - b_weight);
  if (result != 0)
    return result;

  if (a_thread->is_alive() || b_thread->is_alive())
    return compare_count(model, a, b, unused);
  return compare_title(model, a, b, unused);
}


static int
compare_search_weight(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  long a_weight;
  long b_weight;
  int result = 0;
  gtk_tree_model_get(model, a,
		     THREAD_DATA, &a_thread,
		     THREAD_SEARCH_WEIGHT, &a_weight,
		     -1);
  gtk_tree_model_get(model, b,
		     THREAD_DATA, &b_thread,
		     THREAD_SEARCH_WEIGHT, &b_weight,
		     -1);
  result = static_cast<int>(a_weight - b_weight);
  if (result != 0)
    return result;

  if (a_thread->is_alive() || b_thread->is_alive())
    return compare_count(model, a, b, unused);
  return compare_title(model, a, b, unused);
}


}
